package edu.uwyo;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.File;

import edu.uwyo.crawl.Query;
import edu.uwyo.crawl.QueryParams;
import edu.uwyo.crawl.database.Database;
import edu.uwyo.crawl.database.ExcludeList;
import edu.uwyo.crawl.ui.FrameMain;
import net.javamon.dev.Console;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;

/**
 * The main program for running user-configured searches on directories and files.
 * Make sure to call {@link #quit(File)} whenever the application exits.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public final class Main {
    /** Don't allow construction of this class. */
    private Main() { throw new AssertionError(); }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public static FrameMain frameMain;

    private static Database database;
    private static File currSnapshot;
    private static boolean allowWindowKill;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** The main program. */
    public static void main(String[] args) {
        Resources.init(".", "fileCrawler");
        Console.init();
        Database.init();
        Console.windowLoad.addKeyListener(new WindowKillListener());
        database = new Database(new Query(), Query.INSTANCE);

        if (args.length > 0) {
            commandMode(args);
            System.exit(0);
        }

        String developer = Resources.readUserSetting("developerMode");
        if (developer != null) {
            database.setDeveloperMode(Boolean.parseBoolean(developer));
        }

        Console.windowLoad.start(null, null);
        new Thread(new Runnable() {
            @Override
            public void run() {
                String lastSnapshot = Resources.readUserSetting("lastSnapshot");
                if (lastSnapshot != null) {
                    currSnapshot = new File(lastSnapshot);
                    if (!currSnapshot.exists() || !database.loadSearchResults(currSnapshot)) {
                        currSnapshot = null;
                        Resources.removeUserSetting("lastSnapshot");
                    }
                }

                frameMain = new FrameMain();
                Console.windowLoad.stop();
                frameMain.show(currSnapshot);
            }
        }).start();
    }

    /** @return the {@link Database} in which search results will be stored */
    public static Database database() {
        return database;
    }

    /** @return the {@link Query} used for searching */
    public static Query query() {
        return (Query) database.results().searchControl();
    }

    /**
     * @return the {@link File} representing the current results snapshot,
     *  or {@code null} if none
     */
    public static File currSnapshot() {
        return currSnapshot;
    }

    /** @return the {@link ExcludeList}, or {@code null} if none */
    public static ExcludeList excludeList() {
        return database.results().excludeList();
    }

    /**
     * Sets whether the loading window can be "killed".
     * If {@code true}, triggering the appropriate key binding will kill the entire application.
     * @param allowKill a boolean, {@code true} if the loading window can be killed
     */
    public static void allowWindowKill(boolean allowKill) {
        allowWindowKill = allowKill;
    }

    /**
     * Loads the given search results snapshot.
     * If the current results cannot be saved or the new snapshot cannot be loaded,
     *  an error message will be displayed and this method will do nothing.
     * @param snapshot the {@link File} of the search snapshot to load
     * @param destination the {@link File} where the current results should be saved,
     *  or {@code null} if the current results should not be saved
     */
    public static void loadSnapshot(final File snapshot, final File destination) {
        if (frameMain != null) { frameMain.hide(); }
        Console.windowLoad.start(Resources.__("mainStartLoading"), null);
        new Thread(new Runnable() {
            @Override
            public void run() {
                if (destination != null && frameMain != null) {
                    frameMain.saveView();
                }

                if (destination == null || database.saveSearchResults(destination)) {
                    if (database.loadSearchResults(snapshot)) {
                        currSnapshot = snapshot;
                        Resources.writeUserSetting("lastSnapshot", currSnapshot.getAbsolutePath());
                        if (frameMain != null) { frameMain.save(); }
                    } else {
                        Prompt.message(Resources.__("mainSnapshotLoadErrorH"),
                                Resources.__("mainSnapshotLoadErrorMsg", snapshot.getAbsolutePath()));
                    }
                } else {
                    Prompt.message(Resources.__("mainSnapshotSaveErrorH"),
                            Resources.__("mainSnapshotSaveErrorMsg", snapshot.getAbsolutePath()));
                }
                Console.windowLoad.stop();
                if (frameMain != null) { frameMain.show(currSnapshot); }
            }
        }).start();
    }

    /**
     * Saves the current search results and settings to the given destination.
     * @param snapshot the {@link File} where the current search results should be saved
     */
    public static void saveSnapshot(final File snapshot) {
        frameMain.hide();
        Console.windowLoad.start(null, null);
        new Thread(new Runnable() {
            @Override
            public void run() {
                frameMain.saveView();
                if (database.saveSearchResults(snapshot)) {
                    currSnapshot = snapshot;
                }

                if (currSnapshot != null) {
                    Resources.writeUserSetting("lastSnapshot", currSnapshot.getAbsolutePath());
                }

                Console.windowLoad.stop();
                frameMain.save();
                frameMain.show(currSnapshot);
            }
        }).start();
    }

    /**
     * Reads a setting.
     * @param setting the String name of the setting to retrieve
     * @return the String setting value, or {@code null} if it has not been set
     */
    public static String readSetting(String setting) {
        return database.results().readSetting(setting);
    }

    /**
     * Writes a setting.
     * @param setting the String name of the setting to write
     * @param value the String value of the setting to write
     */
    public static void writeSetting(String setting, String value) {
        database.results().writeSetting(setting, value);
    }

    /**
     * Obtains the flag associated with the given file path.
     * @param path the String file path whose flag should be obtained
     * @return the {@link Byte} flag associated with {@code path},
     *  or {@code null} if no flag has been set
     */
    public static Byte getFlag(String path) {
        return database.results().getFlag(path);
    }

    /**
     * Associates a flag with the given file path.
     * If {@code path} already has a flag, that flag will be overwritten.
     * @param path the String file path to be given a flag
     * @param flag the byte flag to set
     */
    public static void setFlag(String path, byte flag) {
        database.results().setFlag(path, flag);
    }

    /**
     * Removes the current flag associated with {@code path}, if any.
     * @param path the String file path whose flag should be removed
     */
    public static void removeFlag(String path) {
        database.results().removeFlag(path);
    }

    /**
     * Removes all flags for all file paths.
     * After this method is called, {@link #getFlag(String)} will return {@code null}
     *  for <i>any and every</i> value of {@code path}.
     */
    public static void removeAllFlags() {
        database.results().removeAllFlags();
    }

    /**
     * Quits the application by exporting the current search results.
     * <b>This method <u>MUST</u> be called any time the application exits!</b>
     * <p>
     * Automatically hides the main application frame.
     * @param snapshot the {@link File} where the current search results should be saved,
     *  or {@code null} if results should not be saved
     */
    public static void quit(final File snapshot) {
        if (snapshot == null) {
            if (currSnapshot != null) {
                Resources.writeUserSetting("lastSnapshot", currSnapshot.getAbsolutePath());
            }
            System.exit(0);
        }

        frameMain.hide();
        Console.windowLoad.start(Resources.__("mainExitLoading"), null);
        new Thread(new Runnable() {
            @Override
            public void run() {
                if (snapshot != null) {
                    frameMain.saveView();
                    if (database.saveSearchResults(snapshot)) {
                        currSnapshot = snapshot;
                    }
                }

                if (currSnapshot != null) {
                    Resources.writeUserSetting("lastSnapshot", currSnapshot.getAbsolutePath());
                }

                Console.windowLoad.stop();
                System.exit(0);
            }
        }).start();
    }

    /**
     * Runs this application on the command-line.
     * No graphical components will be created.
     * @param args the array of String command-line tokens to parse
     */
    private static void commandMode(String[] args) {
        QueryParams queryParams = new QueryParams(database);
        if (queryParams.parse(args)) {
            queryParams.run();
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CLASSES                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** A listener for killing the loading window. */
    private static class WindowKillListener implements KeyListener {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public void keyTyped(KeyEvent e) {}

        @Override
        public void keyPressed(KeyEvent e) {
            if (e.getKeyCode() == KeyEvent.VK_ESCAPE && allowWindowKill) {
                System.exit(1);
            }
        }

        @Override
        public void keyReleased(KeyEvent e) {}
    }
}
