package edu.uwyo.crawl;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.uwyo.Main;
import edu.uwyo.crawl.contentMatcher.ContentMatcher;
import edu.uwyo.crawl.database.Database;
import edu.uwyo.crawl.database.DuplicateFinder;
import edu.uwyo.crawl.database.FileSearchControl;
import edu.uwyo.crawl.ui.panels.PanelOptions;
import edu.uwyo.crawl.ui.panels.PanelOptionsFiles;
import edu.uwyo.crawl.ui.panels.PanelOptionsSize;
import edu.uwyo.crawl.ui.panels.PanelTime;
import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import net.javamon.dev.Ext;
import net.javamon.dev.FileUtils;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;

/**
 * A representation of the current search parameters that have been chosen by the user.
 * Call {@link #update()} to fetch parameters from the current display
 *  and {@link #run(Database, boolean)} to perform a search using those parameters.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class Query implements FileSearchControl {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used for serialization; <b>do not modify!</b> */
    public static final Query INSTANCE = new Query();

    /**
     * Indicates that the file's size fell within the expected range.
     * Used in conjunction with {@link #getSizeDescriptor(String)}.
     * <p>
     * This is the default value returned for all files,
     *  regardless of whether size-checking is enabled.
     */
    public static final int SIZE_OKAY = 0;

    /**
     * Indicates that the file's size is probably too small/large.
     * Used in conjunction with {@link #getSizeDescriptor(String)}.
     */
    public static final int SIZE_WARNING = 1;

    /**
     * Indicates that the file's size is far too small/large.
     * Used in conjunction with {@link #getSizeDescriptor(String)}.
     */
    public static final int SIZE_ERROR = 2;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used only by command-line searches. */
    protected int headerSize;

    /** Used only by command-line searches. */
    protected int bufferNet;

    /** Used only by command-line searches. */
    protected int bufferLocal;

    /** Used only by command-line searches. */
    protected int chunkSize;

    /** Used only by command-line searches. */
    protected String[] excludedExtensions;

    /** Used only by command-line searches. */
    protected String[] excludedFiles;

    /** Used only by command-line searches. */
    protected Pattern excludedPattern;

    protected File toSearch;
    protected int nesting;
    protected String[] extensions;

    protected Pattern patternName;
    protected Pattern patternContent;

    protected int duplicateMode;
    protected boolean duplicateExact;

    protected long createdStart;
    protected long createdEnd;

    protected long modifiedStart;
    protected long modifiedEnd;

    protected long accessedStart;
    protected long accessedEnd;

    protected long sizeSmallerWarning;
    protected long sizeSmallerError;

    protected long sizeLargerWarning;
    protected long sizeLargerError;

    protected long sizeSmaller;
    protected long sizeLarger;

    protected String fileMode;
    private boolean isCommandLine;

    private final ContentMatcher contentMatcher;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link Query}. */
    public Query() {
        contentMatcher = new ContentMatcher();
        clear(null);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public boolean isSearchable(String path, BasicFileAttributes attributes) {
        if (exclude(path)) { return false; }
        if (fileMode != null && !fileMode.equals(path)) { return false; } //only allow selected file

        try {
            return containsName(path) && withinDate(attributes) && withinSize(attributes)
                    && containsContent(path);
        } catch (OutOfMemoryError e) {
            e.printStackTrace();
            if (!isCommandLine) {
                try {
                    Prompt.errorScroll(Resources.__("mainOOMH"),
                            Resources.__("mainOOMMsg", Main.database().numSearched(), path));
                } catch (OutOfMemoryError e1) {} //better to exit program than hang if message fails
            }
            System.exit(1);
            return false;
        } catch (Exception e) {
            return false;
        }
    }

    @Override
    public Path root() {
        return Paths.get(toSearch.getAbsolutePath());
    }

    @Override
    public int numLevels() {
        return nesting;
    }

    @Override
    public void clear(File file) {
        isCommandLine = false;
        toSearch = file;
        fileMode = null;
        contentMatcher.setPattern(null);

        nesting = Integer.MAX_VALUE;
        extensions = null;
        patternName = null;
        patternContent = null;
        duplicateMode = 0;
        duplicateExact = false;
        createdStart = -1L;
        createdEnd = -1L;
        modifiedStart = -1L;
        modifiedEnd = -1L;
        accessedStart = -1L;
        accessedEnd = -1L;
        sizeSmallerWarning = -1L;
        sizeSmallerError = -1L;
        sizeLargerWarning = -1L;
        sizeLargerError = -1L;
        sizeSmaller = -1L;
        sizeLarger = -1L;
    }

    @Override
    public void update() {
        if (Main.frameMain == null) { return; }
        File oldFile = toSearch;
        clear(new File(Main.frameMain.main().options.folder.getText()));
        Main.excludeList().setFiles(Main.frameMain.exclude().main().files.getExcludedFiles());
        Main.excludeList().setExt(Main.frameMain.exclude().main().extensions.getExcludedTypes());
        Main.excludeList().setPattern(Main.frameMain.exclude().main().regex.getRegex());

        nesting = Main.frameMain.main().options.recurse.getSelectedIndex();
        if (nesting == PanelOptions.RECURSE_INFINITE) {
            nesting = Integer.MAX_VALUE;
        } else if (nesting == PanelOptions.RECURSE_FILE) {
            fileMode = toSearch.getAbsolutePath();
            toSearch = toSearch.getParentFile();
            nesting = 1; //force search of enclosing folder to ensure proper result tree display
        } else if (nesting == PanelOptions.RECURSE_RESULTS) {
            toSearch = oldFile;
        }

        String fileTypes = Main.frameMain.main().files.fileTypes.getText();
        if (fileTypes.length() == 0 || fileTypes.equals(PanelOptionsFiles.ALL_FILES)) {
            extensions = null;
        } else {
            extensions = fileTypes.split(PanelOptionsFiles.DELIMITERS);
        }

        String nameText = Main.frameMain.main().options.patternName.getText();
        if (nameText.length() > 0) {
            patternName = Pattern.compile(nameText);
        } else {
            patternName = null;
        }

        String contentText = Main.frameMain.main().options.patternContent.getText();
        if (contentText.length() > 0) {
            patternContent = Pattern.compile(contentText);
            contentMatcher.setPattern(patternContent);
        } else {
            patternContent = null;
        }

        duplicateMode = Main.frameMain.main().options.duplicate.getSelectedIndex();
        duplicateExact = Main.frameMain.main().options.duplicateExact.isSelected();

        if (Main.frameMain.main().time.modes[PanelTime.CREATED].enabled.isSelected()) {
            PanelTime created = Main.frameMain.main().time.modes[PanelTime.CREATED];
            switch (created.comparison.getSelectedIndex()) {
                case PanelTime.OLDER:   createdStart = -1L;
                                        createdEnd = created.getTime(created.dateMax);
                                        break;

                case PanelTime.NEWER:   createdStart = created.getTime(created.dateMax);
                                        createdEnd = -1L;
                                        break;

                default:                createdStart = created.getTime(created.dateMin);
                                        createdEnd = created.getTime(created.dateMax);
                                        break;
            }
        }

        if (Main.frameMain.main().time.modes[PanelTime.MODIFIED].enabled.isSelected()) {
            PanelTime modified = Main.frameMain.main().time.modes[PanelTime.MODIFIED];
            switch (modified.comparison.getSelectedIndex()) {
                case PanelTime.OLDER:   modifiedStart = -1L;
                                        modifiedEnd = modified.getTime(modified.dateMax);
                                        break;

                case PanelTime.NEWER:   modifiedStart = modified.getTime(modified.dateMax);
                                        modifiedEnd = -1L;
                                        break;

                default:                modifiedStart = modified.getTime(modified.dateMin);
                                        modifiedEnd = modified.getTime(modified.dateMax);
                                        break;
            }
        }

        if (Main.frameMain.main().time.modes[PanelTime.ACCESSED].enabled.isSelected()) {
            PanelTime accessed = Main.frameMain.main().time.modes[PanelTime.ACCESSED];
            switch (accessed.comparison.getSelectedIndex()) {
                case PanelTime.OLDER:   accessedStart = -1L;
                                        accessedEnd = accessed.getTime(accessed.dateMax);
                                        break;

                case PanelTime.NEWER:   accessedStart = accessed.getTime(accessed.dateMax);
                                        accessedEnd = -1L;
                                        break;

                default:                accessedStart = accessed.getTime(accessed.dateMin);
                                        accessedEnd = accessed.getTime(accessed.dateMax);
                                        break;
            }
        }

        if (Main.frameMain.main().size.mode.getSelectedIndex() != PanelOptionsSize.NONE) {
            PanelOptionsSize size = Main.frameMain.main().size;
            switch (size.mode.getSelectedIndex()) {
                case PanelOptionsSize.SMALLER:  sizeSmallerWarning = size.getWarningBytes();
                                                sizeSmallerError = size.getErrorBytes();
                                                break;

                case PanelOptionsSize.LARGER:   sizeLargerWarning = size.getWarningBytes();
                                                sizeLargerError = size.getErrorBytes();
                                                break;

                case PanelOptionsSize.BETWEEN:  sizeSmaller = size.getWarningBytes();
                                                sizeLarger = size.getErrorBytes();
                                                break;
            }
        }
    }

    @Override
    public void loadSearchHash(String hash) {
        if (hash == null) { return; }
        String[] tokens = hash.split(HASH_DELIMITER);
        if (tokens.length < 18) {
            return;
        }

        toSearch = tokens[0].equals("null") ? null : new File(tokens[0]);
        nesting = Integer.parseInt(tokens[1]);
        patternName = tokens[2].equals("null") ? null : Pattern.compile(tokens[2]);
        patternContent = tokens[3].equals("null") ? null : Pattern.compile(tokens[3]);
        duplicateMode = Integer.parseInt(tokens[4]);
        duplicateExact = Boolean.parseBoolean(tokens[5]);
        createdStart = Long.parseLong(tokens[6]);
        createdEnd = Long.parseLong(tokens[7]);
        modifiedStart = Long.parseLong(tokens[8]);
        modifiedEnd = Long.parseLong(tokens[9]);
        accessedStart = Long.parseLong(tokens[10]);
        accessedEnd = Long.parseLong(tokens[11]);
        sizeSmallerWarning = Long.parseLong(tokens[12]);
        sizeSmallerError = Long.parseLong(tokens[13]);
        sizeLargerWarning = Long.parseLong(tokens[14]);
        sizeLargerError = Long.parseLong(tokens[15]);
        sizeSmaller = Long.parseLong(tokens[16]);
        sizeLarger = Long.parseLong(tokens[17]);

        extensions = null;
        if (tokens.length > 18) {
            extensions = new String[tokens.length - 18];
            for (int i = 0; i < extensions.length; i++) {
                extensions[i] = tokens[18 + i];
            }
        }
    }

    @Override
    public String getSearchHash() {
        StringBuilder builder = new StringBuilder();
        builder.append((toSearch == null) ? "null" : toSearch.getAbsolutePath());
        builder.append(HASH_DELIMITER);
        builder.append(nesting);
        builder.append(HASH_DELIMITER);
        builder.append((patternName == null) ? "null" : patternName);
        builder.append(HASH_DELIMITER);
        builder.append((patternContent == null) ? "null" : patternContent);
        builder.append(HASH_DELIMITER);
        builder.append(duplicateMode);
        builder.append(HASH_DELIMITER);
        builder.append(duplicateExact);
        builder.append(HASH_DELIMITER);
        builder.append(createdStart);
        builder.append(HASH_DELIMITER);
        builder.append(createdEnd);
        builder.append(HASH_DELIMITER);
        builder.append(modifiedStart);
        builder.append(HASH_DELIMITER);
        builder.append(modifiedEnd);
        builder.append(HASH_DELIMITER);
        builder.append(accessedStart);
        builder.append(HASH_DELIMITER);
        builder.append(accessedEnd);
        builder.append(HASH_DELIMITER);
        builder.append(sizeSmallerWarning);
        builder.append(HASH_DELIMITER);
        builder.append(sizeSmallerError);
        builder.append(HASH_DELIMITER);
        builder.append(sizeLargerWarning);
        builder.append(HASH_DELIMITER);
        builder.append(sizeLargerError);
        builder.append(HASH_DELIMITER);
        builder.append(sizeSmaller);
        builder.append(HASH_DELIMITER);
        builder.append(sizeLarger);

        if (extensions != null) {
            for (int i = 0; i < extensions.length; i++) {
                builder.append(HASH_DELIMITER);
                builder.append(extensions[i]);
            }
        }
        return builder.toString();
    }

    @Override
    public Query create() {
        return new Query();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Performs a search using the current query settings.
     * <p>
     * <i>This method may perform a very large number of I/O operations.
     * Do <u>not</u> run this method from the main thread or the application might freeze!</i>
     * @param database the {@link Database} to use when running the search
     * @param isCommandLine a boolean, {@code true} if this query
     *  is being run from the command-line
     * @return the {@link FileDisplayTree} of files that were found by the search
     */
    public FileDisplayTree run(Database database, boolean isCommandLine) {
        this.isCommandLine = isCommandLine;
        if (isCommandLine) {
            DuplicateFinder.setHeaderSize(headerSize);
            if (FileUtils.isRemote(toSearch)) {
                contentMatcher.tuneParameters(bufferNet, chunkSize);
            } else {
                contentMatcher.tuneParameters(bufferLocal, chunkSize);
            }
            Main.excludeList().setFiles(excludedFiles);
            Main.excludeList().setExt(excludedExtensions);
            Main.excludeList().setPattern(excludedPattern);

        } else {
            DuplicateFinder.setHeaderSize(Main.frameMain.developer().main().getHeaderSize());
            if (FileUtils.isRemote(toSearch)) {
                contentMatcher.tuneParameters(Main.frameMain.developer().main().getBufferNet(),
                        Main.frameMain.developer().main().getChunkSize());
            } else {
                contentMatcher.tuneParameters(Main.frameMain.developer().main().getBufferLocal(),
                        Main.frameMain.developer().main().getChunkSize());
            }
        }

        final boolean curr = (nesting == PanelOptions.RECURSE_RESULTS);
        if (curr && !isCommandLine) {
            Main.frameMain.main().options.folder.setText(toSearch.toString());
            Main.frameMain.main().options.saveView();
        }

        switch (duplicateMode) {
            case PanelOptions.DUPLICATE_NAME:       return database.findDuplicates(false,
                                                            true, false, duplicateExact, curr,
                                                            null, null, null);
            case PanelOptions.DUPLICATE_CONTENT:    return database.findDuplicates(false,
                                                            false, true, duplicateExact, curr,
                                                            null, null, null);
            case PanelOptions.DUPLICATE_ALL:        return database.findDuplicates(false,
                                                            true, true, duplicateExact, curr,
                                                            null, null, null);
            default:                                return database.search(false, curr);
        }
    }

    /**
     * Performs a duplicate search using the current query settings.
     * Use this method only when performing a search
     *  on a selected subset of files {@code searchPaths}.
     * All other searches should use {@link #run(Database, boolean)}.
     * <i>This method cannot be used on the command-line
     *  as it requires a user selection of files.</i>
     * <p>
     * Because the files to search have already been provided,
     *  most query settings will be <i>ignored</i> in this method.
     * However, changes in the performance tuning will still be applied as needed.
     * @param database the {@link Database} to use when running the search
     * @param searchPaths the array of String paths to search, or {@code null} if current
     *  or new search results should be used;
     *  if this is not {@code null}, an equal-length {@code searchSizes} must also be provided
     * @param searchSizes the array of long sizes to search
     * @param searchRoot the String search root; only used if {@code searchPaths != null}
     * @param names a boolean, {@code true} if files must have identical names
     *  to be included in the results
     * @param content a boolean, {@code true} if files must have similar content
     *  to be included in the results
     * @param exact a boolean, {@code true} if files must have <u>identical</u> contents
     *  to be considered a match; otherwise, files are only compared using their header bytes
     * @return the {@link FileDisplayTree} of files that were found by the search
     */
    public FileDisplayTree runDuplicates(Database database, String[] searchPaths,
            long[] searchSizes, String searchRoot, boolean names, boolean content, boolean exact) {
        isCommandLine = false;
        DuplicateFinder.setHeaderSize(Main.frameMain.developer().main().getHeaderSize());
        if (FileUtils.isRemote(toSearch)) {
            contentMatcher.tuneParameters(Main.frameMain.developer().main().getBufferNet(),
                    Main.frameMain.developer().main().getChunkSize());
        } else {
            contentMatcher.tuneParameters(Main.frameMain.developer().main().getBufferLocal(),
                    Main.frameMain.developer().main().getChunkSize());
        }

        return database.findDuplicates(false, names, content, exact, false, searchPaths,
                searchSizes, searchRoot);
    }

    /**
     * Obtains a size descriptor for the given file,
     *  used to highlight how severely the file's size falls outside a desired range.
     * If no size checking was enabled, this method always returns {@link #SIZE_OKAY}.
     * <p>
     * Should only be called after at least one call to {@link #run(Database, boolean)}.
     * Performs no I/O operations.
     * @param path the String file path for which to obtain a size descriptor
     * @return the integer size descriptor: {@link #SIZE_OKAY}, {@link #SIZE_WARNING},
     *  or {@link #SIZE_ERROR}
     */
    public int getSizeDescriptor(String path) {
        final long size = new File(path).length();

        if (sizeSmaller >= 0L && sizeLarger >= 0L) {
            if (!withinRange(size, sizeSmaller, sizeLarger)) {
                return SIZE_ERROR;
            }

        } else if (sizeSmallerWarning >= 0L && sizeSmallerError >= 0L) {
            if (withinRange(size, -1L, sizeSmallerError)) {
                return SIZE_ERROR;
            } else if (withinRange(size, -1L, sizeSmallerWarning)) {
                return SIZE_WARNING;
            }

        } else if (sizeLargerWarning >= 0L && sizeLargerError >= 0L) {
            if (withinRange(size, sizeLargerError, -1L)) {
                return SIZE_ERROR;
            } else if (withinRange(size, sizeLargerWarning, -1L)) {
                return SIZE_WARNING;
            }
        }

        return SIZE_OKAY;
    }

    /**
     * Determines whether a file should be manually excluded from the search.
     * If there are no exclusion lists, this method always returns {@code false}.
     * @param path the String absolute file path to check
     * @return a boolean, {@code true} if the file should be excluded from the search
     */
    private boolean exclude(String path) {
        return Main.excludeList().isFileExcluded(path, Ext.get(path));
    }

    /**
     * Determines whether the file's name (including extension) matches any constraints.
     * If name or extension matching have been disabled, this method always returns {@code true}.
     * @param path the String file path to check
     * @return a boolean, {@code true} if the file's name/extension match the current constraints
     */
    private boolean containsName(String path) {
        boolean endsWith = (extensions == null);
        if (extensions != null) {
            for (int i = 0; i < extensions.length; i++) {
                if (path.toLowerCase().endsWith(extensions[i])) {
                    endsWith = true;
                    break;
                }
            }
        }

        boolean matches = true;
        String name = new File(path).getName();
        if (patternName != null) {
            Matcher m = patternName.matcher(name);
            matches = m.find();
        }

        return endsWith && matches;
    }

    /**
     * Determines whether the file's content matches any constraints.
     * If content-checking has been disabled, this method always returns {@code true}.
     * <p>
     * If the file has previously been checked for the same constraints on a prior search,
     *  this method will return the result of that prior search.
     * Otherwise, <b>this method reads the entire file's contents,
     *  which can significantly slow down the search.</b>
     * @param path the String absolute file path to check
     * @return a boolean, {@code true} if the file's contents match the current constraints
     * @throws Exception if something went wrong while reading the file
     */
    private boolean containsContent(String path) throws Exception {
        if (patternContent != null) {
            return contentMatcher.find(path);
        }
        return true;
    }

    /**
     * Determines whether a file's creation, modification, or access dates
     *  fall inside the given constraints.
     * If no date ranges have been specified, this method always returns {@code true}.
     * @param attributes the {@link BasicFileAttributes} to check
     * @return a boolean, {@code true} if the file falls within any and all date constraints
     */
    private boolean withinDate(BasicFileAttributes attributes) {
        return withinRange(attributes.creationTime().toMillis(), createdStart, createdEnd)
                && withinRange(attributes.lastModifiedTime().toMillis(), modifiedStart, modifiedEnd)
                && withinRange(attributes.lastAccessTime().toMillis(), accessedStart, accessedEnd);
    }

    /**
     * Determines whether a file falls within the given size constraints.
     * If size checking is disabled, this method always returns {@code true}.
     * @param attributes the {@link BasicFileAttributes} to check
     * @return a boolean, {@code true} if the file falls within any and all size constraints
     */
    private boolean withinSize(BasicFileAttributes attributes) {
        final long size = attributes.size();

        if (sizeSmaller >= 0L && sizeLarger >= 0L) {
            return withinRange(size, sizeSmaller, sizeLarger);
        }

        if (sizeSmallerWarning >= 0L && sizeSmallerError >= 0L) {
            return withinRange(size, -1L, sizeSmallerWarning);
        }

        if (sizeLargerWarning >= 0L && sizeLargerError >= 0L) {
            return withinRange(size, sizeLargerWarning, -1L);
        }

        return true;
    }

    /**
     * Determines whether a long value falls within the provided range,
     *  where both range constraints are optional:
     * <ul>
     *  <li><b>If both constraints are negative,</b> this method always returns {@code true}.</li>
     *  <li><b>If both constraints are >= 0,</b> then {@code value} must fall within
     *  both {@code start} and {@code end}, inclusive.</li>
     *  <li><b>If {@code start >= 0},</b> then {@code value}
     *  must be greater than or equal to {@code start}, with infinite upward range.</li>
     *  <li><b>If {@code end >= 0},</b> then {@code value}
     *  must be less than or equal to {@code end}, with infinite downward range.</li>
     * </ul>
     * @return a boolean, {@code true} if the given time is within the provided interval
     */
    private boolean withinRange(long value, long start, long end) {
        if (start < 0L && end < 0L) {
            return true;
        } else if (end >= 0L && start >= 0L) {
            return value >= start && value <= end;
        } else if (end < 0L) {
            return value >= start;
        } else {
            return value <= end;
        }
    }
}
