package edu.uwyo.crawl;

import java.io.File;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import edu.uwyo.Main;
import edu.uwyo.crawl.database.Database;
import edu.uwyo.crawl.ui.frameDeveloper.panels.PanelMain;
import edu.uwyo.crawl.ui.panels.PanelFiles;
import edu.uwyo.crawl.ui.panels.PanelOptions;
import edu.uwyo.crawl.ui.panels.PanelOptionsSize;
import edu.uwyo.crawl.ui.panels.PanelTime;
import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import net.javamon.dev.Console;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.commandParser.CommandLineParser;
import net.javamon.dev.utils.commandParser.CommandLineParser.MissingArgumentException;
import net.javamon.dev.utils.commandParser.CommandLineParser.MissingOptionException;
import net.javamon.dev.utils.commandParser.CommandLineParser.MultipleDefinitionException;
import net.javamon.dev.utils.commandParser.CommandLineParser.UnknownOptionException;
import net.javamon.dev.utils.commandParser.OptionMap;
import net.javamon.dev.utils.commandParser.option.BooleanOptionBuilder;
import net.javamon.dev.utils.commandParser.option.IntegerOptionBuilder;
import net.javamon.dev.utils.commandParser.option.Option;
import net.javamon.dev.utils.commandParser.option.OptionType.BadTypeException;
import net.javamon.dev.utils.commandParser.option.StringOptionBuilder;

/**
 * A representation of search parameters for {@link Query} that have been obtained
 *  from the command-line (no UI is involved in this class's usage).
 * Call {@link #parse(String[])} to update the parameters with a command-line string
 *  and {@link #run()} to run the resulting query.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class QueryParams {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final Pattern VALID_EXTENSION = Pattern.compile("\\.?[a-zA-Z0-9_]+");

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private File input;
    private File output;
    private int sizeMode;

    private Query query;
    private final Database database;
    private final OptionMap options;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link QueryParams}.
     * @param database the {@link Database} to use when parsing query parameters
     */
    public QueryParams(Database database) {
        this.database = database;

        Option<Boolean> help = new BooleanOptionBuilder(false, false, "h", "help")
                .description(Resources.__("cmdHelp"))
                .build();
        options = new OptionMap(help);

        options.put(new StringOptionBuilder(true, true, "f", "file")
                .description(Resources.__("cmdSearch")));

        options.put(new StringOptionBuilder(true, true, "o", "output")
                .description(Resources.__("cmdOutput")));

        options.put(new StringOptionBuilder(false, true, "i", "input")
                .description(Resources.__("cmdInput")));

        options.put(new IntegerOptionBuilder(false, true, "r", "recurse")
                .defaultArgument(PanelOptions.RECURSE_FILE)
                .description(Resources.__("cmdRecurse")));

        options.put(new IntegerOptionBuilder(false, true, "d", "duplicateMode")
                .defaultArgument(0)
                .description(Resources.__("cmdDuplicate")));

        options.put(new BooleanOptionBuilder(false, false, "D", "duplicateExact")
                .description(Resources.__("cmdDuplicateExact")));

        options.put(new StringOptionBuilder(false, true, "p", "patternName")
                .description(Resources.__("cmdPatternName")));

        options.put(new StringOptionBuilder(false, true, "P", "patternContent")
                .description(Resources.__("cmdPatternContent")));

        options.put(new StringOptionBuilder(false, true, "c", "created")
                .defaultArgument("-1:-1")
                .description(Resources.__("cmdCreated")));

        options.put(new StringOptionBuilder(false, true, "m", "modified")
                .defaultArgument("-1:-1")
                .description(Resources.__("cmdModified")));

        options.put(new StringOptionBuilder(false, true, "a", "accessed")
                .defaultArgument("-1:-1")
                .description(Resources.__("cmdAccessed")));

        options.put(new StringOptionBuilder(false, true, "s", "size")
                .defaultArgument(PanelOptionsSize.NONE + ",-1:-1")
                .description(Resources.__("cmdSize")));

        options.put(new StringOptionBuilder(false, true, "x", "extInclude")
                .description(Resources.__("cmdExt")));

        options.put(new StringOptionBuilder(false, true, "X", "extExclude")
                .description(Resources.__("cmdExtExclude")));

        options.put(new StringOptionBuilder(false, true, "F", "fileExclude")
                .description(Resources.__("cmdFileExclude")));

        options.put(new StringOptionBuilder(false, true, "N", "filePatternExclude")
                .description(Resources.__("cmdFilePatternExclude")));

        options.put(new IntegerOptionBuilder(false, true, "H", "headerSize")
                .defaultArgument(PanelMain.DEFAULT_HEADER_SIZE)
                .description(Resources.__("cmdHeaderSize")));

        options.put(new IntegerOptionBuilder(false, true, "b", "bufferLocal")
                .defaultArgument(PanelMain.DEFAULT_BUFFER_LOCAL)
                .description(Resources.__("cmdBufferLocal")));

        options.put(new IntegerOptionBuilder(false, true, "B", "bufferNet")
                .defaultArgument(PanelMain.DEFAULT_BUFFER_NET)
                .description(Resources.__("cmdBufferNet")));

        options.put(new IntegerOptionBuilder(false, true, "C", "chunkSize")
                .defaultArgument(PanelMain.DEFAULT_CHUNK_SIZE)
                .description(Resources.__("cmdChunkSize")));

        options.put(new BooleanOptionBuilder(false, false, "v", "verbose")
                .description(Resources.__("cmdVerbose")));
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Parses the given command-line, converting parsed arguments into the fields of this class.
     * @param args the array of String command-line tokens to parse
     * @return a boolean, {@code true} if parsing was successful
     */
    public boolean parse(String[] args) {
        try {
            if (CommandLineParser.parse(args, 0, options)) {
                showHelp();
                return false;
            }
        } catch (UnknownOptionException e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorUnknownOption", e.token));
            showHelp();
            return false;
        } catch (MissingArgumentException e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorMissingArgument", e.token));
            showHelp();
            return false;
        } catch (MissingOptionException e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorMissingOption", e.option.getTokens()));
            showHelp();
            return false;
        } catch (BadTypeException e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorUnknownOption", e.value,
                    e.typeConverter.getDisplayType()));
            showHelp();
            return false;
        } catch (MultipleDefinitionException e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorMultiple", e.token));
            showHelp();
            return false;
        }

        database.setCommandLine(options.getBoolean("verbose") != null);
        try {
            setInput();
            setFiles();
            setNesting();
            setDuplicates();
            setPatternName();
            setPatternContent();
            setCreated();
            setModified();
            setAccessed();
            setSize();
            setExt();
            setExcludeExt();
            setExcludeFiles();
            setExcludePattern();
            setHeaderSize();
            setBufferLocal();
            setBufferNet();
            setChunkSize();
        } catch (RuntimeException e) {
            query = null;
            return false;
        }
        return true;
    }

    /**
     * Runs the query that has been defined by the previously-read parameters.
     * This method does nothing unless {@link #parse(String[])} returned {@code true}.
     */
    public void run() {
        if (query != null) {
            FileDisplayTree results = query.run(database, true);

            writeSettings();
            if (database.saveSearchResults(output)) {
                Console.logInfo(Resources.__("cmdSave", output.getAbsolutePath()) + "\n");
            } else {
                Console.logInfo(Resources.__("cmdErrorSave", output.getAbsolutePath()) + "\n");
                String[] paths = results.getPaths();
                for (int i = 0; i < paths.length; i++) {
                    Console.logInfo("\n" + paths[i]);
                }
            }
        }
    }

    /** Shows the help message. */
    private void showHelp() {
        Console.logInfo("\n" + Resources.__("cmdUsage"));
        Console.logInfo(options.getHelp());
    }

    /** Sets the input file, if any. */
    private void setInput() {
        input = null;
        String i = options.getString("input");
        if (i != null) {
            input = new File(i);
            if (!input.isFile()) {
                Console.logInfo("\n\n" + Resources.__("cmdErrorInput", i));
                throw new RuntimeException();
            }
        }

        if (input != null && !database.loadSearchResults(input)) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorInputRead", input.getAbsolutePath()));
            throw new RuntimeException();
        }
        query = Main.query(); //reflect loaded query, or blank otherwise
    }

    /** Sets the input/output files. */
    private void setFiles() {
        query.toSearch = new File(options.getString("file"));
        output = new File(options.getString("output"));
    }

    /** Sets the nesting options. */
    private void setNesting() {
        query.nesting = Math.min(PanelOptions.RECURSE_INFINITE,
                Math.max(0, options.getInt("recurse")));
        if (input != null && input.isFile()) {
            query.nesting = PanelOptions.RECURSE_RESULTS;
        }

        if (query.nesting == PanelOptions.RECURSE_FILE && !query.toSearch.isFile()) {
            Console.logInfo("\n\n"
                    + Resources.__("cmdErrorSearchFile", query.toSearch.getAbsolutePath()));
            throw new RuntimeException();
        } else if (query.nesting != PanelOptions.RECURSE_FILE && !query.toSearch.isDirectory()) {
            Console.logInfo("\n\n"
                    + Resources.__("cmdErrorSearchDir", query.toSearch.getAbsolutePath()));
            throw new RuntimeException();
        }

        if (query.nesting == PanelOptions.RECURSE_INFINITE) {
            query.nesting = Integer.MAX_VALUE;
        } else if (query.nesting == PanelOptions.RECURSE_FILE) {
            query.fileMode = query.toSearch.getAbsolutePath();
            query.toSearch = query.toSearch.getParentFile();
            query.nesting = 1; //force search of enclosing folder to ensure proper tree display
        }
    }

    /** Sets the duplicate search settings. */
    private void setDuplicates() {
        query.duplicateMode = options.getInt("duplicateMode");
        if (query.duplicateMode < 0 || query.duplicateMode > 3) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorDuplicate", query.duplicateMode));
            throw new RuntimeException();
        }

        query.duplicateExact = (options.getBoolean("duplicateExact") != null
                && query.duplicateMode > 0);
    }

    /** Sets the name pattern. */
    private void setPatternName() {
        query.patternName = null;
        String p = options.getString("patternName");
        if (p != null) {
            try {
                query.patternName = Pattern.compile(p);
            } catch (PatternSyntaxException e) {
                Console.logInfo("\n\n" + Resources.__("cmdErrorPatternName", p));
                throw new RuntimeException();
            }
        }
    }

    /** Sets the content pattern. */
    private void setPatternContent() {
        query.patternContent = null;
        String P = options.getString("patternName");
        if (P != null) {
            try {
                query.patternName = Pattern.compile(P);
            } catch (PatternSyntaxException e) {
                Console.logInfo("\n\n" + Resources.__("cmdErrorPatternContent", P));
                throw new RuntimeException();
            }
        }
    }

    /** Sets the creation time filter. */
    private void setCreated() {
        String c = options.getString("created");
        try {
            String[] cTokens = c.split(":");
            query.createdStart = Long.parseLong(cTokens[0]);
            query.createdEnd = Long.parseLong(cTokens[1]);
            if ((query.createdStart >= query.createdEnd && query.createdStart != -1L)
                    || query.createdStart < -1L
                    || query.createdEnd < -1L) {
                throw new RuntimeException();
            }
        } catch (Exception e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorCreateFormat", c));
            throw new RuntimeException();
        }
    }

    /** Sets the modification time filter. */
    private void setModified() {
        String m = options.getString("modified");
        try {
            String[] mTokens = m.split(":");
            query.modifiedStart = Long.parseLong(mTokens[0]);
            query.modifiedEnd = Long.parseLong(mTokens[1]);
            if ((query.modifiedStart >= query.modifiedEnd && query.modifiedStart != -1L)
                    || query.modifiedStart < -1L
                    || query.modifiedEnd < -1L) {
                throw new RuntimeException();
            }
        } catch (Exception e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorModifyFormat", m));
            throw new RuntimeException();
        }
    }

    /** Sets the access time filter. */
    private void setAccessed() {
        String a = options.getString("accessed");
        try {
            String[] aTokens = a.split(":");
            query.accessedStart = Long.parseLong(aTokens[0]);
            query.accessedEnd = Long.parseLong(aTokens[1]);
            if ((query.accessedStart >= query.accessedEnd && query.accessedStart != -1L)
                    || query.accessedStart < -1L
                    || query.accessedEnd < -1L) {
                throw new RuntimeException();
            }
        } catch (Exception e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorAccessFormat", a));
            throw new RuntimeException();
        }
    }

    /** Sets the size filter. */
    private void setSize() {
        query.sizeSmallerWarning = -1L;
        query.sizeSmallerError = -1L;
        query.sizeLargerWarning = -1L;
        query.sizeLargerError = -1L;
        query.sizeSmaller = -1L;
        query.sizeLarger = -1L;

        String s = options.getString("size");
        try {
            String sSub = s.substring(0, s.indexOf(','));
            sizeMode = Integer.parseInt(sSub);
            String[] sTokens = s.substring(s.indexOf(',') + 1).split(":");

            switch (sizeMode) {
                case PanelOptionsSize.SMALLER:  query.sizeSmallerWarning = Long.parseLong(sTokens[0]);
                                                query.sizeSmallerError = Long.parseLong(sTokens[1]);
                                                if (query.sizeSmallerWarning < query.sizeSmallerError
                                                        || query.sizeSmallerWarning <= 0L
                                                        || query.sizeSmallerError <= 0L) {
                                                    throw new RuntimeException();
                                                }
                                                break;

                case PanelOptionsSize.LARGER:   query.sizeLargerWarning = Long.parseLong(sTokens[0]);
                                                query.sizeLargerError = Long.parseLong(sTokens[1]);
                                                if (query.sizeLargerWarning > query.sizeLargerError
                                                        || query.sizeLargerWarning <= 0L
                                                        || query.sizeLargerError <= 0L) {
                                                    throw new RuntimeException();
                                                }
                                                break;

                case PanelOptionsSize.BETWEEN:  query.sizeSmaller = Long.parseLong(sTokens[0]);
                                                query.sizeLarger = Long.parseLong(sTokens[1]);
                                                if (query.sizeSmaller > query.sizeLarger
                                                        || query.sizeSmaller <= 0L
                                                        || query.sizeLarger <= 0L) {
                                                    throw new RuntimeException();
                                                }
                                                break;

                case PanelOptionsSize.NONE:     break;

                default:                        throw new RuntimeException();
            }

        } catch (Exception e) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorSizeFormat", s));
            throw new RuntimeException();
        }
    }

    /** Sets the included extensions. */
    private void setExt() {
        query.extensions = null;
        String x = options.getString("extInclude");
        if (x != null) {
            query.extensions = x.split(",");
            for (String extension : query.extensions) {
                if (!VALID_EXTENSION.matcher(extension).matches()) {
                    Console.logInfo("\n\n" + Resources.__("cmdErrorExt", extension));
                    throw new RuntimeException();
                }
            }
        }
    }

    /** Sets the excluded extensions. */
    private void setExcludeExt() {
        query.excludedExtensions = null;
        String X = options.getString("extExclude");
        if (X != null) {
            query.excludedExtensions = X.split(",");
            for (String extension : query.excludedExtensions) {
                if (!VALID_EXTENSION.matcher(extension).matches()) {
                    Console.logInfo("\n\n" + Resources.__("cmdErrorExcludeExt", extension));
                    throw new RuntimeException();
                }
            }
        }
    }

    /** Sets the excluded files. */
    private void setExcludeFiles() {
        query.excludedFiles = null;
        String F = options.getString("fileExclude");
        if (F != null) {
            query.excludedFiles = F.split(",");
        }
    }

    /** Sets the excluded pattern. */
    private void setExcludePattern() {
        query.excludedPattern = null;
        String N = options.getString("filePatternExclude");
        if (N != null) {
            try {
                query.excludedPattern = Pattern.compile(N);
            } catch (PatternSyntaxException e) {
                Console.logInfo("\n\n" + Resources.__("cmdErrorPatternExclude", N));
                throw new RuntimeException();
            }
        }
    }

    /** Sets the header size. */
    private void setHeaderSize() {
        query.headerSize = options.getInt("headerSize");
        if (query.headerSize < 1) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorHeaderSize", query.headerSize));
            throw new RuntimeException();
        }
    }

    /** Sets the local buffer size. */
    private void setBufferLocal() {
        query.bufferLocal = options.getInt("bufferLocal");
        if (query.bufferLocal < 1) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorBufferLocal", query.bufferLocal));
            throw new RuntimeException();
        }
    }

    /** Sets the network buffer size. */
    private void setBufferNet() {
        query.bufferNet = options.getInt("bufferNet");
        if (query.bufferNet < 1) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorBufferNet", query.bufferNet));
            throw new RuntimeException();
        }
    }

    /** Sets the chunk size. */
    private void setChunkSize() {
        query.chunkSize = options.getInt("chunkSize");
        if (query.chunkSize < 1) {
            Console.logInfo("\n\n" + Resources.__("cmdErrorChunkSize", query.chunkSize));
            throw new RuntimeException();
        }
    }

    /** Writes this query's search settings to the current results file. */
    private void writeSettings() {
        Main.writeSetting("lastFolder", query.toSearch.getAbsolutePath());
        writePatternSetting("Name", query.patternName);
        writePatternSetting("Content", query.patternContent);
        writeNestingSetting();
        Main.writeSetting("lastDuplicate", query.duplicateMode + "");
        Main.writeSetting("lastDuplicateExact", query.duplicateExact ? "True" : "False");
        writeExtensionSettings();
        writeSizeSettings();
        writeDateSetting("Created", query.createdStart, query.createdEnd);
        writeDateSetting("Modified", query.modifiedStart, query.modifiedEnd);
        writeDateSetting("Accessed", query.accessedStart, query.accessedEnd);

        for (int i = 0; i < PanelFiles.getNumFlags(); i++) {
            Main.writeSetting("lastFlagFilter" + i, "True");
        }
    }

    /** Writes a single pattern setting. */
    private void writePatternSetting(String patternName, Pattern pattern) {
        if (pattern == null) {
            Main.writeSetting("lastPattern" + patternName, "");
        } else {
            Main.writeSetting("lastPattern" + patternName, pattern.toString());
        }
    }

    /** Writes the nesting setting. */
    private void writeNestingSetting() {
        if (query.nesting == 1 && query.fileMode != null) {
            Main.writeSetting("lastRecurse", PanelOptions.RECURSE_FILE + "");
        } else if (query.nesting == Integer.MAX_VALUE) {
            Main.writeSetting("lastRecurse", PanelOptions.RECURSE_INFINITE + "");
        } else {
            Main.writeSetting("lastRecurse", query.nesting + "");
        }
    }

    /** Writes the included extension settings. */
    private void writeExtensionSettings() {
        if (query.extensions == null) {
            Main.writeSetting("lastFileTypes", "");
            return;
        }

        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < query.extensions.length; i++) {
            if (query.extensions[i].startsWith(".")) {
                builder.append(query.extensions[i]);
            } else {
                builder.append("." + query.extensions[i]);
            }

            if (i < query.extensions.length - 1) {
                builder.append(" ");
            }
        }
        Main.writeSetting("lastFileTypes", builder.toString());
    }

    /** Writes the size settings. */
    private void writeSizeSettings() {
        Main.writeSetting("lastSizeMode", sizeMode + "");
        switch (sizeMode) {
            case PanelOptionsSize.SMALLER:  writeSize(query.sizeSmallerWarning,
                                                    query.sizeSmallerError);
                                            break;

            case PanelOptionsSize.LARGER:   writeSize(query.sizeLargerWarning,
                                                    query.sizeLargerError);
                                            break;

            case PanelOptionsSize.BETWEEN:  writeSize(query.sizeSmaller, query.sizeLarger);
                                            break;

            default:                        writeSize(0L, 0L);
                                            break;
        }
        Main.writeSetting("lastSizeWarningUnits", "0");
        Main.writeSetting("lastSizeErrorUnits", "0");
    }

    /** Writes a single size setting. */
    private void writeSize(long lower, long upper) {
        Main.writeSetting("lastSizeWarning", String.valueOf(lower / 1024));
        Main.writeSetting("lastSizeError", String.valueOf(upper / 1024));
    }

    /** Writes a single date setting. */
    private void writeDateSetting(String dateLabel, long start, long end) {
        Main.writeSetting("lastTime" + dateLabel + "Enabled",
                (start > -1L || end > -1L) ? "True" : "False");
        Main.writeSetting("lastTime" + dateLabel + "Comparison", getDateMode(start, end) + "");
        Main.writeSetting("lastTime" + dateLabel + "DateMin", start + "");
        Main.writeSetting("lastTime" + dateLabel + "DateMax", end + "");
    }

    /** @return the integer date mode associated with the given start/end times */
    private int getDateMode(long start, long end) {
        if (start == -1L && end == -1L) {
            return 0;
        } else if (start == -1L) {
            return PanelTime.OLDER;
        } else if (end == -1L) {
            return PanelTime.NEWER;
        }
        return PanelTime.BETWEEN;
    }
}
