package edu.uwyo.crawl.contentMatcher;

import java.util.ArrayList;

/**
 * An abstract token representing a single syntactical element of a regular expression,
 *  such as a character to be matched, a capture group, or a character class.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public abstract class AbstractExpression {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    protected static final int EXACT = 0;
    protected static final int ZERO_OR_MORE = 1;
    protected static final int ONE_OR_MORE = 2;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    protected int currOrGroup;
    protected ArrayList<Integer> length;
    protected ArrayList<ArrayList<AbstractExpression>> expressions;

    private int mode;
    private int repeatMin;
    private int repeatMax;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link AbstractExpression}. */
    protected AbstractExpression() {
        expressions = new ArrayList<ArrayList<AbstractExpression>>();
        length = new ArrayList<Integer>();

        expressions.add(new ArrayList<AbstractExpression>());
        length.add(0);
        currOrGroup = expressions.size() - 1; //nextOrGroup might be overridden

        mode = EXACT;
        repeatMin = -1;
        repeatMax = -1;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                ABSTRACT METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Converts this expression into its natural string form.
     * This method must guarantee that the result of {@link StringBuilder#toString()}
     *  will produce a <i>legal, one-to-one regular expression</i>; that is, the output must be
     *  a regular expression <u>identical</u> to the original input pattern.
     * @param builder the {@link StringBuilder} to which this expression should be appended
     * @see #appendNumMatches(StringBuilder)
     */
    public abstract void toString(StringBuilder builder);

    /**
     * Adds an expression to this expression.
     * Used e.g. for capture groups that contain one or more sub-expressions.
     * <p>
     * <i>If this expression supports the union operator, this method must use {@link #currOrGroup}
     *  when adding elements to ensure the correct union behavior is achieved.</i>
     * @param expression the {@link AbstractExpression} to add
     * @param length the integer length of {@code expression};
     *  typically {@link AbstractExpression#length()}, but may be overridden here as needed
     */
    protected abstract void add(AbstractExpression expression, int length);

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Obtains the <i>minimal length</i> of this expression, equivalent to the length
     *  of the shortest string that could be matched.
     * Useful in determining how far to jump ahead when using e.g. a Boyer-Moore algorithm
     *  to more efficiently find patterns.
     * Considers all sub-expressions, if any.
     * <p>
     * Expressions with a Kleene closure will always have length zero,
     *  while a union will always return the shortest possible branch.
     * Repeats will return the expanded character count (e.g. <code>(ab){3}</code> will return 6).
     * All other expression types should return the minimum number of characters
     *  required for a match.
     * @return the integer minimal length of this expression
     */
    public int length() {
        int minLength = Integer.MAX_VALUE;
        for (int i = 0; i < length.size(); i++) {
            if (repeatMin >= 0) {
                minLength = Math.min(minLength, length.get(i) * repeatMin);
                continue;
            }

            switch (mode) {
                case ONE_OR_MORE:   minLength = Math.min(minLength, 1); break;
                case ZERO_OR_MORE:  minLength = Math.min(minLength, 0); break;
                default:            minLength = Math.min(minLength, length.get(i)); break;
            }
        }
        return minLength;
    }

    /**
     * Indicates that this expression must be matched zero or more times.
     * This is equivalent to appending a {@code *} character to the expression.
     * <p>
     * <i>This method overwrites the previous quantifier, if any.</i>
     */
    protected void zeroOrMore() {
        mode = ZERO_OR_MORE;
    }

    /**
     * Indicates that this expression must be matched one or more times.
     * This is equivalent to appending a {@code +} character to the expression.
     * <p>
     * <i>This method overwrites the previous quantifier, if any.</i>
     */
    protected void oneOrMore() {
        mode = ONE_OR_MORE;
    }

    /**
     * Indicates that this expression must be repeated a given number of times.
     * This is equivalent to the <code>{x,y}</code> syntax.
     * @param min the integer minimum number of repeats; if -1, all repeating will be disabled
     * @param max the integer maximum number of repeats; if -1, there will be no maximum
     */
    protected void repeat(int min, int max) {
        repeatMin = min;
        repeatMax = max;
    }

    /**
     * Progresses this expression to its next or-group.
     * Used when the union operation {@code |} is encountered.
     */
    protected void nextOrGroup() {
        expressions.add(new ArrayList<AbstractExpression>());
        length.add(0);
        currOrGroup = expressions.size() - 1;
    }

    /**
     * A utility method for printing via {@link #toString(StringBuilder)}.
     * Automatically appends any match quantifiers <code>*, +, {3}, ...</code>
     *  to the printed expression.
     * Only expressions that support quantifiers should call this method.
     * @param builder the {@link StringBuilder} to which this expression should be appended
     */
    protected void appendNumMatches(StringBuilder builder) {
        if (repeatMin >= 0) {
            builder.append("{");
            builder.append(repeatMin);
            if (repeatMax == repeatMin) {
                builder.append(",");
            } else if (repeatMax > repeatMin) {
                builder.append("," + repeatMax);
            }
            builder.append("}");
        }

        if (isZeroOrMore()) {
            builder.append("*");
        } else if (isOneOrMore()) {
            builder.append("+");
        }
    }

    /** @return a boolean, {@code true} if this expression must be matched zero or more times */
    private boolean isZeroOrMore() {
        return mode == ZERO_OR_MORE;
    }

    /** @return a boolean, {@code true} if this expression must be matched one or more times */
    private boolean isOneOrMore() {
        return mode == ONE_OR_MORE;
    }
}
