package edu.uwyo.crawl.contentMatcher;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.regex.Pattern;

import net.javamon.dev.Console;
import net.javamon.dev.Ext;
import net.javamon.dev.ZipUtils;

/**
 * A class for reading a file and determining whether its contents match a desired pattern.
 * The pattern (a regular expression) is set via {@link #setPattern(Pattern)},
 *  and will then be used on all subsequent invocations of {@link #find(String)}.
 * Performance can be tuned using {@link #tuneParameters(int, int)}.
 * <p>
 * Non-readable characters (e.g. null characters and other lower-order ASCII bytes)
 *  will be automatically stripped out of the read files to prevent inclusion in the search.
 * Consequently, searches for those characters (particularly the string terminator {@code \0})
 *  will never return any results.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class ContentMatcher implements TunableContentSearch {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private boolean enabled;
    private int maxChunkChars;
    private char[] buffer;

    private final PatternParser parser;
    private final StringBuilder fileChars;
    private final ZipMatcher zipMatcher;
    private final ZipMatcher noXMLMatcher;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link ContentMatcher}. */
    public ContentMatcher() {
        parser = new PatternParser();
        fileChars = new StringBuilder();
        zipMatcher = new ZipMatcher(fileChars, parser, false, Ext.BINARY);
        noXMLMatcher = new ZipMatcher(fileChars, parser, true, Ext.BINARY,
                "docProps/core.xml", "docProps/app.xml");

        maxChunkChars = 256000000; //512 MB (2 bytes per char)
        buffer = new char[8192];
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void tuneParameters(int bufferSize, int maxChunkBytes) {
        if (bufferSize / 2 != buffer.length) {
            Console.logInfo("buffer size set to " + bufferSize);
            buffer = new char[bufferSize / 2];
            zipMatcher.tuneParameters(bufferSize, maxChunkBytes);
            noXMLMatcher.tuneParameters(bufferSize, maxChunkBytes);
        }
        maxChunkChars = maxChunkBytes / 2;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the pattern (a regular expression) for which to search.
     * This method must be called exactly once before a search (on one or multiple files) begins.
     * The pattern set here will be used on all future calls to {@link #find(String)}
     *  until a new pattern is set.
     * <p>
     * Although a {@link Pattern} must be provided to this method,
     *  the Java specification for regular expressions <u>is not</u> perfectly supported;
     *  some meta-characters e.g. {@code \w} may behave differently than expected.
     * See {@link PatternParser} for an enumeration of supported syntax.
     * @param pattern the {@link Pattern} for which to search;
     *  if {@code null}, searching via {@link #find(String)} will be disabled
     */
    public void setPattern(Pattern pattern) {
        enabled = (pattern != null);
        if (pattern != null) {
            parser.setPattern(pattern.toString());
        }
    }

    /**
     * Reads the file at {@code path} and attempts to find the pattern that was specified
     *  by the last call to {@link #setPattern(Pattern)}.
     * The file is assumed to exist; but if anything goes wrong while reading the file,
     *  an exception will be thrown.
     * Special support exists for reading files in the following formats: {@link Ext#XLSX},
     *  {@link Ext#DOCX}, {@link Ext#PPTX}, and {@link Ext#ZIP}.
     * <p>
     * Non-readable characters, including the null terminator {@code \0},
     *  will be stripped out of every read file.
     * Consequently, searches for lower-order ASCII bytes will never return any results.
     * <p>
     * If a file is too large to be read in memory, as determined
     *  by {@link #tuneParameters(int, int)}, it will be split into multiple segments.
     * The search will then run separately on each segment.
     * In the case that a match would occur across the split, that match will <u>not</u> be found.
     * However, this scenario is very unlikely given a large split size.
     * @param path the String path of the file to search
     * @return a boolean, {@code true} if the pattern was found and {@code false} otherwise
     * @throws Exception if something went wrong while reading the file
     */
    public boolean find(String path) throws Exception {
        if (!enabled) { return false; }

        fileChars.setLength(0);
        String lower = path.toLowerCase();
        if (lower.endsWith(Ext.XLSX) || lower.endsWith(Ext.DOCX) || lower.endsWith(Ext.PPTX)) {
            return ZipUtils.invoke(new File(path), noXMLMatcher);
        } else if (lower.endsWith(Ext.ZIP)) {
            return ZipUtils.invoke(new File(path), zipMatcher);
        }

        BufferedReader reader = new BufferedReader(
                new InputStreamReader(new FileInputStream(path), "UTF-8"), buffer.length);
        int charsRead = 0;

        while (charsRead != -1) {
            charsRead = reader.read(buffer);
            if (charsRead != -1) {
                for (int i = 0; i < charsRead; i++) {
                    if ((buffer[i] > 31 && buffer[i] != 127)
                            || (buffer[i] > 8 && buffer[i] < 14)) {
                        fileChars.append(buffer[i]);
                    }
                }

                if (fileChars.length() >= maxChunkChars) {
                    if (parser.find(fileChars)) {
                        reader.close();
                        return true;
                    }
                    fileChars.setLength(0); //can't read entire file at once; break into segments
                }
            }
        }
        reader.close();
        return parser.find(fileChars);
    }
}
