package edu.uwyo.crawl.contentMatcher;

import java.util.Stack;

import dk.brics.automaton.AutomatonMatcher;
import dk.brics.automaton.RegExp;
import dk.brics.automaton.RunAutomaton;
import net.javamon.dev.Console;

/**
 * A fast parser for regular expressions.
 * Setting a new parser via {@link #setPattern(String)} involves the construction of
 *  a complete finite deterministic automaton (DFA) for that pattern and can be quite expensive.
 * However, all subsequent searches on the DFA will be <i>very</i> fast;
 *  often an order of magnitude faster than Java's built-in parser.
 * Uses the {@link dk.brics.automaton} package.
 * <p>
 * The following syntax is supported:
 * <ul>
 *  <li>Repeat sequences ?, *, +, {x}, {x,}, and {x,y}</li>
 *  <li>Whitespace meta-characters \s and \S</li>
 *  <li>Digit meta-characters \d and \D</li>
 *  <li>Word meta-characters \w and \W (matches [a-zA-Z0-9]; <i>not</i> Unicode-aware)</li>
 *  <li>Horizontal whitespace meta-characters \h and \H</li>
 *  <li>Vertical whitespace meta-characters \v and \V</li>
 *  <li>Unicode characters are directly supported (e.g. ) or can be embedded
 *  via hex meta-characters \xYY and \x{YYYY} (e.g. \x{00E9})</li>
 *  <li>Character classes [] and [^] (negation)</li>
 *  <li>Capture groups ()</li>
 *  <li>Character wildcard .</li>
 *  <li>Union operator |</li>
 *  <li>Escapes via \ (e.g. \( matches a left parenthesis)</li>
 *  <li>^ to denote the start of a string; <i>only supported if provided at the start of a pattern
 *  and if not part of a union (will be stripped out of the pattern prior to processing)</i></li>
 *  <li>$ to denote the end of a string; <i>only supported if provided at the end of a pattern
 *  and if not part of a union (will be stripped out of the pattern prior to processing)</i></li>
 * </ul>
 * <p>
 * Because ^ and $ are only processed outside of the regular expression itself,
 *  these characters <i>do not</i> need to be escaped elsewhere within the regular expression.
 * However, escaping those characters will not affect the compiled pattern.
 * <p>
 * <i>This class does not discriminate between single-line and multi-line matches.
 * It is up to the caller to determine what is passed into {@link #find(CharSequence)}:
 *  single lines, a multi-line string, the contents of an entire file, etc.</i>
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class PatternParser {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private RunAutomaton toFind;
    private boolean atStart;
    private boolean atEnd;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determines whether a full file/string is required to call {@link #find(CharSequence)}.
     * @return a boolean, {@code true} if the pattern requires a full string/file
     *  to return valid results
     */
    public boolean requireFull() {
        return atEnd;
    }

    /**
     * Searches the character sequence {@code s} for the pattern
     *  set by {@link #setPattern(String)}.
     * If no pattern was previously set, this method always returns {@code false}.
     * @param s the {@link CharSequence} to search
     * @return a boolean, {@code true} if the pattern was found and {@code false} otherwise
     * @see #find(CharSequence, int, int)
     */
    public boolean find(CharSequence s) {
        return find(s, 0, s.length());
    }

    /**
     * Searches the character sequence {@code s} for the pattern
     *  set by {@link #setPattern(String)}.
     * If no pattern was previously set, this method always returns {@code false}.
     * @param s the {@link CharSequence} to search
     * @param startOffset the integer index in {@code s} at which to start searching
     * @param endOffset the integer index in {@code s} at which to end searching
     * @return a boolean, {@code true} if the pattern was found and {@code false} otherwise
     */
    public boolean find(CharSequence s, int startOffset, int endOffset) {
        if (toFind == null) { return false; }
        AutomatonMatcher matcher = toFind.newMatcher(s, startOffset, endOffset);

        boolean found;
        while (true) {
            found = matcher.find();

            if (found) {
                if ((!atStart || matcher.start() == 0)
                        && (!atEnd || matcher.end() == s.length())) {
                    return true;
                }
            } else {
                return false;
            }
        }
    }

    /**
     * Sets the regular expression for which to search.
     * This method automatically expands meta-characters and creates a DFA for running the search.
     * @param pattern the String regular expression to use when searching
     */
    public void setPattern(String pattern) {
        ExpressionTokenizer tokenizer = parsePattern(pattern);
        String toFindPattern = tokenizer.toString();

        atStart = toFindPattern.startsWith("^");
        if (atStart) {
            toFindPattern = toFindPattern.substring(1);
        }

        atEnd = toFindPattern.endsWith("$") && !toFindPattern.endsWith("\\$");
        if (atEnd) {
            toFindPattern = toFindPattern.substring(0, toFindPattern.length() - 1);
        }

        Console.logInfo("expanded " + pattern + " to " + toFindPattern);
        toFind = new RunAutomaton(new RegExp(toFindPattern).toAutomaton());
    }

    /**
     * Parses a regular expression into tokens, which can then be recompiled
     *  into a form usable by {@link RegExp}.
     * @param pattern the String pattern to parse
     * @return an {@link ExpressionTokenizer} representing {@code pattern};
     *  call {@link ExpressionTokenizer#toString()} to get a usable string representation
     *  of the pattern
     */
    private ExpressionTokenizer parsePattern(String pattern) {
        pattern = pattern.replace("\\n", "\n").replace("\\t", "\t").replace("\\r", "\r")
                .replace("\\0", "\0").replace("\\f", "\f").replace("\\b", "\b");
        ExpressionTokenizer tokenizer = new ExpressionTokenizer();

        int i = 0;
        boolean forceContinue;
        Stack<GroupExpression> groupExpressions = new Stack<GroupExpression>();

        while (i < pattern.length()) {
            forceContinue = false;

            /* UNION */
            if (!forceContinue && pattern.charAt(i) == '|') {
                if (groupExpressions.size() > 0) {
                    groupExpressions.peek().nextOrGroup();
                } else {
                    tokenizer.nextOrGroup();
                }
                forceContinue = true;
            }

            /* ESCAPE: META-CHARACTER OR REGULAR CHARACTER */
            if (!forceContinue && pattern.charAt(i) == '\\') {
                ConstantExpression constantExpression = new ConstantExpression();

                int jump = parseMetaCharacter(i + 1, pattern, constantExpression);
                if (jump == 0) {
                    constantExpression.add(new CharExpression(
                            pattern.charAt(i) + "" + pattern.charAt(i + 1)), 1);
                    jump = 1;
                }
                i += jump;

                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '{') {
                        i += parseRepeat(i + 1, pattern, constantExpression);
                    }
                }

                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '*') {
                        i++;
                        constantExpression.zeroOrMore();
                    } else if (pattern.charAt(i + 1) == '+') {
                        i++;
                        constantExpression.oneOrMore();
                    }
                }

                if (groupExpressions.size() == 0) {
                    tokenizer.add(constantExpression, constantExpression.length());
                } else {
                    groupExpressions.peek().add(constantExpression, constantExpression.length());
                }
                forceContinue = true;
            }

            /* CHARACTER CLASS */
            if (!forceContinue && pattern.charAt(i) == '[') {
                BracketExpression bracketExpression = new BracketExpression();
                i++;
                while (i < pattern.length()) {
                    if (pattern.charAt(i) == '\\') {
                        if (pattern.charAt(i + 1) == 'x') {
                            i += parseHex(i + 2, pattern, bracketExpression);
                        } else {
                            bracketExpression.add(new CharExpression(
                                    pattern.charAt(i) + "" + pattern.charAt(i + 1)), 1);
                            i++;
                        }
                    } else if (pattern.charAt(i) == ']') {
                        if (i + 1 < pattern.length()) {
                            if (pattern.charAt(i + 1) == '*') {
                                i++;
                                bracketExpression.zeroOrMore();
                            } else if (pattern.charAt(i + 1) == '+') {
                                i++;
                                bracketExpression.oneOrMore();
                            }
                        }
                        if (groupExpressions.size() == 0) {
                            tokenizer.add(bracketExpression, bracketExpression.length());
                        } else {
                            groupExpressions.peek().add(bracketExpression, bracketExpression.length());
                        }
                        forceContinue = true;
                        break;
                    } else {
                        bracketExpression.add(new CharExpression(pattern.charAt(i)), 1);
                    }
                    i++;
                }
                //error
            }

            /* START OF CAPTURE GROUP */
            if (!forceContinue && pattern.charAt(i) == '(') {
                groupExpressions.push(new GroupExpression());
                forceContinue = true;
            }

            /* END OF CAPTURE GROUP */
            if (!forceContinue && groupExpressions.size() > 0 && pattern.charAt(i) == ')') {
                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '{') {
                        i += parseRepeat(i + 1, pattern, groupExpressions.peek());
                    }
                }

                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '*') {
                        i++;
                        groupExpressions.peek().zeroOrMore();
                    } else if (pattern.charAt(i + 1) == '+') {
                        i++;
                        groupExpressions.peek().oneOrMore();
                    }
                }
                tokenizer.add(groupExpressions.peek(), groupExpressions.peek().length());
                groupExpressions.pop();
                forceContinue = true;
            }

            /* START ANCHOR */
            if (!forceContinue && pattern.charAt(i) == '^') {
                StartExpression startExpression = new StartExpression();
                if (groupExpressions.size() == 0) {
                    tokenizer.add(startExpression, startExpression.length());
                } else {
                    groupExpressions.peek().add(startExpression, startExpression.length());
                }
                forceContinue = true;
            }

            /* END ANCHOR */
            if (!forceContinue && pattern.charAt(i) == '$') {
                EndExpression endExpression = new EndExpression();
                if (groupExpressions.size() == 0) {
                    tokenizer.add(endExpression, endExpression.length());
                } else {
                    groupExpressions.peek().add(endExpression, endExpression.length());
                }
                forceContinue = true;
            }

            /* REGULAR CHARACTERS */
            if (!forceContinue) {
                ConstantExpression constantExpression = new ConstantExpression();
                constantExpression.add(new CharExpression(pattern.charAt(i)), 1);
                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '{') {
                        i += parseRepeat(i + 1, pattern, constantExpression);
                    }
                }

                if (i + 1 < pattern.length()) {
                    if (pattern.charAt(i + 1) == '*') {
                        i++;
                        constantExpression.zeroOrMore();
                    } else if (pattern.charAt(i + 1) == '+') {
                        i++;
                        constantExpression.oneOrMore();
                    }
                }
                if (groupExpressions.size() == 0) {
                    tokenizer.add(constantExpression, constantExpression.length());
                } else {
                    groupExpressions.peek().add(constantExpression, constantExpression.length());
                }
                forceContinue = true;
            }

            i++;
        }

        return tokenizer;
    }

    /**
     * Parses a meta-character (including hex characters).
     * If the character at {@code i} is not a meta-character, this method returns 0.
     * @param i the integer index in the current {@code pattern} where the meta-character was found
     * @param pattern the String pattern to examine
     * @param expression the {@link AbstractExpression} to which any characters, character classes,
     *  etc. should be added
     * @return the integer number of characters that were processed by this method;
     *  when added to {@code i}, this represents the next position in the pattern to be parsed
     */
    private int parseMetaCharacter(int i, String pattern, AbstractExpression expression) {
        BracketExpression bracketExpression = new BracketExpression();
        switch (pattern.charAt(i)) {
            case 's':   bracketExpression.add(new CharExpression(" \\t\\n\\r"), 1); break;
            case 'S':   bracketExpression.add(new CharExpression("^ \\t\\n\\r"), 1); break;
            case 'd':   bracketExpression.add(new CharExpression("0-9"), 1); break;
            case 'D':   bracketExpression.add(new CharExpression("^0-9"), 1); break;
            case 'w':   bracketExpression.add(new CharExpression("a-zA-Z0-9_"), 1); break;
            case 'W':   bracketExpression.add(new CharExpression("^a-zA-Z0-9_"), 1); break;
            case 'x':   return parseHex(i, pattern, expression);
            case 'h':   bracketExpression.add(new CharExpression(" \\t"), 1); break;
            case 'H':   bracketExpression.add(new CharExpression("^ \\t"), 1); break;
            case 'v':   bracketExpression.add(new CharExpression("\\n\\r"), 1); break;
            case 'V':   bracketExpression.add(new CharExpression("^\\n\\r"), 1); break;
            default:    return 0;
        }

        expression.add(bracketExpression, 1);
        return 1;
    }

    /**
     * Parses a hex meta-character.
     * If the character at {@code i} is not a hex meta-character, this method returns 0.
     * @param i the integer index in the current {@code pattern} where the meta-character was found
     * @param pattern the String pattern to examine
     * @param expression the {@link AbstractExpression} to which any characters, character classes,
     *  etc. should be added
     * @return the integer number of characters that were processed by this method;
     *  when added to {@code i}, this represents the next position in the pattern to be parsed
     */
    private int parseHex(int i, String pattern, AbstractExpression expression) {
        if (pattern.charAt(i) != 'x') { return 0; }
        String hexCode = null;
        int jump = 0;

        if (pattern.charAt(i + 1) == '{' && i + 6 < pattern.length()
                && pattern.charAt(i + 6) == '}') {
            hexCode = pattern.substring(i + 2, i + 6);
            jump = 7;
        } else if (i + 2 < pattern.length()) {
            hexCode = "00" + pattern.substring(i + 1, i + 3);
            jump = 3;
        }

        if (hexCode != null) {
            try {
                char unicode = (char) Integer.parseInt(hexCode, 16);
                expression.add(new CharExpression(unicode), 1);
            } catch (NumberFormatException e) {
                jump = 0;
            }
        }
        return jump;
    }

    /**
     * Parses a repeat meta-sequence.
     * @param i the integer index in the current {@code pattern} where the meta-sequence was found
     * @param pattern the String pattern to examine
     * @param expression the {@link AbstractExpression} to which any characters, character classes,
     *  etc. should be added
     * @return the integer number of characters that were processed by this method;
     *  when added to {@code i}, this represents the next position in the pattern to be parsed
     */
    private int parseRepeat(int i, String pattern, AbstractExpression expression) {
        int start = i + 1;
        int end = i + 1;

        int seek = i;
        while (seek < pattern.length()) {
            if (pattern.charAt(seek) == '}') {
                end = seek;
                break;
            }
            seek++;
        }

        String[] startEnd = pattern.substring(start, end).split(",");
        if (startEnd.length > 0 && startEnd.length <= 2) {
            try {
                int startVal = Integer.parseInt(startEnd[0]);
                int endVal = -1;
                if (startEnd.length == 2) {
                    endVal = Integer.parseInt(startEnd[1]);
                }

                expression.repeat(startVal, endVal);
                return pattern.substring(start, end).length() + 2;
            } catch (NumberFormatException e) {}
        }

        return 0;
    }
}
