package edu.uwyo.crawl.contentMatcher;

import java.io.IOException;
import java.util.regex.Pattern;
import java.util.zip.ZipInputStream;

import edu.uwyo.Main;
import net.javamon.dev.Ext;
import net.javamon.dev.ZipUtils.ZipInvokable;

/**
 * A content matcher for zip files, including Microsoft Word and Excel documents.
 * This class unzips files in memory (<i>without</i> creating any files on disk),
 *  then reads each zipped entry one-at-a-time.
 * If any matches are found, the rest of the file will <u>not</u> be read,
 *  thereby minimizing disk I/O.
 * <p>
 * Typical usage involves passing this object
 *  into {@link net.javamon.dev.ZipUtils#invoke(java.io.File, ZipInvokable)}.
 * The current search parameters are always determined by the objects passed in
 *  during construction, but are otherwise mutable.
 * <p>
 * XML tags {@literal <}{@code ...}{@literal >} can optionally be stripped from the zipped file,
 *  preventing metadata from being searched.
 * Additionally, specific files or file types within the zip archive can be excluded.
 * <p>
 * Some archives may be too large to be easily represented in memory.
 * Similar to {@link ContentMatcher#find(String)}, files that are considered to be too large
 *  will be split, and a separate search will run on each chunk.
 * In addition to false negatives occurring across the split point,
 *  <b>splitting files may create false positives</b> if XML tags need to be stripped,
 *  as a split may occur in the middle of a tag.
 * Splits will preferentially occur between entries in the zip file, but cannot be guaranteed.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class ZipMatcher implements ZipInvokable, TunableContentSearch {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CONSTANTS                    ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final Pattern EXCLUDE_XML = Pattern.compile("<[^>]*>");

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private int maxChunkChars;
    private byte[] buffer;

    private final StringBuilder toParse;
    private final PatternParser parser;
    private final boolean stripXML;
    private final StringBuilder builder;
    private final String[] exclude;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link ZipMatcher}.
     * @param toParse a {@link StringBuilder} that will contain the string
     *  to be searched; each read file should be appended to this builder
     * @param parser the {@link PatternParser} indicating the current search parameters
     * @param stripXML a boolean, {@code true} if XML tags should be removed
     *  from the searched file; <i>this is required for Word and Excel documents
     *  to prevent false positives</i>
     * @param exclude a variable number of String file names/extensions indicating zip entries
     *  to exclude from the search, <i>in addition to those already in the global exclusion list</i>
     */
    protected ZipMatcher(StringBuilder toParse, PatternParser parser, boolean stripXML,
            String... exclude) {
        this.toParse = toParse;
        this.parser = parser;
        this.stripXML = stripXML;
        this.exclude = exclude;

        maxChunkChars = 256000000; //512 MB (2 bytes per char)
        buffer = new byte[8192];
        builder = new StringBuilder();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void start() {
        builder.setLength(0);
    }

    @Override
    public boolean invoke(ZipInputStream zipIn, String name) throws IOException {
        if (endsWith(name)) { return false; }
        clearBuffer();
        int count = zipIn.read(buffer);

        while (count != -1) {
            if (count > 0) {
                String chars = new String(buffer, 0, count, "UTF-8");
                for (int i = 0; i < chars.length(); i++) {
                    if ((chars.charAt(i) > 31 && chars.charAt(i) != 127)
                            || (chars.charAt(i) > 8 && chars.charAt(i) < 14)) {
                        builder.append(chars.charAt(i));
                    }
                }

                if (builder.length() >= maxChunkChars) {
                    append();
                    if (parser.find(toParse)) {
                        builder.setLength(0);
                        toParse.setLength(0);
                        return true;
                    }
                    builder.setLength(0);
                    toParse.setLength(0); //can't read entire entry at once; break into segments
                }
            }
            count = zipIn.read(buffer);
        }

        append();
        builder.setLength(0);
        if (toParse.length() >= maxChunkChars) {
            if (parser.find(toParse)) {
                toParse.setLength(0);
                return true;
            }
            toParse.setLength(0); //can't read entire file at once; break into segments
        }
        return parser.find(toParse);
    }

    @Override
    public void tuneParameters(int bufferSize, int maxChunkBytes) {
        buffer = new byte[bufferSize];
        maxChunkChars = maxChunkBytes / 2;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Clears the current data buffer of all contents. */
    private void clearBuffer() {
        for (int i = 0; i < buffer.length; i++) {
            buffer[i] = (byte) 0;
        }
    }

    /**
     * Appends the current {@link #builder} to {@link #toParse},
     *  automatically stripping out XML tags if necessary.
     */
    private void append() {
        String toAppend;
        if (stripXML) {
            toAppend = EXCLUDE_XML.matcher(builder).replaceAll("");
        } else {
            toAppend = builder.toString();
        }
        toParse.append(toAppend);
    }

    /** @return a boolean, {@code true} if {@code name} ends with a prohibited extension */
    private boolean endsWith(String name) {
        if (exclude == null) { return false; }
        if (Main.excludeList().isFileExcluded(null, Ext.get(name))) { return true; }

        for (int i = 0; i < exclude.length; i++) {
            if (name.endsWith(exclude[i])) {
                return true;
            }
        }
        return false;
    }
}
