package edu.uwyo.crawl.database;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitor;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;

import edu.uwyo.Main;

/**
 * A visitor class integrated with {@link Database} and used for crawling directories.
 * Call {@link #reset(ExcludeList)} at the start of each new search
 *  to ensure that file statistics are updated correctly.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class BasicFileVisitor implements FileVisitor<Path> {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CONSTANTS                    ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private final Object LOCK = new Object();

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private int numFound;
    private int numSkipped;
    private ExcludeList excludeList;

    private final Crawler crawler;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link BasicFileVisitor}.
     * @param crawler the {@link Crawler} owning this visitor
     */
    protected BasicFileVisitor(Crawler crawler) {
        this.crawler = crawler;
        reset(null);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs)
            throws IOException {
        if (excludeList.isDirExcluded(dir.toString())) {
            return FileVisitResult.SKIP_SUBTREE;
        }
        return FileVisitResult.CONTINUE;
    }

    @Override
    public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
        if (!attrs.isDirectory() && crawler.add(file, attrs)) {
            synchronized (LOCK) {
                numFound++;
            }
        } else {
            synchronized (LOCK) {
                numSkipped++;
            }
        }

        return FileVisitResult.CONTINUE;
    }

    @Override
    public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
        return FileVisitResult.CONTINUE;
    }

    @Override
    public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
        return FileVisitResult.CONTINUE;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * @return the integer total number of files crawled during the last search
     *  (this method is thread-safe)
     */
    protected int numTotal() {
        synchronized (LOCK) {
            return numFound + numSkipped;
        }
    }

    /**
     * @return the integer number of files found in the last search (this method is thread-safe)
     */
    protected int numFound() {
        synchronized (LOCK) {
            return numFound;
        }
    }

    /**
     * @return the integer number of files skipped in the last search (this method is thread-safe)
     */
    protected int numSkipped() {
        synchronized (LOCK) {
            return numSkipped;
        }
    }

    /**
     * Resets this visitor's statistics.
     * This method is thread-safe.
     * @param excludeList the {@link ExcludeList} of global file and extension exclusions
     */
    protected void reset(ExcludeList excludeList) {
        synchronized (LOCK) {
            this.excludeList = excludeList;
            numFound = 0;
            numSkipped = 0;
        }
    }

    /**
     * Searches the given list of files.
     * This method functions identically to the {@link FileVisitor} operations in this class,
     *  but operates instead on a list of provided files, rather than walking a file tree.
     * @param toSearch the array of String file paths to search
     * @throws IOException if there was a problem accessing or reading a file in {@code toSearch}
     */
    protected void searchFiles(String[] toSearch) throws IOException {
        for (int i = 0; i < toSearch.length; i++) {
            if (toSearch[i] == null) { continue; } //sanity check; should never actually happen
            Path path = Paths.get(toSearch[i]);
            BasicFileAttributes attrs;
            try {
                attrs = Files.readAttributes(path, BasicFileAttributes.class);
            } catch (IOException e) {
                synchronized (LOCK) {
                    numSkipped++;
                }
                continue;
            }

            if (!attrs.isDirectory() && !Main.excludeList().isFileInExcludedDir(toSearch[i])
                    && crawler.add(path, attrs)) {
                synchronized (LOCK) {
                    numFound++;
                }
            } else {
                synchronized (LOCK) {
                    numSkipped++;
                }
            }
        }
    }
}
