package edu.uwyo.crawl.database;

import java.io.IOException;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.EnumSet;
import java.util.Set;

import net.javamon.dev.Console;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;

/**
 * A class used for crawling directories for files.
 * Each found file must be added via {@link #add(Path, BasicFileAttributes)}.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class Crawler {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private FileSearchControl searchControl;
    private boolean commandLine;
    private boolean verbose;

    private final Database database;
    private final BasicFileVisitor visitor;
    private final Set<FileVisitOption> visitOptions;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link Crawler}.
     * @param database the {@link Database} used for storing results
     */
    protected Crawler(Database database) {
        this.database = database;
        visitor = new BasicFileVisitor(this);
        visitOptions = EnumSet.noneOf(FileVisitOption.class);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the integer number of files searched so far (this method is thread-safe) */
    protected int numSearched() {
        return visitor.numTotal();
    }

    /**
     * Enables command-line operation of the file crawler.
     * Regular displayed messages will be logged to the console instead.
     * @param verbose a boolean; if {@code true}, additional console messages will be printed
     */
    protected void setCommandLine(boolean verbose) {
        this.commandLine = true;
        this.verbose = verbose;
    }

    /**
     * Adds a file to the database, but only if it is searchable
     *  under the current settings of {@link #searchControl}.
     * Only files (not directories) should be passed into this method.
     * @param path the {@link Path} of the file to be added
     * @param attributes the {@link BasicFileAttributes} of the file
     * @return a boolean, {@code true} if the file was added and {@code false} otherwise
     */
    protected boolean add(Path path, BasicFileAttributes attributes) {
        if (!searchControl.isSearchable(path.toString(), attributes)) {
            return false;
        }

        if (verbose) {
            Console.logInfo(path.toString());
        }
        database.addResult(path.toFile());
        return true;
    }

    /**
     * Crawls the given directory using the current settings of {@code searchControl}.
     * This method performs a multitude of generally fast I/O operations,
     *  but may be <u>extremely slow</u> if files are being searched for content.
     * @param searchControl the {@link FileSearchControl} used for determining which files
     *  will be returned by a search
     * @param excludeList the {@link ExcludeList} of global file and extension exclusions
     * @param displayPrompt a boolean, {@code true} if results should be displayed to the user
     * @param toSearch an array of String files to search; overrides the default search mechanism
     *  if not {@code null}
     * @throws IOException if an error occurred while attempting to crawl the directory
     */
    protected void run(FileSearchControl searchControl, ExcludeList excludeList,
            boolean displayPrompt, String[] toSearch) throws IOException {
        this.searchControl = searchControl;
        visitor.reset(excludeList);

        long start = System.nanoTime();
        if (toSearch != null) {
            visitor.searchFiles(toSearch);
        } else {
            Files.walkFileTree(searchControl.root(), visitOptions, searchControl.numLevels(),
                    visitor);
        }
        long end = System.nanoTime();

        float time = (end - start) / 1000000000.0f;
        float rate = visitor.numTotal() / time;

        if (commandLine) {
            Console.logInfo(Resources.__("cmdCrawlResultMsg", searchControl.root().toString(),
                    visitor.numTotal(), time, rate, visitor.numFound(), visitor.numSkipped()));
        } else if (displayPrompt) {
            Prompt.messageScroll(Resources.__("mainCrawlResultH"),
                    Resources.__("mainCrawlResultMsg", searchControl.root().toString(),
                            visitor.numTotal(), time, rate, visitor.numFound(),
                            visitor.numSkipped()));
        }
    }
}
