package edu.uwyo.crawl.database;

import java.io.IOException;
import java.io.InputStream;

/**
 * An input stream designed to be paired with {@link Serializer} for reading objects.
 * The current stream can be reassigned via {@link #setInput(InputStream)}.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class DataInput {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private InputStream input;
    private long bytesRead;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the long total number of bytes that have been read thus far */
    public long bytesRead() {
        return bytesRead;
    }

    /**
     * Reads {@code buffer.length} bytes from the current input stream.
     * @param buffer the array into which bytes should be read
     * @return the integer total number of bytes read
     * @throws IOException if an I/O error occurs
     */
    public int read(byte[] buffer) throws IOException {
        int read = input.read(buffer);
        if (read == -1) {
            throw new IOException("end of input stream");
        }
        bytesRead += read;
        return read;
    }

    /**
     * Reads {@code length} bytes from the current input stream into the given {@code buffer},
     *  starting at {@code offset}.
     * @param buffer the array into which bytes should be read
     * @param offset the integer offset in {@code buffer} at which to start placing the read bytes
     * @param length the integer maximum number of bytes to read
     * @return the integer total number of bytes read
     * @throws IOException if an I/O error occurs
     */
    public int read(byte[] buffer, int offset, int length) throws IOException {
        int read = input.read(buffer, offset, length);
        if (read == -1) {
            throw new IOException("end of input stream");
        }
        bytesRead += read;
        return read;
    }

    /**
     * Reads a signed byte from the current input stream.
     * @return the byte that was read
     * @throws IOException if an I/O error occurs
     */
    public byte readByte() throws IOException {
        return (byte) readUnsignedByte();
    }

    /**
     * Reads an unsigned byte from the current input stream.
     * @return the integer unsigned byte that was read
     * @throws IOException if an I/O error occurs
     */
    public int readUnsignedByte() throws IOException {
        int readByte = input.read();
        if (readByte == -1) {
            throw new IOException("end of input stream");
        }
        int toReturn = (byte) readByte & 0xFF;
        bytesRead++;
        return toReturn;
    }

    /**
     * Reads a 1-byte boolean from the current input stream.
     * A byte value of 0 is {@code false} and all other values are {@code true}.
     * @return the boolean that was read
     * @throws IOException if an I/O error occurs
     */
    public boolean readBoolean() throws IOException {
        return readByte() != 0;
    }

    /**
     * Reads a 2-byte signed short from the current input stream.
     * @return the short that was read
     * @throws IOException if an I/O error occurs
     */
    public short readShort() throws IOException {
        return (short) readUnsignedShort();
    }

    /**
     * Reads a 2-byte unsigned short from the current input stream.
     * @return the integer unsigned short that was read
     * @throws IOException if an I/O error occurs
     */
    public int readUnsignedShort() throws IOException {
        return (readUnsignedByte() * 256) + readUnsignedByte();
    }

    /**
     * Reads a 2-byte character from the current input stream.
     * @return the character that was read
     * @throws IOException if an I/O error occurs
     */
    public char readChar() throws IOException {
        return (char) ((readUnsignedByte() << 8) | (readUnsignedByte() & 0xFF));
    }

    /**
     * Reads an integer from the current input stream.
     * Depending on the size of the integer, 1-5 bytes may actually be read.
     * @return the integer that was read
     * @throws IOException if an I/O error occurs
     */
    public int readInt() throws IOException {
        return unpackInt();
    }

    /**
     * Reads a long from the current input stream.
     * Depending on the size of the long, 1-10 bytes may actually be read.
     * @return the long that was read
     * @throws IOException if an I/O error occurs
     */
    public long readLong() throws IOException {
        return unpackLong();
    }

    /**
     * Reads a float from the current input stream.
     * Depending on the size of the float, 1-5 bytes may actually be read.
     * @return the float that was read
     * @throws IOException if an I/O error occurs
     */
    public float readFloat() throws IOException {
        return Float.intBitsToFloat(readInt());
    }

    /**
     * Reads a double from the current input stream.
     * Depending on the size of the double, 1-10 bytes may actually be read.
     * @return the double that was read
     * @throws IOException if an I/O error occurs
     */
    public double readDouble() throws IOException {
        return Double.longBitsToDouble(readLong());
    }

    /**
     * Reads a UTF-8 character string from the current input stream.
     * This method first reads a variable-size integer header indicating the string's length,
     *  followed by the string itself.
     * @return the String that was read
     * @throws IOException if an I/O error occurs
     */
    public String readUTF() throws IOException {
        char[] characters = new char[unpackInt()];
        for (int i = 0; i < characters.length; i++) {
            characters[i] = (char) unpackInt();
        }
        return new String(characters);
    }

    /**
     * Sets the current input stream.
     * The calling class is responsible for opening and closing this stream.
     * @param input the {@link InputStream} to use
     */
    protected void setInput(InputStream input) {
        this.input = input;
        bytesRead = 0L;
    }

    /**
     * Closes the current input stream, if any.
     * @throws IOException if an I/O error occurs.
     */
    protected void close() throws IOException {
        if (input != null) {
            input.close();
        }
    }

    /**
     * Reads 1-5 bytes from the current output stream and unpacks them into an integer.
     * Optimized for smaller integer values, in which only a single byte might need to be read.
     * @return the integer that was read
     * @throws IOException if an I/O error occurs
     */
    private int unpackInt() throws IOException {
        int toReturn = 0;
        int valueByte;
        do {
            valueByte = readByte();
            toReturn = (toReturn << 7 ) | (valueByte & 0x7F);
        } while ((valueByte & 0x80) == 0);

        return toReturn;
    }

    /**
     * Reads 1-10 bytes from the current output stream and unpacks them into a long.
     * Optimized for smaller long values, in which only a single byte might need to be read.
     * @return the long that was read
     * @throws IOException if an I/O error occurs
     */
    private long unpackLong() throws IOException {
        long toReturn = 0;
        int valueByte;
        do {
            valueByte = readByte();
            toReturn = (toReturn << 7) | (valueByte & 0x7F);
        } while ((valueByte & 0x80)==0);

        return toReturn;
    }
}
