package edu.uwyo.crawl.database;

import java.io.IOException;
import java.io.OutputStream;

/**
 * An output stream designed to be paired with {@link Serializer} for writing objects.
 * The current stream can be reassigned via {@link #setOutput(OutputStream)}.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class DataOutput {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private OutputStream output;
    private long bytesWritten;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the long total number of bytes written thus far */
    public long bytesWritten() {
        return bytesWritten;
    }

    /**
     * Writes all bytes in {@code bytes} to the current output stream.
     * No bytes will be written if {@code bytes} is empty.
     * @param bytes the array of bytes to be written
     * @throws IOException if an I/O error occurs
     */
    public void write(byte[] bytes) throws IOException {
        output.write(bytes);
        bytesWritten += bytes.length;
    }

    /**
     * Writes {@code length} bytes from {@code bytes} to the current output stream,
     *  starting at {@code offset}.
     * @param bytes the array of bytes to be written
     * @param offset the integer offset from which to start writing
     * @param length the integer number of bytes to write
     * @throws IOException if an I/O error occurs
     */
    public void write(byte[] bytes, int offset, int length) throws IOException {
        output.write(bytes, offset, length);
        bytesWritten += length;
    }

    /**
     * Writes a single byte to the current output stream.
     * @param value the integer byte to write; all preceding (non-byte) bits are ignored
     * @throws IOException if an I/O error occurs
     */
    public void writeByte(int value) throws IOException {
        output.write(value);
        bytesWritten++;
    }

    /**
     * Writes a 1-byte boolean to the current output stream.
     * The value {@code 0} is considered {@code false}, while all other values are {@code true}.
     * @param value the boolean to write
     * @throws IOException if an I/O error occurs
     */
    public void writeBoolean(boolean value) throws IOException {
        writeByte(value ? 1 : 0);
    }

    /**
     * Writes a 2-byte short to the current output stream.
     * @param value the short to write; all preceding (non-short) bits are ignored
     * @throws IOException if an I/O error occurs
     */
    public void writeShort(short value) throws IOException {
        writeByte((0xFF & (value >> 8)));
        writeByte((0xFF & value));
    }

    /**
     * Writes a 2-byte character to the current output stream.
     * @param value the character to write
     * @throws IOException if an I/O error occurs
     */
    public void writeChar(char value) throws IOException {
        writeByte(value >>> 8);
        writeByte(value);
    }

    /**
     * Writes an integer to the current output stream.
     * The integer may occupy between 1 to 5 bytes, but is optimized for small numbers,
     *  in which case the fewest necessary bytes will be written.
     * @param value the integer to write
     * @throws IOException if an I/O error occurs
     */
    public void writeInt(int value) throws IOException {
        packInt(value);
    }

    /**
     * Writes a long to the current output stream.
     * The long may occupy between 1 to 10 bytes, but is optimized for small numbers,
     *  in which case the fewest necessary bytes will be written.
     * @param value the long to write
     * @throws IOException if an I/O error occurs
     */
    public void writeLong(long value) throws IOException {
        packLong(value);
    }

    /**
     * Writes a float to the current output stream.
     * @param value the float to write; will be converted internally to an integer
     * @throws IOException if an I/O error occurs
     */
    public void writeFloat(float value) throws IOException {
        writeInt(Float.floatToIntBits(value));
    }

    /**
     * Writes a double to the current output stream.
     * @param value the double to write; will be converted internally to a long
     * @throws IOException if an I/O error occurs
     */
    public void writeDouble(double value) throws IOException {
        writeLong(Double.doubleToLongBits(value));
    }

    /**
     * Writes a UTF-8 character string to the current output stream.
     * This method always writes an integer header containing the string's length,
     *  followed by the string itself.
     * Optimized to reduce space for ASCII strings.
     * @param value the String to write
     * @throws IOException if an I/O error occurs
     */
    public void writeUTF(String value) throws IOException {
        int length = value.length();
        packInt(length);
        for (int i = 0; i < length; i++) {
            packInt(value.charAt(i));
        }
    }

    /**
     * Sets the current output stream.
     * The calling method is responsible for opening and closing this stream.
     * @param output the {@link OutputStream} to use
     */
    protected void setOutput(OutputStream output) {
        this.output = output;
        bytesWritten = 0L;
    }

    /**
     * Closes the current output stream, if any.
     * @throws IOException if an I/O error occurs.
     */
    protected void close() throws IOException {
        if (output != null) {
            output.close();
        }
    }

    /**
     * Packs an integer into a smaller number of bytes if possible,
     *  then places it in the current output stream.
     * The resulting integer may occupy between 1 to 5 bytes and is optimized for smaller values.
     * @param value the integer to pack
     * @throws IOException if an I/O error occurs
     */
    private void packInt(int value) throws IOException {
        int shift = (value & ~0x7F);
        if (shift != 0) {
            shift = 31 - Integer.numberOfLeadingZeros(value);
            shift -= shift % 7; //round down to nearest multiple of 7
            while (shift != 0) {
                writeByte((byte) ((value >>> shift) & 0x7F));
                shift -= 7;
            }
        }
        writeByte((byte) ((value & 0x7F) | 0x80));
    }

    /**
     * Packs a long into a smaller number of bytes if possible,
     *  then places it in the current output stream.
     * The resulting long may occupy between 1 to 10 bytes and is optimized for smaller values.
     * @param value the long to pack
     * @throws IOException if an I/O error occurs
     */
    private void packLong(long value) throws IOException {
        int shift = 63 - Long.numberOfLeadingZeros(value);
        shift -= shift % 7; //round down to nearest multiple of 7
        while (shift!=0) {
            writeByte((byte) ((value>>>shift) & 0x7F));
            shift -= 7;
        }
        writeByte((byte) ((value & 0x7F) | 0x80));
    }
}
