package edu.uwyo.crawl.database;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import javax.swing.JOptionPane;

import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import net.javamon.dev.Console;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;

/**
 * A database containing file results and user settings.
 * Searches can be run via {@link #search(boolean, boolean)}
 *  and {@link #findDuplicates(boolean, boolean, boolean, boolean, boolean, String[], long[], String)}.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class Database {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static int LARGE_NAME_SEARCH;
    private static int LARGE_CONTENT_SEARCH;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private Results results;
    private boolean developerMode;
    private boolean commandLine;

    private final DataInput input;
    private final DataOutput output;
    private final Crawler crawler;
    private final DuplicateFinder duplicateFinder;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link Database}.
     *  line and should not produce any interactive messages
     * @param searchControl the {@link FileSearchControl} to use for determining which files
     *  are searched
     * @param searchControlSerializer the {@link Serializer} used for writing and reading
     *  all search settings
     */
    public Database(FileSearchControl searchControl,
            Serializer<FileSearchControl> searchControlSerializer) {
        input = new DataInput();
        output = new DataOutput();
        crawler = new Crawler(this);
        duplicateFinder = new DuplicateFinder();

        Results.searchControl = searchControl;
        Results.INSTANCE.searchControlSerializer = searchControlSerializer;
        results = new Results();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Initializes static database parameters.
     * Call this once at application start, before any instances of this class have been created.
     */
    public static void init() {
        try {
            LARGE_NAME_SEARCH = Integer.parseInt(
                    Resources.readSetting("dbDuplicateWarningName"));
        } catch (Exception e) {
            LARGE_NAME_SEARCH = 50000;
        }

        try {
            LARGE_CONTENT_SEARCH = Integer.parseInt(
                    Resources.readSetting("dbDuplicateWarningContent"));
        } catch (Exception e) {
            LARGE_CONTENT_SEARCH = 1000;
        }
    }

    /**
     * Determines how many files are in the current result set.
     * @return the integer total number of files stored in this result set
     */
    public int numFiles() {
        return results.resultTree().size();
    }

    /**
     * Sets whether developer messages should be displayed.
     * @param enable a boolean, {@code true} if developer messages should be enabled
     */
    public void setDeveloperMode(boolean enable) {
        developerMode = enable;
    }

    /**
     * Enables command-line operation of the database.
     * Regular displayed messages will be logged to the console instead.
     * @param verbose a boolean; if {@code true}, additional console messages will be printed
     */
    public void setCommandLine(boolean verbose) {
        commandLine = true;
        crawler.setCommandLine(verbose);
    }

    /** @return the current search {@link Results} */
    public Results results() {
        return results;
    }

    /** @return the integer number of files searched so far (this method is thread-safe) */
    public int numSearched() {
        return crawler.numSearched();
    }

    /**
     * Loads a search result set into the database.
     * <b>Any current search results and search settings will be lost.</b>
     * After this method is called, the user interface should typically be refreshed
     *  to reflect the loaded results.
     * @param snapshot the {@link File} containing search results to be loaded
     * @return a boolean, {@code true} if the snapshot was successfully loaded
     */
    public boolean loadSearchResults(File snapshot) {
        try {
            input.setInput(new BufferedInputStream(new FileInputStream(snapshot), 8192));
            Results newResults = Results.INSTANCE.deserialize(input);
            input.close();
            results = newResults;
            return true;
        } catch (Exception e) {
            if (commandLine) {
                Console.logInfo(Resources.__("cmdErrorMsg", "loadSearchResults(File)",
                                snapshot.getAbsolutePath(), Console.getStack(e, false)));
            } else {
                e.printStackTrace();
                if (developerMode) {
                    Prompt.errorScroll(Resources.__("mainDeveloperErrorH"),
                            Resources.__("mainDeveloperErrorMsg", "loadSearchResults(File)",
                                    snapshot.getAbsolutePath(), Console.getStack(e, true)));
                }
            }
            return false;
        }
    }

    /**
     * Saves a snapshot of the current search results (and the settings that produced them)
     *  to a file.
     * This file can be loaded via {@link #loadSearchResults(File)}.
     * @param destination the {@link File} where search results should be saved
     * @return a boolean, {@code true} if the snapshot was successfully saved
     */
    public boolean saveSearchResults(File destination) {
        try {
            output.setOutput(new BufferedOutputStream(new FileOutputStream(destination), 8192));
            Results.INSTANCE.serialize(output, results);
            output.close();
            return true;
        } catch (Exception e) {
            if (commandLine) {
                Console.logInfo(Resources.__("cmdErrorMsg", "saveSearchResults(File)",
                                destination.getAbsolutePath(), Console.getStack(e, false)));
            } else {
                e.printStackTrace();
                if (developerMode) {
                    Prompt.errorScroll(Resources.__("mainDeveloperErrorH"),
                            Resources.__("mainDeveloperErrorMsg", "saveSearchResults(File)",
                                    destination.getAbsolutePath(), Console.getStack(e, true)));
                }
            }
            return false;
        }
    }

    /**
     * Searches for files that meet the current search criteria.
     * Metadata-based searches (size, date modified, etc.) will be generally fast,
     *  while content-based searches can be <u>very slow</u> depending on disk and network speed.
     * <p>
     * <i>This method changes the current search results if {@code useCached == false}.
     * If a snapshot was previously loaded via {@link #loadSearchResults(File)},
     *  that snapshot will be lost (however, the snapshot's file will <u>not</u> be modified).</i>
     * @param useCached a boolean; if {@code true}, the previous search results will be returned
     *  <i>without</i> performing a new search
     * @param searchCurrResults a boolean; if {@code true}, the search will be performed
     *  on the current search results only
     * @return a {@link FileDisplayTree} containing the result of the search
     */
    public FileDisplayTree search(boolean useCached, boolean searchCurrResults) {
        if (!useCached) {
            String[] toSearch = null;
            if (searchCurrResults) {
                toSearch = results.resultTree().getPaths();
            }

            results.resultTree().clear(results.searchControl().root().toString());
            results.resultTree().setFlatMode(false);
            try {
                crawler.run(results.searchControl(), results.excludeList(),
                        developerMode && !commandLine, toSearch);
            } catch (Exception e) {
                if (commandLine) {
                    Console.logInfo(Resources.__("cmdErrorMsg", "search(boolean)",
                                    "", Console.getStack(e, false)));
                } else {
                    e.printStackTrace();
                    if (developerMode) {
                        Prompt.errorScroll(Resources.__("mainDeveloperErrorH"),
                                Resources.__("mainDeveloperErrorMsg", "search(boolean)",
                                        "", Console.getStack(e, true)));
                    }
                }
            }
        }
        return results.resultTree();
    }

    /**
     * Searches for duplicate files by comparing file contents.
     * This method first calls {@link #search(boolean, boolean)}, then filters the list of results
     *  such that only duplicate files will be displayed.
     * <b>Searching for duplicates can be <u>extremely slow</u> depending on the search mode
     *  and should be avoided on large directories.</b>
     * This is especially true for {@code exact} file searches,
     *  which should only be run on a few files at a time.
     * <p>
     * <i>This method changes the current search results if {@code useCached == false}.
     * If a snapshot was previously loaded via {@link #loadSearchResults(File)},
     *  that snapshot will be lost (however, the snapshot's file will <u>not</u> be modified).</i>
     * @param useCached a boolean; if {@code true}, the previous search results will be returned
     *  <i>without</i> performing a new search
     * @param names a boolean, {@code true} if files must have identical names
     *  to be included in the results
     * @param content a boolean, {@code true} if files must have similar content
     *  to be included in the results
     * @param exact a boolean, {@code true} if files must have <u>identical</u> contents
     *  to be considered a match; otherwise, files are only compared using their header bytes
     * @param searchCurrResults a boolean; if {@code true}, the search will be performed
     *  on the current search results only (duplicates will be filtered from these results)
     * @param searchPaths the array of String paths to search, or {@code null} if current
     *  or new search results should be used;
     *  if this is not {@code null}, an equal-length {@code searchSizes} must also be provided
     * @param searchSizes the array of long sizes to search
     * @param searchRoot the String search root; only used if {@code searchPaths != null}
     * @return a {@link FileDisplayTree} containing the result of the search
     */
    public FileDisplayTree findDuplicates(boolean useCached, boolean names, boolean content,
            boolean exact, boolean searchCurrResults, String[] searchPaths, long[] searchSizes,
            String searchRoot) {
        if (!useCached) {
            if (!content && !names) { names = true; }

            if (!searchCurrResults && searchPaths == null) {
                search(false, false); //if searching current results, no need to re-search files
            }

            final int numFiles = numFiles();
            if (!commandLine) {
                if ((content && numFiles >= LARGE_CONTENT_SEARCH)
                        || (!content && numFiles >= LARGE_NAME_SEARCH)) {
                    int choice = Prompt.show(Resources.__("mainDuplicateSearchConfirmH"),
                            Resources.__("mainDuplicateSearchConfirmMsg", numFiles));
                    if (choice != JOptionPane.YES_OPTION) {
                        return results.resultTree();
                    }
                }
            }

            if (searchPaths != null) {
                results.resultTree().clear(searchRoot);
                for (int i = 0; i < searchPaths.length; i++) {
                    results.resultTree().add(null, searchPaths[i], false, searchSizes[i]);
                }
            }

            long start = 0L;
            long end = 0L;
            if (!commandLine) { Console.windowLoad.suppressInvokable(true); }
            try {
                start = System.nanoTime();
                duplicateFinder.find(results.resultTree(), names, content, exact);
                end = System.nanoTime();

            } catch (Exception e) {
                if (commandLine) {
                    Console.logInfo(Resources.__("cmdErrorMsg",
                                    "findDuplicates(boolean, boolean, boolean, boolean)",
                                    useCached + ", " + names + ", " + content + ", " + exact,
                                    Console.getStack(e, false)));
                } else {
                    e.printStackTrace();
                    if (developerMode) {
                        Prompt.errorScroll(Resources.__("mainDeveloperErrorH"),
                                Resources.__("mainDeveloperErrorMsg",
                                        "findDuplicates(boolean, boolean, boolean, boolean)",
                                        useCached + ", " + names + ", " + content + ", " + exact,
                                        Console.getStack(e, true)));
                    }
                }
            }

            if ((developerMode || commandLine) && start != 0L && end != 0L) {
                float time = (end - start) / 1000000000.0f;
                float rate = numFiles / time;

                if (commandLine) {
                    Console.logInfo(Resources.__("cmdDuplicateResultMsg",
                            results.searchControl().root().toString(),
                            numFiles, time, rate, names + "", content + "", exact + ""));
                } else {
                    Prompt.messageScroll(Resources.__("mainCrawlDuplicateResultH"),
                            Resources.__("mainCrawlDuplicateResultMsg",
                                    results.searchControl().root().toString(),
                                    numFiles, time, rate, names + "", content + "", exact + ""));
                }
            }
        }
        return results.resultTree();
    }

    /**
     * Adds a file to the current result tree.
     * @param file the {@link File} to add
     */
    protected void addResult(File file) {
        results.resultTree().add(null, file.getAbsolutePath(), file.isDirectory(), file.length());
    }
}
