package edu.uwyo.crawl.database;

import java.io.File;
import java.io.FileInputStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import net.javamon.dev.ReaderTools;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.HTML;

/**
 * A class used for finding duplicate entries within a set of files that have already been found.
 * <i>This search mechanism is generally very slow because it must read file contents on disk.</i>
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class DuplicateFinder {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static int headerSize = 1024;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link DuplicateFinder}. */
    protected DuplicateFinder() {}

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the size of the header used for detecting duplicates.
     * @param headerSize the integer header size to read from each file, in bytes;
     *  files containing an identical header are considered identical unless exact matching
     *  is enabled
     */
    public static void setHeaderSize(int headerSize) {
        DuplicateFinder.headerSize = headerSize;
    }

    /**
     * Crawls the given result list and searches for duplicate files.
     * The result of this search will be stored in {@code results} after this method completes,
     *  and the original search {@code results} will be lost.
     * <p>
     * Depending on the search mode and number of files,
     *  this method may be <u>extremely</u> slow.
     * In particular, {@code exact} file searches might consume large amounts of memory
     *  and should only be run on small data sets (preferably, only a few files at a time).
     * If a files contents cannot be read, it will not be included in the results.
     * @param results the {@link FileDisplayTree} containing all previously found results;
     *  this list will be filtered for duplicates
     * @param names a boolean, {@code true} if files must have identical names
     *  to be included in the results
     * @param content a boolean, {@code true} if files must have similar content
     *  to be included in the results
     * @param exact a boolean, {@code true} if files must have <u>identical</u> contents
     *  to be considered a match; otherwise, files are only compared
     *  using their first {@link #headerSize} bytes
     */
    protected void find(FileDisplayTree results, boolean names, boolean content, boolean exact) {
        Scratch scratchName = new Scratch();
        Scratch scratchSize = new Scratch();
        Scratch scratchHeader = new Scratch();
        Scratch scratchExact = new Scratch();

        if (names) {
            putIntoScratch(results, scratchName, true, false);
        } else {
            putIntoScratch(results, scratchSize, false, true);
        }

        if (names && !content) {
            results.clear(Resources.__("mainDuplicateNameOnly"));
        } else if (!names && content && !exact) {
            results.clear(Resources.__("mainDuplicateContentOnly"));
        } else if (!names && content && exact) {
            results.clear(Resources.__("mainDuplicateContentOnlyExact"));
        } else if (names && content && !exact) {
            results.clear(Resources.__("mainDuplicateAll"));
        } else {
            results.clear(Resources.__("mainDuplicateAllExact"));
        }
        results.setFlatMode(true);

        if (names) {
            Set<String> nameSet = scratchName.keySet();
            for (String name : nameSet) {
                HashSet<String> duplicates = scratchName.get(name);
                if (duplicates.size() > 1) {
                    for (String path : duplicates) {
                        if (!content) {
                            results.add(name, path, false, new File(path).length());
                        } else {
                            scratchSize.add(new File(path).length() + "", path);
                        }
                    }
                }
            }
            if (!content) { return; }
        }

        Set<String> sizeSet = scratchSize.keySet();
        for (String size : sizeSet) {
            HashSet<String> duplicates = scratchSize.get(size);
            if (duplicates.size() > 1) {
                for (String path : duplicates) {
                    String header = getHeader(path);
                    if (header != null) {
                        scratchHeader.add(header, path);
                    }
                }
            }
        }

        if (exact) {
            Set<String> headerSet = scratchHeader.keySet();
            for (String header : headerSet) {
                HashSet<String> duplicates = scratchHeader.get(header);
                if (duplicates.size() > 1) {
                    for (String path : duplicates) {
                        String contents = getExact(path);
                        if (contents != null) {
                            scratchExact.add(contents, path);
                        }
                    }
                }
            }
            scratchHeader = scratchExact;
        }

        Set<String> headerSet = scratchHeader.keySet();
        for (String header : headerSet) {
            HashSet<String> duplicates = scratchHeader.get(header);
            String index = null;
            if (duplicates.size() > 1) {
                for (String path : duplicates) {
                    if (names) {
                        File file = new File(path);
                        results.add(file.getName(), path, false, file.length());
                    } else {
                        if (index == null) {
                            index = HTML.wrap(Resources.__("mainDuplicateFound", path));
                        }
                        results.add(index, path, false, new File(path).length());
                    }
                }
            }
        }
    }

    /**
     * Recursively puts the contents of {@code curr} into a scratch set.
     * Should be called at the start of {@link #find(FileDisplayTree, boolean, boolean, boolean)}.
     * <p>
     * Only one of {@code name} or {@code size} can be {@code true} when this method is called.
     * @param curr the {@link FileDisplayTree} to be placed into the scratch set
     * @param dest the {@link Scratch} where entries from {@code curr} should be placed
     * @param name a boolean; if {@code true}, file names will be placed into the scratch set
     * @param size a boolean; if {@code true}, file sizes will be placed into the scratch set
     */
    private void putIntoScratch(FileDisplayTree curr, Scratch dest, boolean name, boolean size) {
        if (curr.getFilePath() != null) {
            File file = new File(curr.getFilePath());
            if (!file.isDirectory()) {
                if (name) {
                    dest.add(file.getName(), curr.getFilePath());
                } else if (size) {
                    dest.add(file.length() + "", curr.getFilePath());
                }
                return;
            }
        }

        HashMap<String, FileDisplayTree> childrenMap = curr.getChildren();
        Set<String> children = childrenMap.keySet();
        for (String child : children) {
            putIntoScratch(childrenMap.get(child), dest, name, size);
        }
    }

    /**
     * Obtains the header for a given file.
     * This method reads up to {@link #headerSize} bytes from disk and can be quite slow.
     * @param path the String file path to be read
     * @return a String representation of the file's header,
     *  or {@code null} if it could not be read
     */
    private static String getHeader(String path) {
        try {
            FileInputStream input = new FileInputStream(path);
            byte[] tempHeader = new byte[headerSize];
            input.read(tempHeader);
            input.close();
            return ReaderTools.toHex(tempHeader);
        } catch (Exception e) {}

        return null;
    }

    /**
     * Obtains the exact file contents.
     * This method reads the entire file and can be both slow and memory intensive;
     *  consequently, exact-file searches should only be run on small file sets.
     * @param path the String file path to be read
     * @return a String representation of the file's contents,
     *  or {@code null} if it could not be read
     */
    private static String getExact(String path) {
        try {
            byte[] contents = Files.readAllBytes(Paths.get(path));
            return ReaderTools.toHex(contents);
        } catch (Exception e) {}

        return null;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CLASSES                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** A helper class for storing the temporary results of a duplicate search. */
    private static class Scratch {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* FIELDS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        private HashMap<String, HashSet<String>> map;

        ///////////////////////////////////////////////////////////////////////////////////////////
        /* CONSTRUCTORS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        /** Creates a new {@link Scratch}. */
        protected Scratch() {
            map = new HashMap<String, HashSet<String>>();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////
        /* METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        /** @return the {@link Set} of keys currently contained in this scratch set */
        protected Set<String> keySet() {
            return map.keySet();
        }

        /** @return the {@link HashSet} entries associated with {@code key} */
        protected HashSet<String> get(String key) {
            return map.get(key);
        }

        /**
         * Adds a new entry to this scratch set.
         * A new underlying hash set will be automatically created for {@code key}
         *  if it does not already exist.
         * @param key the String key of the value to add
         * @param value the String value to be added
         */
        protected void add(String key, String value) {
            if (map.get(key) == null) {
                map.put(key, new HashSet<String>());
            }
            map.get(key).add(value);
        }
    }
}
