package edu.uwyo.crawl.database;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A list containing files and file types that should be excluded from all searches.
 * A file or extension can be checked against this list
 *  by calling {@link #isFileExcluded(String, String)}.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class ExcludeList implements Serializable, Serializer<ExcludeList> {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used for serialized; <b>do not modify!</b> */
    public static final ExcludeList INSTANCE = new ExcludeList();

    private static final long serialVersionUID = -5626671390023171389L;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private HashSet<String> files;
    private HashSet<String> extensions;
    private Pattern pattern;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link ExcludeList}. */
    protected ExcludeList() {
        files = new HashSet<String>();
        extensions = new HashSet<String>();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void serialize(DataOutput output, ExcludeList instance) throws IOException {
        output.writeInt(instance.files.size());
        for (String file : instance.files) {
            output.writeUTF(file);
        }

        output.writeInt(instance.extensions.size());
        for (String extension : instance.extensions) {
            output.writeUTF(extension);
        }

        output.writeBoolean(instance.pattern != null);
        if (instance.pattern != null) {
            output.writeUTF(instance.pattern.toString());
        }
    }

    @Override
    public ExcludeList deserialize(DataInput input) throws IOException {
        ExcludeList toReturn = new ExcludeList();

        int numFiles = input.readInt();
        for (int i = 0; i < numFiles; i++) {
            toReturn.files.add(input.readUTF());
        }

        int numExtensions = input.readInt();
        for (int i = 0; i < numExtensions; i++) {
            toReturn.extensions.add(input.readUTF());
        }

        if (input.readBoolean()) {
            toReturn.pattern = Pattern.compile(input.readUTF());
        }

        return toReturn;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determines whether a file or extension is in the exclusion list.
     * If both {@code file} and {@code ext} are {@code null}, this method returns {@code false}.
     * <p>
     * If a file is provided, this method checks:
     * <ol>
     *  <li>Whether the specific file has been excluded.</li>
     *  <li>Whether the file's name matches the exclusion pattern
     *  set by {@link #setPattern(Pattern)}.</li>
     * </ol>
     * @param path the String absolute file path to check, or {@code null} if none
     * @param ext the String file extension to check, or {@code null} if none;
     *  should start with a leading period/dot
     * @return a boolean, {@code true} if the file or extension is in the exclusion list
     */
    public boolean isFileExcluded(String path, String ext) {
        if (ext != null && !ext.startsWith(".")) {
            ext = "." + ext;
        }

        if ((ext != null && extensions.contains(ext.toLowerCase()))
                || (path != null && files.contains(path))) {
            return true;
        }

        if (path != null && pattern != null) {
            Matcher m = pattern.matcher(path);
            if (m.find()) {
                return true;
            }
        }

        return false;
    }

    /**
     * Determines whether a folder is in the exclusion list.
     * If the folder is excluded, none of its children should be examined.
     * @param path the String absolute folder path to check
     * @return a boolean, {@code true} if the folder is in the exclusion list
     */
    public boolean isDirExcluded(String path) {
        return files.contains(path);
    }

    /**
     * Determines whether a file is located within an excluded directory.
     * This method must search/compare all files in the exclusion list to determine if {@code file}
     *  is a child of one of those files; consequently, it will tend to be slower
     *  than {@link #isDirExcluded(String)} and {@link #isFileExcluded(String, String)}.
     * @param filePath the {@link File} to check
     * @return a boolean, {@code true} if the file is a child of any folder in the exclusion list
     */
    public boolean isFileInExcludedDir(String filePath) {
        for (String path : files) {
            if (filePath.startsWith(path)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Sets an exclusion pattern; any files that match this pattern will be excluded
     *  when {@link #isFileExcluded(String, String)} is called.
     * The pattern is compared against a file's absolute path, including extension.
     * @param pattern the {@link Pattern} to use for determining which files to exclude,
     *  or {@code null} if pattern-based exclusions should be disabled
     */
    public void setPattern(Pattern pattern) {
        this.pattern = pattern;
    }

    /**
     * Adds a new file to the exclusion list, if it did not already exist.
     * @param path the String absolute file path to add
     */
    public void addFile(String path) {
        if (path == null) { return; }
        files.add(path);
    }

    /**
     * Adds a new extension to the exclusion list, if it did not already exist.
     * @param ext the String file extension to add; should start with a leading period/dot
     */
    public void addExt(String ext) {
        if (ext == null) { return; }

        if (!ext.startsWith(".")) {
            ext = "." + ext;
        }
        extensions.add(ext);
    }

    /**
     * Removes the given file from the exclusion list, if it exists.
     * @param path the String absolute file path to be removed
     */
    public void removeFile(String path) {
        if (path == null) { return; }
        files.remove(path);
    }

    /**
     * Removes the given file extension from the exclusion list, if it exists.
     * @param ext the String file extension to be removed
     */
    public void removeExt(String ext) {
        if (ext == null) { return; }

        if (!ext.startsWith(".")) {
            ext = "." + ext;
        }
        extensions.remove(ext);
    }

    /**
     * Determines whether an array of excluded extensions
     *  matches the current excluded extension set.
     * The order of extensions does not matter.
     * @param extensions the array of String excluded extensions to check
     * @return a boolean, {@code true} if the strings in {@code extensions} matches
     *  the current excluded extension set exactly
     */
    public boolean matchesExt(String[] extensions) {
        if (extensions.length != this.extensions.size()) { return false; }

        for (int i = 0; i < extensions.length; i++) {
            if (!this.extensions.contains(extensions[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * Determines whether an array of excluded files
     *  matches the current excluded file set.
     * The order of files does not matter.
     * @param files the array of String excluded files to check
     * @return a boolean, {@code true} if the strings in {@code files} matches
     *  the current excluded file set exactly
     */
    public boolean matchesFiles(String[] files) {
        if (files.length != this.files.size()) { return false; }

        for (int i = 0; i < files.length; i++) {
            if (!this.files.contains(files[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * Obtains a string view of all current excluded extensions.
     * @return an array of String extensions to be excluded, or {@code null} if none
     */
    public String[] getExt() {
        if (extensions.size() == 0) { return null; }
        String[] toReturn = new String[extensions.size()];

        int curr = 0;
        for (String extension : extensions) {
            toReturn[curr] = extension;
            curr++;
        }
        return toReturn;
    }

    /**
     * Obtains a string view of all current excluded files.
     * @return an array of String filenames to be excluded, or {@code null} if none
     */
    public String[] getFiles() {
        if (files.size() == 0) { return null; }
        String[] toReturn = new String[files.size()];

        int curr = 0;
        for (String path : files) {
            toReturn[curr] = path;
            curr++;
        }
        return toReturn;
    }

    /**
     * Obtains a string view of all pattern-excluded files.
     * @return a String pattern indicating the files to be excluded, or the empty string if none
     */
    public String getPattern() {
        return pattern == null ? "" : pattern.toString();
    }

    /**
     * Sets the excluded extensions.
     * This method overwrites the existing exclusion set.
     * @param extensions the array of String excluded extensions
     */
    public void setExt(String[] extensions) {
        this.extensions.clear();
        if (extensions != null) {
            for (String ext : extensions) {
                this.extensions.add(ext);
            }
        }
    }

    /**
     * Sets the excluded files.
     * This method overwrites the existing exclusion set.
     * @param files the array of String excluded files
     */
    public void setFiles(String[] files) {
        this.files.clear();
        if (files != null) {
            for (String path : files) {
                this.files.add(path);
            }
        }
    }
}
