package edu.uwyo.crawl.database;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;

/**
 * An interface representing the settings used by a search.
 * Implementers of this interface are responsible for determining
 *  on which directory a search should be performed
 *  and whether a candidate file should be included in that search.
 * <p>
 * For maximum flexibility, this interface makes no assumptions
 *  about which parameters of a file are important for determining its inclusion or exclusion.
 * However, keep in mind that <b>examining a file's contents in <i>any</i>
 *  of this interface's methods will incur a <u>very</u> high I/O overhead.</b>
 * File contents should <i>not</i> be examined by default unless expressly requested by the user.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public interface FileSearchControl extends Serializer<FileSearchControl> {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * A delimiter used for separating different search settings
     *  when generating a hash via {@link #getSearchHash()}.
     * In the unlikely case that this string might itself be part of a search term,
     *  a different delimiter may need to be used.
     */
    public static final String HASH_DELIMITER = "__FileSearchControlQueryHash__";

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Determines whether a candidate file should be added to the search results.
     * This method typically examines the file's name, last modified date, etc.
     *  to determine whether the file falls within the user-requested search criteria.
     * <p>
     * <i>Searching a file's contents in this method is allowed,
     *  but will result in <u>much slower</u> searches.
     * Users should be instructed to avoid searching for file contents on large directories.</i>
     * <p>
     * Files examined in this method are <u>not</u> guaranteed to exist,
     *  especially if the file system is changing while the search is running.
     * This method must be prepared to deal with race conditions
     *  e.g. if a file is deleted while it is being read.
     * Any exceptions of this nature should be silently caught and returned as {@code false}.
     * @param path the String absolute file path to be included in the search;
     *  guaranteed to never be a directory
     * @param attributes the {@link BasicFileAttributes} associated with {@code path}
     * @return a boolean, {@code true} if the file should be included in this search
     *  and {@code false} otherwise
     */
    public boolean isSearchable(String path, BasicFileAttributes attributes);

    /**
     * Determines which path should be crawled during the next search.
     * This is usually a directory, but might also be a single file for some searches.
     * The number of subfolders that will be examined is specified by {@link #numLevels()}.
     * @return the current {@link Path} to search
     */
    public Path root();

    /**
     * The amount of recursion to apply when searching:
     * <ul>
     *  <li><b>If zero (0):</b> only a single file will be searched,
     *  and {@link #root()} must be a file.</li>
     *  <li><b>If one (1):</b> only the current directory will be searched,
     *  and {@link #root()} must be a folder.</li>
     *  <li><b>If two (2):</b> only the current directory and its immediate subfolders
     *  will be searched.</li>
     *  <li>...</li>
     *  <li><b>If {@link Integer#MAX_VALUE}:</b> the current directory
     *  and all subfolders will be searched.</li>
     * </ul>
     * @return the integer number of levels of recursion to apply when searching
     */
    public int numLevels();

    /**
     * Clears the current search parameters to perform a fully recursive search
     *  with no other parameters, starting at the given root.
     * @param root the {@link File} root at which to begin searching
     */
    public void clear(File root);

    /** Updates the search query to reflect current values. Performs no I/O operations. */
    public void update();

    /**
     * Loads a search hash from disk.
     * Typically used in combination with {@link #deserialize(DataInput)}
     *  to restore the previous search settings.
     * If {@code hash == null}, this object has not been saved to disk yet,
     *  and should reset itself to the default settings.
     * <p>
     * <i>The default implementation of {@link #deserialize(DataInput)} calls {@link #create()},
     *  followed by this method, to read this object from disk.
     * If this object cannot be represented solely by its search hash,
     *  the default deserialization method must be overridden.</i>
     * @param hash the String hash to be loaded into this object, or {@code null} if none
     */
    public void loadSearchHash(String hash);

    /**
     * Calculates a hash of the current search settings.
     * Typically used in combination with {@link #serialize(DataOutput, FileSearchControl)}
     *  to persist a search's settings to disk.
     * <p>
     * <b>It is essential that a unique string exists for every possible combination
     *  of search settings;</b> otherwise, it would be impossible to know the settings
     *  for a file loaded from disk.
     * A simple way to accomplish this is to separate each setting via {@link #HASH_DELIMITER}.
     * However, any algorithm for generating the hash can be used,
     *  provided it guarantees uniqueness.
     * <p>
     * <i>The default implementation of {@link #serialize(DataOutput, FileSearchControl)}
     *  calls this method and writes the result to disk.
     * If this object cannot be represented solely by its search hash,
     *  the default serialization method must be overridden.</i>
     * A {@link StringBuilder} should typically be used to improve performance
     *  when there are many search settings to be hashed.
     * @return a unique String representing the current search settings
     */
    public String getSearchHash();

    /**
     * Creates an instance of this class.
     * Should be called only by {@link #deserialize(DataInput)}.
     * @return the created {@link FileSearchControl}
     */
    public FileSearchControl create();

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public default void serialize(DataOutput output, FileSearchControl instance)
            throws IOException {
        output.writeUTF(instance.getSearchHash());
    }

    @Override
    public default FileSearchControl deserialize(DataInput input) throws IOException {
        FileSearchControl toReturn = create();
        toReturn.loadSearchHash(input.readUTF());
        return toReturn;
    }
}
