package edu.uwyo.crawl.database;

import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Set;

import edu.uwyo.crawl.ui.tree.FileDisplayTree;

/**
 * A class containing the current search settings and any associated results.
 * Can be serialized to disk as a "snapshot" of the current search results.
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public class Results implements Serializable, Serializer<Results> {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used for serialization; <b>do not modify!</b> */
    public static final Results INSTANCE = new Results();

    private static final String FOOTER = "FileCrawlerResultsSuffixTotalBytesFollows";
    private static final long serialVersionUID = 5365148610698773172L;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used for controlling the search parameters. Shared for all instances of this class. */
    protected static FileSearchControl searchControl;

    /**
     * Used for determining how to serialize this class.
     * Only relevant for {@link #INSTANCE}; other instances can be set, but will have no effect.
     */
    protected Serializer<FileSearchControl> searchControlSerializer;

    private FileDisplayTree resultTree;
    private HashMap<String, String> settings;
    private ExcludeList excludeList;
    private HashMap<String, Byte> flags;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link Results}. */
    protected Results() {
        resultTree = new FileDisplayTree(null);
        settings = new HashMap<String, String>();
        excludeList = new ExcludeList();
        flags = new HashMap<String, Byte>();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void serialize(DataOutput output, Results instance) throws IOException {
        INSTANCE.searchControlSerializer.serialize(output, searchControl);
        FileDisplayTree.INSTANCE.serialize(output, instance.resultTree);

        output.writeInt(instance.settings.size());

        Set<String> settingsKeySet = instance.settings.keySet();
        for (String setting : settingsKeySet) {
            output.writeUTF(setting);

            String value = instance.settings.get(setting);
            output.writeBoolean(value != null);
            if (value != null) {
                output.writeUTF(value);
            }
        }

        Set<String> flagsKeySet = instance.flags.keySet();
        output.writeInt(flagsKeySet.size());
        for (String path : flagsKeySet) {
            output.writeUTF(path);
            output.writeByte(instance.flags.get(path));
        }

        ExcludeList.INSTANCE.serialize(output, instance.excludeList);

        output.writeUTF(FOOTER);
        output.writeLong(output.bytesWritten());
    }

    @Override
    public Results deserialize(DataInput input) throws IOException {
        Results toReturn = new Results();

        searchControl = INSTANCE.searchControlSerializer.deserialize(input);
        toReturn.resultTree = FileDisplayTree.INSTANCE.deserialize(input);

        int numSettings = input.readInt();
        for (int i = 0; i < numSettings; i++) {
            String setting = input.readUTF();
            String value = null;
            if (input.readBoolean()) {
                value = input.readUTF();
            }
            toReturn.settings.put(setting, value);
        }

        int numFlags = input.readInt();
        for (int i = 0; i < numFlags; i++) {
            String path = input.readUTF();
            byte flag = input.readByte();
            toReturn.flags.put(path, flag);
        }

        toReturn.excludeList = ExcludeList.INSTANCE.deserialize(input);

        String footer = input.readUTF();
        if (!footer.equals(FOOTER)) {
            throw new IOException("expected file footer " + FOOTER + ", got " + footer);
        }

        long bytesRead = input.bytesRead();
        long expectedBytesRead = input.readLong();
        if (bytesRead != expectedBytesRead) {
            throw new IOException("expected " + expectedBytesRead + " bytes, read " + bytesRead);
        }

        return toReturn;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the {@link FileDisplayTree} containing all search results */
    public FileDisplayTree resultTree() {
        return resultTree;
    }

    /** @return the {@link FileSearchControl} containing all search settings */
    public FileSearchControl searchControl() {
        return searchControl;
    }

    /** @return the {@link ExcludeList} indicating which files/folders are excluded from search */
    public ExcludeList excludeList() {
        return excludeList;
    }

    /**
     * Obtains a user setting.
     * @param setting the String setting to obtain
     * @return the String value of the given setting, or {@code null} if it does not exist
     */
    public String readSetting(String setting) {
        return settings.get(setting);
    }

    /**
     * Adds or updates the given setting.
     * @param setting the String setting to add or update
     * @param value the String setting value
     */
    public void writeSetting(String setting, String value) {
        settings.put(setting, value);
    }

    /**
     * Obtains the flag associated with the given file path.
     * @param path the String file path whose flag should be obtained
     * @return the {@link Byte} flag associated with {@code path},
     *  or {@code null} if no flag has been set
     */
    public Byte getFlag(String path) {
        return flags.get(path);
    }

    /**
     * Associates a flag with the given file path.
     * If {@code path} already has a flag, that flag will be overwritten.
     * @param path the String file path to be given a flag
     * @param flag the byte flag to set
     */
    public void setFlag(String path, byte flag) {
        flags.put(path, flag);
    }

    /**
     * Removes the current flag associated with {@code path}, if any.
     * @param path the String file path whose flag should be removed
     */
    public void removeFlag(String path) {
        flags.remove(path);
    }

    /**
     * Removes all flags for all file paths.
     * After this method is called, {@link #getFlag(String)} will return {@code null}
     *  for <i>any and every</i> value of {@code path}.
     */
    public void removeAllFlags() {
        flags.clear();
    }
}
