package edu.uwyo.crawl.ui;

import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;

import edu.uwyo.Main;
import edu.uwyo.crawl.RestorableView;
import edu.uwyo.crawl.ui.frameAbout.FrameAbout;
import edu.uwyo.crawl.ui.frameClear.FrameClear;
import edu.uwyo.crawl.ui.frameDeveloper.FrameDeveloper;
import edu.uwyo.crawl.ui.frameExclude.FrameExclude;
import edu.uwyo.crawl.ui.frameFlags.FrameFlags;
import edu.uwyo.crawl.ui.frameJoin.FrameJoin;
import edu.uwyo.crawl.ui.frameRegex.FrameRegex;
import edu.uwyo.crawl.ui.panels.PanelMain;
import net.javamon.dev.AbstractFrame;
import net.javamon.dev.Ext;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.RelativeFileChooser;
import net.javamon.dev.utils.layout.Validate;

/**
 * The main application frame, which allows the user to perform a file crawl.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public class FrameMain implements WindowListener, RestorableView {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    protected LoadListener loadListener;
    protected SaveListener saveListener;
    protected ClearListener clearListener;
    protected JoinListener joinListener;
    protected CloseListener closeListener;

    protected ToggleListener toggleListener;
    protected ExcludeListener excludeListener;
    protected DeveloperModeListener developerModeListener;
    protected DeveloperListener developerListener;
    protected RegexListener regexListener;
    protected AboutListener aboutListener;

    protected JMenuItem menuClear;
    protected JMenuItem menuFlags;
    protected JMenuItem menuExclude;
    protected JMenuItem menuDeveloperMode;
    protected JMenuItem menuDeveloper;

    protected FrameClear clear;
    protected FrameJoin join;
    protected FrameExclude exclude;
    protected FrameRegex regex;
    protected FrameAbout about;
    protected FrameFlags flags;
    protected FrameDeveloper developer;

    private JMenuItem database;
    private boolean unsaved;

    private RelativeFileChooser fileChooser;

    private PanelMain panelMain;
    private JFrame frame;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link FrameMain}. */
    public FrameMain() {
        frame = new JFrame(Resources.__("mainTitleNoHTML"));
        frame.setMinimumSize(new Dimension(620, 699));
        frame.setLayout(new GridLayout(0, 1));
        frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        frame.addWindowListener(this);

        setMenuListeners();
        setMenuItems();
        fileChooser = RelativeFileChooser.build("mainChooseSnapshotH", Ext.FIND, false);
        panelMain = new PanelMain();
        frame.add(panelMain.panel);

        frame.pack();
        frame.setLocationRelativeTo(null);

        clear = new FrameClear(this);
        join = new FrameJoin(this);
        exclude = new FrameExclude(this);
        regex = new FrameRegex(this);
        about = new FrameAbout(this);
        flags = new FrameFlags(this);
        developer = new FrameDeveloper(this);

        Validate.disable();
        restoreView();
        Validate.enable();
        panelMain.validate();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Does nothing by default. */
    @Override
    public void windowIconified(WindowEvent e) {}

    /** Does nothing by default. */
    @Override
    public void windowDeiconified(WindowEvent e) {}

    /** Does nothing by default. */
    @Override
    public void windowActivated(WindowEvent e) {}

    /** Does nothing by default. */
    @Override
    public void windowDeactivated(WindowEvent e) {}

    @Override
    public void windowClosing(WindowEvent e) {
        File snapshot = null;
        if (unsaved) {
            while (true) {
                int choice = Prompt.show(Resources.__("mainResultsSaveH"),
                        Resources.__("mainResultsSaveMsg"));

                if (choice == JOptionPane.YES_OPTION) {
                    if (Main.currSnapshot() != null) {
                        fileChooser.setCurrentDirectory(Main.currSnapshot());
                    }
                    choice = fileChooser.showSaveDialog(frame);
                    if (choice == JFileChooser.APPROVE_OPTION) {
                        snapshot = fileChooser.getSelectedFile();
                        String path = snapshot.getAbsolutePath();
                        if (!path.endsWith(Ext.FIND)) {
                            snapshot = new File(path + Ext.FIND);
                        }

                        if (snapshot.exists()) {
                            choice = Prompt.show(Resources.__("mainResultsSaveConfirmH"),
                                    Resources.__("mainResultsSaveConfirmMsg", snapshot.getAbsolutePath()));
                            if (choice == JOptionPane.YES_OPTION) {
                                break;
                            } else {
                                snapshot = null;
                            }
                        } else {
                            break;
                        }
                    }
                } else {
                    break;
                }
            }
            save();
        }
        Main.quit(snapshot);
    }

    /** Does nothing by default. */
    @Override
    public void windowOpened(WindowEvent e) {}

    /** Does nothing by default. */
    @Override
    public void windowClosed(WindowEvent e) {}

    @Override
    public void restoreView() {
        String developerMode = Resources.readUserSetting("developerMode");
        if (developerMode != null) {
            boolean enable = Boolean.parseBoolean(developerMode);
            ((JCheckBoxMenuItem) menuDeveloperMode).setSelected(enable);
            Main.database().setDeveloperMode(enable);
            menuDeveloper.setVisible(enable);
        }

        panelMain.restoreView();
        flags.restoreView();
        exclude.restoreView();
        developer.restoreView();
    }

    @Override
    public void saveView() {
        panelMain.saveView();
        flags.saveView();
        exclude.saveView();
        developer.saveView();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Shows this frame and associates it with all future {@link Prompt} messages.
     * @param currSnapshot the {@link File} current result snapshot
     */
    public void show(File currSnapshot) {
        frame.setVisible(true);
        updateView(currSnapshot);
        Prompt.setMessageComponent(frame);
    }

    /** Hides this frame and all its children. */
    public void hide() {
        frame.setVisible(false);
        clear.hide();
        join.hide();
        exclude.hide();
        regex.hide();
        about.hide();
        flags.hide();
        developer.hide();
    }

    /** @return this frame's {@link PanelMain} */
    public PanelMain main() {
        return panelMain;
    }

    /** @return this frame's {@link FrameFlags} */
    public FrameFlags flags() {
        return flags;
    }

    /** @return this frame's {@link FrameExclude} */
    public FrameExclude exclude() {
        return exclude;
    }

    /** @return this frame's {@link FrameDeveloper} */
    public FrameDeveloper developer() {
        return developer;
    }

    /** Validates this frame and all its children. */
    public void validate() {
        clear.main().validate();
        join.main().validate();
        exclude.main().validate();
        regex.main().validate();
        about.main().validate();
        flags.main().validate();
        developer.main().validate();
        panelMain.validate();
    }

    /**
     * Updates the current view of this and all other frames,
     *  including both search results and settings.
     * Reloads search settings, etc. from disk and automatically calls {@link #validate()}.
     * @param currSnapshot the {@link File} current result snapshot,
     *  or {@code null} if there is no associated snapshot
     */
    public void updateView(File currSnapshot) {
        Validate.disable();

        restoreView();
        if (currSnapshot != null) {
            try {
                database.setText(
                        Resources.__("mainMenuDatabase", currSnapshot.getCanonicalPath()));
            } catch (Exception e) {}
        } else {
            database.setText("");
        }

        Validate.enable();
        validate();
    }

    /**
     * Indicates that this application's data has been modified in some way.
     * Will also change the window's title.
     */
    public void modify() {
        unsaved = true;
        frame.setTitle(Resources.__("mainTitleNoHTMLModified"));
    }

    /**
     * Indicates that all unsaved changes introduced by {@link #modify()} have been saved.
     * Will also change the window's title.
     */
    public void save() {
        unsaved = false;
        frame.setTitle(Resources.__("mainTitleNoHTML"));
    }

    /** Sets the menu listeners for this frame. */
    private void setMenuListeners() {
        loadListener = new LoadListener(this);
        saveListener = new SaveListener(this);
        clearListener = new ClearListener(this);
        joinListener = new JoinListener(this);
        closeListener = new CloseListener(frame);

        toggleListener = new ToggleListener(this);
        excludeListener = new ExcludeListener(this);
        developerModeListener = new DeveloperModeListener(this);
        developerListener = new DeveloperListener(this);
        regexListener = new RegexListener(this);
        aboutListener = new AboutListener(this);
    }

    /** Sets the menu for this frame. */
    private void setMenuItems() {
        JMenuBar menuBar = new JMenuBar();

        JMenu file = AbstractFrame.createJMenu("mainMenuFile", null, true);
        AbstractFrame.setJMenuItem(file,
                "mainMenuFileLoad", "mainMenuFileLoadTT",
                loadListener, true, false,
                KeyEvent.VK_O, false, false, true);
        AbstractFrame.setJMenuItem(file,
                "mainMenuFileSave", "mainMenuFileSaveTT",
                saveListener, true, false,
                KeyEvent.VK_S, false, false, true);
        menuClear = AbstractFrame.setJMenuItem(file,
                "mainMenuFileClear", "mainMenuFileClearTT",
                clearListener, true, false,
                KeyEvent.VK_F5, false, false, false);
        AbstractFrame.setJMenuItem(file,
                "mainMenuFileJoin", "mainMenuFileJoinTT",
                joinListener, true, false,
                KeyEvent.VK_J, false, false, true);
        AbstractFrame.setJMenuItem(file,
                "mainMenuFileClose", "mainMenuFileCloseTT",
                closeListener, true, false,
                KeyEvent.VK_Q, false, false, true);
        menuBar.add(file);

        JMenu options = AbstractFrame.createJMenu("mainMenuOptions", null, true);
        menuFlags = AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsToggle", "mainMenuOptionsToggleTT",
                toggleListener, true, false,
                KeyEvent.VK_T, false, false, true);
        menuExclude = AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsExclude", "mainMenuOptionsExcludeTT",
                excludeListener, true, false,
                KeyEvent.VK_E, false, false, true);
        menuDeveloperMode = AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsDeveloper", "mainMenuOptionsDeveloperTT",
                developerModeListener, true, true,
                KeyEvent.VK_F12, false, false, false);
        menuDeveloper = AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsDeveloperOpt", "mainMenuOptionsDeveloperOptTT",
                developerListener, true, false,
                KeyEvent.VK_F8, false, false, false);
        AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsRegex", "mainMenuOptionsRegexTT",
                regexListener, true, false,
                KeyEvent.VK_F3, false, false, false);
        AbstractFrame.setJMenuItem(options,
                "mainMenuOptionsAbout", "mainMenuOptionsAboutTT",
                aboutListener, true, false,
                KeyEvent.VK_F1, false, false, false);
        menuBar.add(options);

        database = AbstractFrame.setJMenuItem(menuBar, "mainMenuDatabaseSlug", null,
                null, false, false, 0, false, false, false);

        frame.setJMenuBar(menuBar);
    }
}
