package edu.uwyo.crawl.ui.frameDeveloper.panels;

import javax.swing.JButton;
import javax.swing.JTextField;

import edu.uwyo.Main;
import edu.uwyo.crawl.RestorableView;
import edu.uwyo.crawl.ui.FrameMain;
import edu.uwyo.crawl.ui.panels.PanelOptions;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.JLabelHTML;
import net.javamon.dev.utils.layout.AbstractPanel;
import net.javamon.dev.utils.layout.Validate;
import net.javamon.dev.utils.layout.ValidateFocusListener;
import net.javamon.dev.utils.layout.ValidateKeyListener;

/**
 * The main panel for {@link edu.uwyo.crawl.ui.frameDeveloper.FrameDeveloper}.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public class PanelMain extends AbstractPanel implements RestorableView {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public static final int DEFAULT_BUFFER_LOCAL = 16384;
    public static final int DEFAULT_BUFFER_NET = 16777216;
    public static final int DEFAULT_CHUNK_SIZE = 512;
    public static final int DEFAULT_HEADER_SIZE = 1024;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public JLabelHTML instructions;
    public JTextField bufferLocal;
    public JTextField bufferNet;
    public JTextField chunkSize;
    public JTextField headerSize;
    public JButton changeRoot;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link PanelMain}. */
    public PanelMain(FrameMain parent) {
        super();
        setInstructions();

        layoutColumn.start(5, 5, 5, panel, instructions, panel);
        setBufferLocal();
        setBufferNet();
        setChunkSize();
        setHeaderSize();
        layoutColumn.end();

        setChangeRoot(parent);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void setEnabled(boolean enabled) {
        bufferLocal.setEnabled(enabled);
        bufferNet.setEnabled(enabled);
        chunkSize.setEnabled(enabled);
        headerSize.setEnabled(enabled);
        changeRoot.setEnabled(enabled);
    }

    @Override
    public void validate() {
        String developerMode = Resources.readUserSetting("developerMode");
        if (developerMode == null && !Boolean.parseBoolean(developerMode)) { return; }

        Validate.textConstrainPOT(bufferLocal);
        Validate.textConstrainPOT(bufferNet);
        Validate.textConstrainInt(chunkSize, 1, false);
        Validate.textConstrainInt(headerSize, 1, false);
        saveView();

        final int searchDepth = Main.frameMain.main().options.recurse.getSelectedIndex();
        changeRoot.setEnabled(searchDepth != PanelOptions.RECURSE_FILE
                && Main.frameMain.main().list.tree.getTree() != null);
    }

    @Override
    public void restoreView() {
        String lastBufferLocal = Resources.readUserSetting("lastBufferLocal");
        String lastBufferNet = Resources.readUserSetting("lastBufferNet");
        String lastChunkSize = Resources.readUserSetting("lastChunkSize");
        String lastHeaderSize = Resources.readUserSetting("lastHeaderSize");

        if (lastBufferLocal != null) {
            bufferLocal.setText(lastBufferLocal);
        } else {
            bufferLocal.setText(DEFAULT_BUFFER_LOCAL + "");
        }

        if (lastBufferNet != null) {
            bufferNet.setText(lastBufferNet);
        } else {
            bufferNet.setText(DEFAULT_BUFFER_NET + "");
        }

        if (lastChunkSize != null) {
            chunkSize.setText(lastChunkSize);
        } else {
            chunkSize.setText(DEFAULT_CHUNK_SIZE + "");
        }

        if (lastHeaderSize != null) {
            headerSize.setText(lastHeaderSize);
        } else {
            headerSize.setText(DEFAULT_HEADER_SIZE + "");
        }
    }

    @Override
    public void saveView() {
        Resources.writeUserSetting("lastBufferLocal", bufferLocal.getText());
        Resources.writeUserSetting("lastBufferNet", bufferNet.getText());
        Resources.writeUserSetting("lastChunkSize", chunkSize.getText());
        Resources.writeUserSetting("lastHeaderSize", headerSize.getText());
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the integer current local buffer size */
    public int getBufferLocal() {
        try {
            return Integer.parseInt(bufferLocal.getText());
        } catch (NumberFormatException e) {
            return DEFAULT_BUFFER_LOCAL;
        }
    }

    /** @return the integer current network buffer size */
    public int getBufferNet() {
        try {
            return Integer.parseInt(bufferNet.getText());
        } catch (NumberFormatException e) {
            return DEFAULT_BUFFER_NET;
        }
    }

    /** @return the integer current file chunk size */
    public int getChunkSize() {
        try {
            return Integer.parseInt(chunkSize.getText()) * 1000000;
        } catch (NumberFormatException e) {
            return DEFAULT_CHUNK_SIZE * 1000000;
        }
    }

    /** @return the integer current file header size */
    public int getHeaderSize() {
        try {
            return Integer.parseInt(headerSize.getText());
        } catch (NumberFormatException e) {
            return DEFAULT_HEADER_SIZE;
        }
    }

    /** Sets the instructions field. */
    private void setInstructions() {
        instructions = createLabel("mainDeveloperInstructions", null);
        layout(WEST, instructions, 5, WEST, panel);
        layout(NORTH, instructions, 5, NORTH, panel);
        layout(EAST, instructions, -5, EAST, panel);
        add(instructions);
    }

    /** Sets the local buffer field. */
    private void setBufferLocal() {
        JLabelHTML label = createLabel("mainDeveloperBufferLocal", "mainDeveloperBufferLocalTT");
        bufferLocal = createTextField(null, 7, "mainDeveloperBufferLocalTT");
        bufferLocal.addFocusListener(new ValidateFocusListener(this));
        bufferLocal.addKeyListener(new ValidateKeyListener(this));
        layoutColumn.place(label, bufferLocal);
    }

    /** Sets the network buffer field. */
    private void setBufferNet() {
        JLabelHTML label = createLabel("mainDeveloperBufferNet", "mainDeveloperBufferNetTT");
        bufferNet = createTextField(null, 7, "mainDeveloperBufferNetTT");
        bufferNet.addFocusListener(new ValidateFocusListener(this));
        bufferNet.addKeyListener(new ValidateKeyListener(this));
        layoutColumn.place(label, bufferNet);
    }

    /** Sets the chunk size field. */
    private void setChunkSize() {
        JLabelHTML label = createLabel("mainDeveloperChunkSize", "mainDeveloperChunkSizeTT");
        chunkSize = createTextField(null, 7, "mainDeveloperChunkSizeTT");
        chunkSize.addFocusListener(new ValidateFocusListener(this));
        chunkSize.addKeyListener(new ValidateKeyListener(this));
        layoutColumn.place(label, chunkSize);
    }

    /** Sets the header size field. */
    private void setHeaderSize() {
        JLabelHTML label = createLabel("mainDeveloperHeaderSize", "mainDeveloperHeaderSizeTT");
        headerSize = createTextField(null, 7, "mainDeveloperHeaderSizeTT");
        headerSize.addFocusListener(new ValidateFocusListener(this));
        headerSize.addKeyListener(new ValidateKeyListener(this));
        layoutColumn.place(label, headerSize);
    }

    /** Sets the change root button. */
    private void setChangeRoot(FrameMain parent) {
        changeRoot = createButton("mainDeveloperChangeRoot", "mainDeveloperChangeRootTT", -1, -1);
        changeRoot.addActionListener(new ChangeRootListener(parent));
        layout(WEST, changeRoot, 5, WEST, panel);
        layout(SOUTH, changeRoot, -5, SOUTH, panel);
        add(changeRoot);
    }
}
