package edu.uwyo.crawl.ui.frameJoin.panels;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.HashSet;

import edu.uwyo.Main;
import edu.uwyo.crawl.ui.panels.PanelFiles;
import net.javamon.dev.Console;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;
import net.javamon.dev.SetUtils;
import net.javamon.dev.StringUtils;

/**
 * A listener for performing the join operation.
 * @version 28 March, 2019
 * @author Christopher Stanton
 */
public class JoinListener implements ActionListener {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final int MODE_UNION = 0;
    private static final int MODE_INTERSECT = 1;
    private static final int MODE_NON_INTERSECT = 2;
    private static final int MODE_TOTAL_INTERSECT = 3;
    private static final int MODE_TOTAL_DIFFERENCE = 4;
    private static final int MODE_SYMMETRIC_DIFFERENCE = 5;
    private static final int MODE_ASYMMETRIC_DIFFERENCE = 6;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private final PanelMain parent;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link JoinListener}. */
    protected JoinListener(PanelMain parent) {
        this.parent = parent;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void actionPerformed(ActionEvent e) {
        final File originalSnapshot = Main.currSnapshot();
        final String[] files = parent.list.getEntries();
        final boolean[] flags = parent.flags.getFlags();
        final int masterIndex = parent.list.getMasterIndex();
        final File master = new File(files[masterIndex]);
        final int mode;
        if (parent.mode.union.isSelected()) {
            mode = MODE_UNION;
        } else if (parent.mode.intersect.isSelected()) {
            mode = MODE_INTERSECT;
        } else if (parent.mode.nonIntersect.isSelected()) {
            mode = MODE_NON_INTERSECT;
        } else if (parent.mode.totalIntersect.isSelected()) {
            mode = MODE_TOTAL_INTERSECT;
        } else if (parent.mode.totalDifference.isSelected()) {
            mode = MODE_TOTAL_DIFFERENCE;
        } else if (parent.mode.symmetricDifference.isSelected()) {
            mode = MODE_SYMMETRIC_DIFFERENCE;
        } else {
            mode = MODE_ASYMMETRIC_DIFFERENCE;
        }

        Main.frameMain.hide();
        Console.windowLoad.start(null, null);
        new Thread(new Runnable() {
            @Override
            public void run() {
                FileResults[] results = new FileResults[files.length];
                boolean flatMode = false;

                for (int i = 0; i < results.length; i++) {
                    if (!Main.database().loadSearchResults(new File(files[i]))) {
                        error("Load", originalSnapshot, files[i]);
                        return;
                    }

                    if (i == 0) {
                        flatMode = Main.database().results().resultTree().isFlat();
                    } else if (Main.database().results().resultTree().isFlat() != flatMode) {
                        error("Flat", originalSnapshot, files[i]);
                        return;
                    }

                    String root = Main.database().results().resultTree().getFilePath();
                    String[] prefixes = Main.database().results().resultTree().getPrefixes();
                    String[] paths = Main.database().results().resultTree().getPaths();
                    long[] sizes = Main.database().results().resultTree().getPathSizes();
                    int[] flags = Main.database().results().resultTree()
                            .getFlags(PanelFiles.NODE_FACTORY);
                    results[i] = new FileResults(root, prefixes, paths, sizes, flags);
                }

                if (!flatMode && !consolidateRoot(results)) {
                    error("Root", originalSnapshot, null);
                    return;
                }

                Main.database().loadSearchResults(master);
                Main.database().results().resultTree().clear(results[masterIndex].root);
                createFileTree(results, masterIndex, flags, mode);

                Console.windowLoad.stop();
                Prompt.message(Resources.__("mainJoinCompleteH"),
                        Resources.__("mainJoinCompleteMsg"));
                Main.frameMain.modify();
                Main.frameMain.show(null);
            }
        }).start();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Displays an error message to the user indicating that the join operation failed,
     *  then restores the original search results and displays the main window.
     * @param error the String error message, as a key fragment to be obtained
     *  from the localized resource bundle; this method expects to find keys
     *  {@code mainJoinError<>H} and {@code mainJoinError<>Msg}, where {@code <>} is {@code error}
     * @param originalSnapshot the {@link File} representing the previously loaded snapshot
     * @param path the String path that was being processed when the error occurred,
     *  or {@code null} if none
     */
    private static void error(String error, File originalSnapshot, String path) {
        String header = Resources.__("mainJoinError" + error + "H");
        String message;
        if (path != null) {
            message = Resources.__("mainJoinError" + error + "Msg", path);
        } else {
            message = Resources.__("mainJoinError" + error + "Msg");
        }

        if (originalSnapshot != null) {
            Main.database().loadSearchResults(originalSnapshot);
        }
        Console.windowLoad.stop();
        Prompt.show(header, message);
        Main.frameMain.show(originalSnapshot);
    }

    /**
     * Consolidates the roots of all search results by finding the shortest common file root.
     * If there is no common root, this method returns {@code false}, indicating that the results
     *  cannot be joined.
     * Otherwise, the root of each {@link FileResults} instance will be changed to match
     *  this longest common root.
     * <p>
     * <i>Do not call this method on "flat" trees.</i>
     * @param results the array of {@link FileResults} to be consolidated
     * @return a boolean, {@code true} if roots were consolidated and {@code false} if there was
     *  no common file root
     */
    private static boolean consolidateRoot(FileResults[] results) {
        String[] roots = new String[results.length];
        for (int i = 0; i < roots.length; i++) {
            roots[i] = results[i].root;
        }

        String commonRoot;
        try {
            commonRoot = StringUtils.longestCommonPrefix(roots);
            if (commonRoot.length() == 0) {
                return false;
            }
        } catch (NullPointerException e) {
            return false;
        }

        boolean sameRoots = true;
        for (int i = 0; i < roots.length; i++) {
            if (roots[i].length() != commonRoot.length()) {
                sameRoots = false; //if roots are all the same, no trailing slash will exist
                break;
            }
        }

        if (!sameRoots) {
            while (!commonRoot.endsWith("/") && !commonRoot.endsWith("\\")) {
                commonRoot = commonRoot.substring(0, commonRoot.length() - 1);
                if (commonRoot.length() == 0) {
                    return false; //longest prefix might "over-match" past the last file separator
                }
            }
            commonRoot = commonRoot.substring(0, commonRoot.length() - 1); //remove trailing slash
        }

        for (int i = 0; i < roots.length; i++) {
            results[i].root = commonRoot;
        }
        return true;
    }

    /**
     * Creates a merged file tree from {@code results}.
     * <i>The target tree must have been cleared before calling this method.</i>
     * <p>
     * During the join operation, duplicate paths may be found.
     * In this case, the <i>master file's</i> entries will always take precedence over the entries
     *  in other search results.
     * @param results the array of {@link FileResults} to add
     * @param masterIndex the integer index of the <i>master file</i> in {@code results}
     * @param flags an array of boolean enabled and disabled flags
     * @param mode the integer mode used for joining {@code results}
     */
    @SuppressWarnings("unchecked")
    private static void createFileTree(FileResults[] results, int masterIndex, boolean[] flags,
            int mode) {
        HashSet<String>[] sets = new HashSet[results.length];
        for (int i = 0; i < sets.length; i++) {
            sets[i] = new HashSet<String>();
            for (int j = 0; j < results[i].paths.length; j++) {
                sets[i].add(results[i].paths[j]);
            }
        }

        HashSet<String> entries;
        switch (mode) {
            case MODE_UNION:                entries = SetUtils.union(sets); break;
            case MODE_INTERSECT:            entries = SetUtils.intersect(sets); break;
            case MODE_NON_INTERSECT:        entries = SetUtils.nonIntersect(sets); break;
            case MODE_TOTAL_INTERSECT:      entries = SetUtils.totalIntersect(sets); break;
            case MODE_TOTAL_DIFFERENCE:     entries = SetUtils.totalDifference(sets); break;
            case MODE_SYMMETRIC_DIFFERENCE: entries = SetUtils.symmetricDifference(sets); break;
            default:                        entries = SetUtils.asymmetricDifference(sets); break;
        }

        FileResults toAdd = results[masterIndex];
        for (int i = 0; i < toAdd.paths.length; i++) {
            addToFileTree(toAdd, i, entries, flags); //ensure master file overrides any duplicates
        }

        for (int i = 0; i < results.length; i++) {
            if (i == masterIndex) { continue; }
            toAdd = results[i];

            for (int j = 0; j < toAdd.paths.length; j++) {
                addToFileTree(toAdd, j, entries, flags);
            }
        }
    }

    /**
     * Adds a single entry within a search result set to the file tree.
     * Entries will only be added if they are contained in the {@code entries} set.
     * If the entry at {@code index} was added, it will be removed from {@code entries}
     *  to prevent it from being added a second time.
     * @param result the {@link FileResults} containing entries to be added
     * @param index the integer index within {@code result} of the path to be added
     * @param entries a {@link HashSet} of String file paths; only paths within this set
     *  can be added to the file tree
     * @param flags an array of boolean enabled and disabled flags
     */
    private static void addToFileTree(FileResults result, int index, HashSet<String> entries,
            boolean[] flags) {
        if (entries.contains(result.paths[index])) {
            if (!flags[Math.max(0, result.flags[index])]) { return; } //flag is disabled

            Main.database().results().resultTree().add(result.prefixes[index], result.paths[index],
                    false, result.sizes[index]);

            if (result.flags[index] == -1) {
                Main.database().results().removeFlag(result.paths[index]);
            } else {
                Main.database().results().setFlag(result.paths[index], (byte) result.flags[index]);
            }
            entries.remove(result.paths[index]);
        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CLASSES                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** A helper class representing all file paths, etc. within a search results file. */
    private static class FileResults {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* FIELDS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        private String root;
        private final String[] prefixes;
        private final String[] paths;
        private final long[] sizes;
        private final int[] flags;

        ///////////////////////////////////////////////////////////////////////////////////////////
        /* CONSTRUCTORS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        /**
         * Creates a new {@link FileResults}.
         * @param root the String root path
         * @param prefixes the array of String prefixes belonging to the search results file
         * @param paths the array of String file paths belonging to the search results file
         * @param sizes the array of long file sizes belonging to the search results file
         * @param flags the array of integer flags belonging to the search results file
         */
        private FileResults(String root, String[] prefixes, String[] paths, long[] sizes,
                int[] flags) {
            this.root = root;
            this.prefixes = prefixes;
            this.paths = paths;
            this.sizes = sizes;
            this.flags = flags;
        }
    }
}
