package edu.uwyo.crawl.ui.panels;

import java.awt.Color;
import java.awt.Event;
import java.awt.event.KeyEvent;

import javax.swing.KeyStroke;

import edu.uwyo.Main;
import edu.uwyo.crawl.RestorableView;
import edu.uwyo.crawl.ui.popup.ColorHashFileNode;
import edu.uwyo.crawl.ui.popup.FilePopupMenu;
import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import edu.uwyo.crawl.ui.tree.PanelFileTree;
import net.javamon.dev.utils.layout.AbstractPanel;

/**
 * The main search file results panel.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public class PanelFiles extends AbstractPanel implements RestorableView {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** The factory instance of {@link ColorHashFileNode}. */
    public static final ColorHashFileNode NODE_FACTORY = new ColorHashFileNode(null, null);

    /** Indicates a normal file without any manual status flags. */
    public static final byte FLAG_OKAY = 0;

    /** Indicates a file that to be ignored. */
    public static final byte FLAG_IGNORE = 1;

    /** Indicates a file that has been fixed. */
    public static final byte FLAG_FIXED = 2;

    /** Indicates a file that needs to be reviewed. */
    public static final byte FLAG_REVIEW = 3;

    /** Indicates a file with an attached warning. */
    public static final byte FLAG_WARNING = 4;

    /** Indicates a file with an attached error. */
    public static final byte FLAG_ERROR = 5;

    /** Contains the display colors of all flags defined in this class. */
    private static final Color[] COLORS = {
            new Color(255, 255, 255),
            new Color(200, 200, 200),
            new Color(150, 150, 255),
            new Color(255, 150, 255),
            new Color(255, 150, 85),
            new Color(255, 100, 100)
    };

    /** Contains the shortcut keys for all flags defined in this class. */
    private static final KeyStroke[] ACCELERATORS = {
            null,
            KeyStroke.getKeyStroke(KeyEvent.VK_I, Event.ALT_MASK),
            KeyStroke.getKeyStroke(KeyEvent.VK_F, Event.ALT_MASK),
            KeyStroke.getKeyStroke(KeyEvent.VK_R, Event.ALT_MASK),
            KeyStroke.getKeyStroke(KeyEvent.VK_W, Event.ALT_MASK),
            KeyStroke.getKeyStroke(KeyEvent.VK_E, Event.ALT_MASK)
    };

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public PanelFileTree tree;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link PanelFiles}. */
    public PanelFiles() {
        super("mainPanelFilesBdr");
        setTree();
        COLORS[FLAG_OKAY] = tree.getTreeBackground();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void setEnabled(boolean enabled) {
        tree.setEnabled(enabled);
    }

    @Override
    public void validate() {
        tree.validate();
    }

    @Override
    public void restoreView() {
        updateView(Main.database().results().resultTree());
    }

    @Override
    public void saveView() {}

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Obtains the index in {@link #COLORS} of the specified color,
     *  analogous to its associated flag.
     * @param color the {@link Color} to be found, or {@code null}
     * @return the integer index within {@link #COLORS} where {@code color} was found,
     *  or -1 if the color was not found
     */
    public static int indexOf(Color color) {
        if (color != null) {
            for (int i = 0; i < COLORS.length; i++) {
                if (COLORS[i].equals(color)) {
                    return i;
                }
            }
        }
        return -1;
    }

    /**
     * Obtains the display color for a flag from this class.
     * @return the {@link Color} for the given flag, or {@code null} if that flag does not exist
     */
    public static Color getFlagColor(int flag) {
        if (flag < 0 || flag >= COLORS.length) { return null; }
        return COLORS[flag];
    }

    /**
     * Obtains the keystroke accelerator for a flag from this class.
     * @return the {@link KeyStroke} for the given flag,
     *  or {@code null} if that flag does not exist
     */
    public static KeyStroke getKeyStroke(int flag) {
        if (flag < 0 || flag >= ACCELERATORS.length) { return null; }
        return ACCELERATORS[flag];
    }

    /** @return the integer total number of possible status flags */
    public static int getNumFlags() {
        return COLORS.length;
    }

    /**
     * Updates the view associated with this panel.
     * Call this method whenever the list of files in this tree should change.
     * @param resultTree the {@link FileDisplayTree} whose contents should be displayed
     *  in this panel, or {@code null} if this panel's tree should be completely hidden
     */
    public void updateView(FileDisplayTree resultTree) {
        tree.updateView(resultTree);
    }

    /** Sets the tree panel. */
    private void setTree() {
        tree = new PanelFileTree(NODE_FACTORY, null, new FilePopupMenu(), false, true, null,
                300, 400);
        layoutFill(tree.panel, 5, panel, panel, panel, panel);
        add(tree.panel);
    }
}
