package edu.uwyo.crawl.ui.panels;

import java.awt.Color;
import java.awt.Dimension;
import java.io.File;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JTextField;

import edu.uwyo.Main;
import edu.uwyo.crawl.RestorableView;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.HTML;
import net.javamon.dev.utils.JLabelHTML;
import net.javamon.dev.utils.RelativeFileChooser;
import net.javamon.dev.utils.comboBox.ExtendableJComboBox;
import net.javamon.dev.utils.layout.AbstractPanel;
import net.javamon.dev.utils.layout.ValidateFocusListener;
import net.javamon.dev.utils.layout.ValidateKeyListener;
import net.javamon.dev.utils.layout.ValidateListener;

/**
 * The main search options panel.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public class PanelOptions extends AbstractPanel implements RestorableView {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public static final int RECURSE_FILE = 0;
    public static final int RECURSE_INFINITE = 7;
    public static final int RECURSE_RESULTS = 8;

    public static final int NO_DUPLICATE = 0;
    public static final int DUPLICATE_NAME = 1;
    public static final int DUPLICATE_CONTENT = 2;
    public static final int DUPLICATE_ALL = 3;

    private static final Color BAD_PATTERN = new Color(255, 200, 200);
    private static final Color GOOD_PATTERN = new JTextField("").getBackground();

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public JTextField folder;
    public JButton chooseFolder;
    public ExtendableJComboBox<String> recurse;
    public JTextField patternName;
    public JTextField patternContent;
    public JButton process;
    public ExtendableJComboBox<String> duplicate;
    public JCheckBox duplicateExact;

    private JLabelHTML badPattern;
    private RelativeFileChooser fileChooser;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link PanelOptions}. */
    public PanelOptions() {
        super("mainPanelOptionsBdr", 100, 175);

        setFolder();
        setRecurse();
        setPatternName();
        setPatternContent();
        setProcess();
        setDuplicate();
        setDuplicateExact();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void setEnabled(boolean enabled) {
        folder.setEnabled(enabled);
        chooseFolder.setEnabled(enabled);
        recurse.setEnabled(enabled);
        patternName.setEnabled(enabled);
        patternContent.setEnabled(enabled);
        process.setEnabled(enabled);
        duplicate.setEnabled(enabled);
        duplicateExact.setEnabled(enabled);
    }

    @Override
    public void validate() {
        validatePattern(patternName, badPattern);
        if (badPattern.isVisible()) {
            validatePattern(patternContent, null);
        } else {
            validatePattern(patternContent, badPattern);
        }

        int nesting = recurse.getSelectedIndex();
        if (nesting == RECURSE_RESULTS && Main.frameMain != null
                && Main.frameMain.main().list.tree.getTree() == null) {
            recurse.setSelectedIndex(0);
            nesting = 0;
        }

        boolean isFile = new File(folder.getText()).isFile();
        boolean correctFile = (nesting == RECURSE_FILE && isFile)
                || (nesting != RECURSE_FILE && !isFile);
        if (!correctFile) {
            folder.setText("");
        }

        int mode = duplicate.getSelectedIndex();
        duplicateExact.setEnabled(mode == DUPLICATE_CONTENT || mode == DUPLICATE_ALL);
        if (mode != DUPLICATE_CONTENT && mode != DUPLICATE_ALL) {
            duplicateExact.setSelected(false);
        }

        process.setEnabled(!badPattern.isVisible()
                && folder.getText().length() > 0
                && correctFile);
    }

    @Override
    public void restoreView() {
        String lastFolder = Main.readSetting("lastFolder");
        String lastPatternName = Main.readSetting("lastPatternName");
        String lastPatternContent = Main.readSetting("lastPatternContent");
        String lastRecurse = Main.readSetting("lastRecurse");
        String lastDuplicate = Main.readSetting("lastDuplicate");
        String lastDuplicateExact = Main.readSetting("lastDuplicateExact");

        if (lastFolder != null) {
            folder.setText(lastFolder);
            folder.getCaret().setDot(folder.getText().length());
        }
        if (lastPatternName != null) {
            patternName.setText(lastPatternName);
        }
        if (lastPatternContent != null) {
            patternContent.setText(lastPatternContent);
        }
        if (lastRecurse != null) {
            recurse.setSelectedIndex(Integer.parseInt(lastRecurse));
        }
        if (lastDuplicate != null) {
            duplicate.setSelectedIndex(Integer.parseInt(lastDuplicate));
        }
        if (lastDuplicateExact != null) {
            duplicateExact.setEnabled(Boolean.parseBoolean(lastDuplicateExact));
        }
    }

    @Override
    public void saveView() {
        Main.writeSetting("lastFolder", folder.getText());
        Main.writeSetting("lastPatternName", patternName.getText());
        Main.writeSetting("lastPatternContent", patternContent.getText());
        Main.writeSetting("lastRecurse", recurse.getSelectedIndex() + "");
        Main.writeSetting("lastDuplicate", duplicate.getSelectedIndex() + "");
        Main.writeSetting("lastDuplicateExact", duplicateExact.isSelected() ? "True" : "False");
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Validates a single regular expression pattern. */
    private void validatePattern(JTextField pattern, JLabelHTML badPattern) {
        pattern.setBackground(GOOD_PATTERN);
        if (badPattern != null) {
            badPattern.setVisible(false);
        }

        if (pattern.isEnabled() && pattern.isEditable()) {
            try {
                Pattern.compile(pattern.getText());
            } catch (PatternSyntaxException e) {
                pattern.setBackground(BAD_PATTERN);
                if (badPattern != null) {
                    badPattern.setVisible(true);
                }
            }
        } else {
            pattern.setText("");
        }
    }

    /** Sets the folder chooser. */
    private void setFolder() {
        JLabelHTML label = createLabel("mainPanelOptionsChoose", "mainPanelOptionsChooseTT");
        layout(WEST, label, 5, WEST, panel);
        layout(NORTH, label, 5, NORTH, panel);
        add(label);

        chooseFolder = createButton("mainPanelOptionsChooseButton", "mainPanelOptionsChooseTT",
                -1, -1);
        fileChooser = RelativeFileChooser.build("mainPanelOptionsChooseH", -1,
                JFileChooser.FILES_AND_DIRECTORIES, null, false);
        chooseFolder.addActionListener(new ChooseFolderListener(this, fileChooser));
        layout(EAST, chooseFolder, -5, EAST, panel);
        layout(NORTH, chooseFolder, 5, NORTH, panel);
        add(chooseFolder);

        folder = createTextField(null, 40, "mainPanelOptionsChooseTT");
        folder.setEditable(false);
        layout(WEST, folder, 5, EAST, label);
        layout(EAST, folder, -5, WEST, chooseFolder);
        layout(NORTH, folder, 5, NORTH, panel);
        add(folder);
    }

    /** Sets the recursive chooser. */
    private void setRecurse() {
        recurse = createComboBox("mainPanelOptionsRecurseTT",
                HTML.wrap(Resources.__("mainPanelOptionsRecurse0")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse1")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse2")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse3")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse4")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse5")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurse6")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurseAll")),
                HTML.wrap(Resources.__("mainPanelOptionsRecurseCurr")));
        recurse.addActionListener(new ValidateListener(this));
        layout(EAST, recurse, -5, EAST, panel);
        layout(NORTH, recurse, 5, SOUTH, chooseFolder);
        add(recurse);

        JLabelHTML label = createLabel("mainPanelOptionsRecurse", "mainPanelOptionsRecurseTT");
        layout(EAST, label, -5, WEST, recurse);
        layout(NORTH, label, 5, SOUTH, chooseFolder);
        add(label);
    }

    /** Sets the name pattern field. */
    private void setPatternName() {
        JLabelHTML label = createLabel("mainPanelOptionsPatternName",
                "mainPanelOptionsPatternNameTT");
        layout(WEST, label, 5, WEST, panel);
        layout(NORTH, label, 5, SOUTH, recurse);
        add(label);

        patternName = createTextField(null, 50, "mainPanelOptionsPatternNameTT");
        patternName.addFocusListener(new ValidateFocusListener(this));
        patternName.addKeyListener(new ValidateKeyListener(this));
        layout(WEST, patternName, 5, EAST, label);
        layout(EAST, patternName, -5, EAST, panel);
        layout(NORTH, patternName, 5, SOUTH, recurse);
        add(patternName);

        badPattern = createLabel("mainPanelOptionsPatternBad", null);
        badPattern.setForeground(new Color(255, 0, 0));
        layout(WEST, badPattern, 5, WEST, panel);
        layout(SOUTH, badPattern, -5, NORTH, patternName);
        add(badPattern);
    }

    /** Sets the content pattern field. */
    private void setPatternContent() {
        JLabelHTML label = createLabel("mainPanelOptionsPatternContent",
                "mainPanelOptionsPatternContentTT");
        layout(WEST, label, 5, WEST, panel);
        layout(NORTH, label, 5, SOUTH, patternName);
        add(label);

        patternContent = createTextField(null, 50, "mainPanelOptionsPatternContentTT");
        patternContent.addFocusListener(new ValidateFocusListener(this));
        patternContent.addKeyListener(new ValidateKeyListener(this));
        layout(WEST, patternContent, 5, EAST, label);
        layout(EAST, patternContent, -5, EAST, panel);
        layout(NORTH, patternContent, 5, SOUTH, patternName);
        add(patternContent);
    }

    /** Sets the process button. */
    private void setProcess() {
        process = createButton("mainPanelOptionsProcess", "mainPanelOptionsProcessTT", -1, -1);
        process.setPreferredSize(new Dimension(150, process.getPreferredSize().height));
        process.addActionListener(new ProcessListener());
        layout(EAST, process, -5, EAST, panel);
        layout(SOUTH, process, -5, SOUTH, panel);
        add(process);
    }

    /** Sets the duplicate chooser. */
    private void setDuplicate() {
        JLabelHTML label = createLabel("mainPanelOptionsDuplicate", "mainPanelOptionsDuplicateTT");
        layout(WEST, label, 5, WEST, panel);
        add(label);

        duplicate = createComboBox("mainPanelOptionsDuplicateTT",
                Resources.__("mainPanelOptionsDuplicate0"),
                Resources.__("mainPanelOptionsDuplicate1"),
                Resources.__("mainPanelOptionsDuplicate2"),
                Resources.__("mainPanelOptionsDuplicate3"));
        duplicate.addActionListener(new ValidateListener(this));
        layout(WEST, duplicate, 5, EAST, label);
        layout(SOUTH, duplicate, -5, SOUTH, panel);
        add(duplicate);

        layout(NORTH, label, 0, NORTH, duplicate);
    }

    /** Sets the exact-duplicates chooser. */
    private void setDuplicateExact() {
        JLabelHTML label = createLabel("mainPanelOptionsDuplicateExact",
                "mainPanelOptionsDuplicateExactTT");
        layout(WEST, label, 5, EAST, duplicate);
        layout(NORTH, label, 0, NORTH, duplicate);
        add(label);

        duplicateExact = createCheckBox("mainPanelOptionsDuplicateExactTT", false);
        layout(WEST, duplicateExact, 5, EAST, label);
        layout(NORTH, duplicateExact, 0, NORTH, duplicate);
        add(duplicateExact);
    }
}
