package edu.uwyo.crawl.ui.panels;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import javax.swing.JCheckBox;
import javax.swing.JTextField;

import edu.uwyo.Main;
import edu.uwyo.crawl.RestorableView;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.JLabelHTML;
import net.javamon.dev.utils.comboBox.ExtendableJComboBox;
import net.javamon.dev.utils.layout.AbstractPanel;
import net.javamon.dev.utils.layout.ValidateFocusListener;
import net.javamon.dev.utils.layout.ValidateKeyListener;
import net.javamon.dev.utils.layout.ValidateListener;

/**
 * The panel representing options for a single time stamp.
 * @version 8 February, 2019
 * @author Christopher Stanton
 */
public class PanelTime extends AbstractPanel implements RestorableView {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public static final int OLDER = 0;
    public static final int NEWER = 1;
    public static final int BETWEEN = 2;

    public static final int CREATED = 0;
    public static final int MODIFIED = 1;
    public static final int ACCESSED = 2;
    public static final int NUM_MODES = 3;

    private static final String DELIMITERS = "[/ :\\-.]+";
    private static final String DATE_FORMAT_STRING = "MM/dd/yyyy HH:mm";
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat(DATE_FORMAT_STRING);
    private static final String DEFAULT_YEAR_PREFIX = "20";

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    public JCheckBox enabled;
    public ExtendableJComboBox<String> comparison;
    public JTextField dateMin;
    public JTextField dateMax;

    public final int mode;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link PanelTime}. */
    public PanelTime(PanelMain parent, int mode) {
        super("mainPanelOptionsTimeBdr" + mode, 230, 82);
        this.mode = mode;

        setEnabled(parent);
        setComparison(parent);
        setDateMin(parent);
        setDateMax(parent);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void setEnabled(boolean enabled) {
        this.enabled.setEnabled(enabled);
        comparison.setEnabled(enabled);
        dateMin.setEnabled(enabled);
        dateMax.setEnabled(enabled);
    }

    @Override
    public void validate() {
        if (!enabled.isEnabled()) {
            enabled.setSelected(false);
            comparison.setSelectedIndex(0);
            dateMin.setText("");
            dateMax.setText("");
            return;
        }

        final boolean isEnabled = enabled.isSelected();
        comparison.setEnabled(isEnabled);
        dateMin.setEnabled(isEnabled && comparison.getSelectedIndex() == BETWEEN
                && comparison.isEnabled());
        dateMax.setEnabled(isEnabled);
        if (!isEnabled) {
            dateMin.setText("");
            dateMax.setText("");
            return;
        }

        dateMin.setVisible(comparison.getSelectedIndex() == 2 && comparison.isEnabled());
        validateDate(dateMax);
        if (dateMin.isVisible()) {
            validateDate(dateMin);

            try {
                Date d1 = DATE_FORMAT.parse(dateMin.getText());
                Date d2 = DATE_FORMAT.parse(dateMax.getText());
                if (d2.getTime() <= d1.getTime()) {
                    Calendar c = Calendar.getInstance();
                    c.setTime(d1);
                    c.add(Calendar.MINUTE, 1);
                    dateMax.setText(DATE_FORMAT.format(c.getTime()));
                }
            } catch (ParseException e) {}
        }
    }

    @Override
    public void restoreView() {
        String time = settingPrefix();
        if (time == null) { return; }

        String lastEnabled = Main.readSetting(time + "Enabled");
        String lastComparison = Main.readSetting(time + "Comparison");
        String lastDateMin = Main.readSetting(time + "DateMin");
        String lastDateMax = Main.readSetting(time + "DateMax");

        if (lastEnabled != null) {
            enabled.setSelected(Boolean.parseBoolean(lastEnabled));
        }
        if (lastComparison != null) {
            comparison.setSelectedIndex(Integer.parseInt(lastComparison));
        }
        if (lastDateMin != null) {
            long dateTime = Long.parseLong(lastDateMin);
            if (dateTime > 0L) {
                dateMin.setText(DATE_FORMAT.format(new Date(dateTime)));
            } else {
                dateMin.setText("");
            }
        }
        if (lastDateMax != null) {
            long dateTime = Long.parseLong(lastDateMax);
            if (dateTime > 0L) {
                dateMax.setText(DATE_FORMAT.format(new Date(dateTime)));
            } else {
                dateMax.setText("");
            }
        }
    }

    @Override
    public void saveView() {
        String time = settingPrefix();
        if (time == null) { return; }

        Main.writeSetting(time + "Enabled", enabled.isSelected() ? "True" : "False");
        Main.writeSetting(time + "Comparison", comparison.getSelectedIndex() + "");
        Main.writeSetting(time + "DateMin", getTime(dateMin) + "");
        Main.writeSetting(time + "DateMax", getTime(dateMax) + "");
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Gets the current time associated with this panel.
     * @return the long time, in milliseconds, since the epoch
     *  (or 0 if no valid time has been specified yet)
     */
    public long getTime(JTextField date) {
        try {
            return DATE_FORMAT.parse(date.getText()).getTime();
        } catch (Exception e) {
            return 0L;
        }
    }

    /** @return the String prefix used for accessing a time settings */
    private String settingPrefix() {
        switch (mode) {
            case CREATED:   return "lastTimeCreated";
            case MODIFIED:  return "lastTimeModified";
            case ACCESSED:  return "lastTimeAccessed";
        }
        return null;
    }

    /** Validates a single date text field. */
    private void validateDate(JTextField date) {
        if (date.isEnabled() && date.getText().length() == 0) {
            date.setText("00/00/0000 00:00");
            return;
        }

        String rawText = date.getText();
        String[] leadingZeroTokens = rawText.split(DELIMITERS);
        if (leadingZeroTokens.length > 1) {
            StringBuilder replaceLeadingZeros = new StringBuilder();
            for (int i = 0; i < leadingZeroTokens.length; i++) {
                if (i == 2) {
                    if (leadingZeroTokens[i].length() == 2) {
                        replaceLeadingZeros.append(DEFAULT_YEAR_PREFIX);
                    }
                } else if (leadingZeroTokens[i].length() == 1) {
                    replaceLeadingZeros.append("0");
                }

                replaceLeadingZeros.append(leadingZeroTokens[i]);
            }
            rawText = replaceLeadingZeros.toString();
        }

        String toFormat = rawText.replaceAll(DELIMITERS, "");
        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < DATE_FORMAT_STRING.length(); i++) {
            if (builder.length() == 2 || builder.length() == 5) {
                builder.append("/");
            } else if (builder.length() == 10) {
                builder.append(" ");
            } else if (builder.length() == 13) {
                builder.append(":");
            }

            if (i < toFormat.length()) {
                builder.append(toFormat.charAt(i));
            } else {
                builder.append("0");
            }

            if (builder.length() == DATE_FORMAT_STRING.length()) {
                break;
            }
        }
        String dateText = builder.toString();

        try {
            DATE_FORMAT.parse(dateText);
            date.setText(dateText);
        } catch (ParseException e) {
            Date recoveryDate = new Date();
            recoveryDate.setTime(0L);
            date.setText(DATE_FORMAT.format(recoveryDate));
        }

        try {
            Date resolvedDate = new Date();
            resolvedDate.setTime(DATE_FORMAT.parse(date.getText()).getTime());
            date.setText(DATE_FORMAT.format(resolvedDate));
        } catch (Exception e) {}
    }

    /** Sets the enabled chooser. */
    private void setEnabled(PanelMain parent) {
        JLabelHTML label = createLabel("mainPanelOptionsTimeEnabled",
                "mainPanelOptionsTimeEnabledTT");
        layout(WEST, label, 5, WEST, panel);
        layout(NORTH, label, 5, NORTH, panel);
        add(label);

        enabled = createCheckBox("mainPanelOptionsTimeEnabledTT", false);
        enabled.addActionListener(new ValidateListener(parent));
        layout(WEST, enabled, 5, EAST, label);
        layout(NORTH, enabled, 5, NORTH, panel);
        add(enabled);
    }

    /** Sets the comparison chooser. */
    private void setComparison(PanelMain parent) {
        comparison = createComboBox("mainPanelOptionsTimeCompareTT",
                Resources.__("mainPanelOptionsTimeCompare0"),
                Resources.__("mainPanelOptionsTimeCompare1"),
                Resources.__("mainPanelOptionsTimeCompare2"));
        comparison.addActionListener(new ValidateListener(parent));
        layout(EAST, comparison, -5, EAST, panel);
        layout(NORTH, comparison, 5, NORTH, panel);
        add(comparison);
    }

    /** Sets the date min field. */
    private void setDateMin(PanelMain parent) {
        dateMin = createTextField(null, 9, "mainPanelOptionsTimeDateTT");
        dateMin.addFocusListener(new ValidateFocusListener(parent));
        dateMin.addKeyListener(new ValidateKeyListener(parent));
        dateMin.setVisible(false);
        layout(WEST, dateMin, 5, WEST, panel);
        layout(NORTH, dateMin, 5, SOUTH, comparison);
        add(dateMin);
    }

    /** Sets the date max field. */
    private void setDateMax(PanelMain parent) {
        dateMax = createTextField(null, 9, "mainPanelOptionsTimeDateTT");
        dateMax.addFocusListener(new ValidateFocusListener(parent));
        dateMax.addKeyListener(new ValidateKeyListener(parent));
        layout(EAST, dateMax, -5, EAST, panel);
        layout(NORTH, dateMax, 5, SOUTH, comparison);
        add(dateMax);
    }
}
