package edu.uwyo.crawl.ui.popup;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.ArrayList;

import javax.swing.JOptionPane;

import edu.uwyo.Main;
import edu.uwyo.crawl.ui.tree.FileDisplayTree;
import net.javamon.dev.Console;
import net.javamon.dev.Prompt;
import net.javamon.dev.Resources;
import net.javamon.dev.windowLoad.WindowLoad.WindowLoadInvokable;

/**
 * A listener for performing a duplicate search on selected files.
 * @version 19 April, 2019
 * @author Christopher Stanton
 */
public class DuplicateCheckListener implements ActionListener {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Performs duplicate matching only on file names. */
    protected static final int NAME = 0;

    /** Performs duplicate matching only on partial file contents. */
    protected static final int CONTENT = 1;

    /** Performs duplicate matching on file names and partial file contents. */
    protected static final int NAME_CONTENT = 2;

    /** Performs duplicate matching only on exact file contents. */
    protected static final int CONTENT_EXACT = 3;

    /** Performs duplicate matching on file names and exact file contents. */
    protected static final int NAME_CONTENT_EXACT = 4;

    protected static final int NUM_MODES = 5;

    private static final WindowLoadInvokable INVOKABLE = new WindowLoadInvokable() {
        @Override
        public String getText() {
            return Resources.__("mainNumSearched", Main.database().numSearched());
        }
    };

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private final FilePopupMenu menu;
    private final int mode;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link DuplicateCheckListener}.
     * @param menu the parent {@link FilePopupMenu}
     * @param mode the integer duplicate checking mode: {@link #NAME}, {@link #CONTENT},
     *  {@link #NAME_CONTENT}, {@link #CONTENT_EXACT}, {@link #NAME_CONTENT_EXACT}
     */
    protected DuplicateCheckListener(FilePopupMenu menu, int mode) {
        this.menu = menu;
        this.mode = mode;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void actionPerformed(ActionEvent e) {
        if (menu.isFlat()) {
            Prompt.error(Resources.__("mainPanelFileMenuDuplicateErrorH"),
                    Resources.__("mainPanelFileMenuDuplicateErrorMsg"));
            return;
        } else if (menu.currNodes().length == 1 && menu.currNodes()[0].getFilePath() != null) {
            Prompt.error(Resources.__("mainPanelFileMenuDuplicateErrorH"),
                    Resources.__("mainPanelFileMenuDuplicateErrorSingleMsg"));
            return;
        }

        int choice = Prompt.show(Resources.__("mainPanelFileMenuDuplicateConfirmH"),
                Resources.__("mainPanelFileMenuDuplicateConfirmMsg"));
        if (choice != JOptionPane.YES_OPTION) {
            return;
        }

        Main.frameMain.hide();
        Main.allowWindowKill(true);
        Console.windowLoad.start(null, INVOKABLE);
        new Thread(new Runnable() {
            @Override
            public void run() {
                ArrayList<String> paths = new ArrayList<String>();
                ArrayList<Long> sizes = new ArrayList<Long>();
                getPathAndSize(paths, sizes);

                String[] pathsArray = paths.toArray(new String[paths.size()]);
                long[] sizesArray = convertLong(sizes);
                String root = menu.getTree().getFilePath();

                boolean names = (mode == NAME || mode == NAME_CONTENT
                        || mode == NAME_CONTENT_EXACT);
                boolean content = (mode == CONTENT || mode == NAME_CONTENT
                        || mode == CONTENT_EXACT || mode == NAME_CONTENT_EXACT);
                boolean exact = (mode == CONTENT_EXACT || mode == NAME_CONTENT_EXACT);

                FileDisplayTree resultTree = Main.query().runDuplicates(Main.database(),
                        pathsArray, sizesArray, root, names, content, exact);
                Main.frameMain.main().list.updateView(resultTree);

                Console.windowLoad.stop();
                Main.frameMain.modify();
                Main.frameMain.show(Main.currSnapshot());
                Main.allowWindowKill(false);
            }
        }).start();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Adds all currently selected nodes and subtrees to the provided {@code paths}
     *  and {@code sizes}, both of which are assumed to be initially empty.
     * @param paths the {@link ArrayList} of String paths to be populated by this method
     * @param sizes the {@link ArrayList} of Long sizes to be populated by this method
     */
    private void getPathAndSize(ArrayList<String> paths, ArrayList<Long> sizes) {
        for (int i = 0; i < menu.currNodes().length; i++) {
            String nodePath = menu.currNodes()[i].getFilePath();
            File f = menu.currNodes()[i].getNodePathAsFile(menu.isFlat());

            if (nodePath != null) {
                FileDisplayTree tree = menu.getTree().getTree(menu.currNodes()[i].getPath());
                paths.add(tree.getFilePath());
                sizes.add(tree.getFileSize());

            } else if (f != null) {
                FileDisplayTree tree = menu.getTree().getTree(menu.currNodes()[i].getPath());
                for (String path : tree.getPaths()) {
                    paths.add(path);
                }
                for (long size : tree.getPathSizes()) {
                    sizes.add(size);
                }
            }
        }
    }

    /** Converts an {@link ArrayList} of {@link Long} to its primitive counterpart. */
    private long[] convertLong(ArrayList<Long> array) {
        long[] toReturn = new long[array.size()];
        for (int i = 0; i < toReturn.length; i++) {
            toReturn[i] = array.get(i);
        }
        return toReturn;
    }
}
