package edu.uwyo.crawl.ui.popup;

import java.awt.Color;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.UIManager;

import edu.uwyo.crawl.ui.panels.PanelFiles;
import edu.uwyo.crawl.ui.tree.AbstractFileNode;
import edu.uwyo.crawl.ui.tree.AbstractFileNodePopup;
import net.javamon.dev.AbstractFrame;
import net.javamon.dev.Resources;
import net.javamon.dev.utils.HTML;
import net.javamon.dev.utils.JAcceleratedPopupMenu;

/**
 * The context menu used when clicking on a file in the search results.
 * Allows the user to manipulate the file, flag it as ignored, etc.
 * @version 19 February, 2019
 * @author Christopher Stanton
 */
public class FilePopupMenu extends AbstractFileNodePopup {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final Color NORMAL = new JMenuItem().getBackground();
    private static Color SELECTED;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private JAcceleratedPopupMenu menu;

    private JMenuItem copyPath;
    private JMenuItem copyName;
    private JMenuItem copyContents;

    private JMenuItem openFile;
    private JMenuItem openFileLoc;

    private JMenu export;
    private JMenuItem exportAll;
    private JMenuItem exportSub;
    private JMenuItem exportSingle;

    private JMenu duplicateCheck;
    private JMenuItem[] duplicateMode;

    private JMenu flagAs;
    private JMenuItem[] flags;

    private boolean flat;
    private AbstractFileNode[] currNodes;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Creates a new {@link FilePopupMenu}. */
    public FilePopupMenu() {
        Object o = UIManager.get("MenuItem.selectionBackground"); //might not exist at runtime
        if (o instanceof Color) {
            SELECTED = (Color) o;
        } else {
            SELECTED = NORMAL;
        }

        menu = new JAcceleratedPopupMenu();

        setCopyPath();
        setCopyName();
        setCopyContents();
        menu.addSeparator();

        setOpenFile();
        setOpenFileLoc();
        menu.addSeparator();

        setExport();
        setDuplicateCheck();
        menu.addSeparator();

        setFlagAs();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    protected void show(AbstractFileNode[] nodes, boolean flat, MouseEvent e) {
        this.currNodes = nodes;
        this.flat = flat;
        boolean folderOnly = true;

        for (int i = 0; i < currNodes.length; i++) {
            if (currNodes[i].getFilePath() != null) {
                folderOnly = false;
                int currFlag = PanelFiles.indexOf(currNodes[i].getColor());
                for (int j = 0; j < flags.length; j++) {
                    flags[j].setBackground((j != currFlag) ? NORMAL : SELECTED);
                    flags[j].setEnabled(true);
                }
                break;
            }
        }

        copyContents.setEnabled(!folderOnly);
        openFile.setEnabled(!folderOnly);
        flagAs.setEnabled(!folderOnly);
        if (folderOnly) {
            for (int i = 0; i < flags.length; i++) {
                flags[i].setEnabled(false);
                flags[i].setBackground(NORMAL);
            }
        }
        menu.show(e.getComponent(), e.getX(), e.getY());
    }

    @Override
    protected boolean pressKey(AbstractFileNode[] nodes, KeyEvent e) {
        this.currNodes = nodes;
        for (int i = 0; i < currNodes.length; i++) {
            if (currNodes[i].getFilePath() != null) {
                return menu.pressKey(e);
            }
        }
        return false;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the array of currently selected {@link AbstractFileNode} */
    protected AbstractFileNode[] currNodes() {
        return currNodes;
    }

    /**
     * @return a boolean, {@code true} if the current tree is "flat"
     *  (i.e. not a hierarchical file view)
     */
    protected boolean isFlat() {
        return flat;
    }

    /** Sets the copy file path option. */
    private void setCopyPath() {
        copyPath = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuCopyPath")));
        copyPath.addActionListener(new CopyPathListener(this));
        AbstractFrame.setShortcut(copyPath, KeyEvent.VK_C, false, false, true);
        menu.addMenuItem(copyPath);
    }

    /** Sets the copy file name option. */
    private void setCopyName() {
        copyName = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuCopyName")));
        copyName.addActionListener(new CopyNameListener(this));
        AbstractFrame.setShortcut(copyName, KeyEvent.VK_C, true, false, true);
        menu.addMenuItem(copyName);
    }

    /** Sets the copy file contents option. */
    private void setCopyContents() {
        copyContents = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuCopyContents")));
        copyContents.addActionListener(new CopyContentsListener(this));
        AbstractFrame.setShortcut(copyContents, KeyEvent.VK_C, false, true, true);
        menu.addMenuItem(copyContents);
    }

    /** Sets the open file option. */
    private void setOpenFile() {
        openFile = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuOpenFile")));
        openFile.addActionListener(new OpenFileListener(this));
        AbstractFrame.setShortcut(openFile, KeyEvent.VK_O, false, true, false);
        menu.addMenuItem(openFile);
    }

    /** Sets the open file location option. */
    private void setOpenFileLoc() {
        openFileLoc = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuOpenFileLoc")));
        openFileLoc.addActionListener(new OpenFileLocationListener(this));
        AbstractFrame.setShortcut(openFileLoc, KeyEvent.VK_O, true, false, false);
        menu.addMenuItem(openFileLoc);
    }

    /** Sets the export results option. */
    private void setExport() {
        export = new JMenu(HTML.wrap(Resources.__("mainPanelFileMenuExport")));

        exportAll = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuExportAll")));
        exportAll.addActionListener(new ExportListener(this, ExportListener.ALL));
        export.add(exportAll);

        exportSub = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuExportSub")));
        exportSub.addActionListener(new ExportListener(this, ExportListener.SUB));
        export.add(exportSub);

        exportSingle = new JMenuItem(HTML.wrap(Resources.__("mainPanelFileMenuExportSingle")));
        exportSingle.addActionListener(new ExportListener(this, ExportListener.SINGLE));
        export.add(exportSingle);

        menu.addMenu(export, true);
    }

    /** Sets the duplicate-checking option. */
    private void setDuplicateCheck() {
        duplicateCheck = new JMenu(HTML.wrap(Resources.__("mainPanelFileMenuDuplicate")));

        duplicateMode = new JMenuItem[DuplicateCheckListener.NUM_MODES];
        for (int i = 0; i < duplicateMode.length; i++) {
            duplicateMode[i] = new JMenuItem(
                    HTML.wrap(Resources.__("mainPanelFileMenuDuplicate" + i)));
            duplicateMode[i].addActionListener(new DuplicateCheckListener(this, i));
            duplicateCheck.add(duplicateMode[i]);
        }

        menu.addMenu(duplicateCheck, true);
    }

    /** Sets the "flag as" submenu and its options. */
    private void setFlagAs() {
        flagAs = new JMenu(HTML.wrap(Resources.__("mainPanelFileMenuFlagAs")));

        flags = new JMenuItem[PanelFiles.getNumFlags()];
        for (int i = 0; i < flags.length; i++) {
            flags[i] = new JMenuItem(HTML.wrap(Resources.__("mainFlag" + i + "NoHTML")));
            flags[i].addActionListener(new ChangeFlagListener(this, i));
            if (i != 0) {
                flags[i].setAccelerator(PanelFiles.getKeyStroke(i));
                flagAs.add(flags[i]); //index 0 is used internally only
            }
        }

        menu.addMenu(flagAs, true);
    }
}
