package edu.uwyo.crawl.ui.tree;

import java.awt.Color;
import java.io.File;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

/**
 * An abstract class used for creating tree nodes that are associated with a specific file.
 * Implementers of this class can use the file name, type, etc. to suggest a color
 *  used for highlighting that file in the displayed tree.
 * <p>
 * <i>Only one instance of this class should be instantiated
 *  and passed into {@link PanelFileTree}.</i>
 * That instance will then be used internally to generate new nodes as needed
 *  via {@link #create(String, Object)}.
 * @version 21 February, 2019
 * @author Christopher Stanton
 */
public abstract class AbstractFileNode extends DefaultMutableTreeNode {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final long serialVersionUID = -5165045093185370972L;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private String toolTipText;

    private final String path;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link AbstractFileNode}.
     * @param path the String absolute file path to be associated with this node
     * @param contents the {@link Object} contents to be placed in this node;
     *  in most cases, this is the String to be displayed by the node
     */
    protected AbstractFileNode(String path, Object contents) {
        super(contents);
        this.path = path;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                ABSTRACT METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Obtains the color that should be used to highlight this node.
     * If the implementing class does not use highlighting,
     *  this method will always return {@code null}.
     * <p>
     * <i>This method always defers to the color set by {@link #setColor(Color)}.</i>
     * @return the {@link Color} to highlight this node, or {@code null} if none
     * @see #getFlagForPath(String)
     */
    public abstract Color getColor();

    /**
     * Sets the color that should be used to highlight this node.
     * If the implementing class does not use highlighting, this method can be left blank.
     * <p>
     * <i>This method guarantees that {@code color} will be returned
     *  by the next call to {@link #getColor()} on this node.</i>
     * @param color the {@link Color} to use for highlighting this node;
     *  if {@code null}, the current set color will be removed and {@link #getColor()} will behave
     *  as if no color was ever manually set
     */
    public abstract void setColor(Color color);

    /**
     * Creates an instance of this node.
     * The created node will have no relation whatsoever to this instance.
     * <p>
     * This is a <i>builder method</i> used to allow a single instance of {@link AbstractFileNode}
     *  to generate multiple new nodes of the same derived type,
     *  for easier integration with {@link PanelFileTree}.
     * @param path the String absolute file path to be associated with this node
     * @param contents the {@link Object} contents to be placed in this node;
     *  in most cases, this is the String to be displayed by the node
     * @return the created {@link AbstractFileNode}
     */
    public abstract AbstractFileNode create(String path, Object contents);

    /**
     * Determines whether this node should be excluded from the current tree view.
     * Used for filtering which nodes can appear in the tree.
     * @return a boolean, {@code true} if this node should not appear in the tree
     */
    public abstract boolean excludeFromTree();

    /**
     * Obtains the flag for a given file.
     * <i>This is a factory method that does not use any part of the instance node's state.</i>
     * @param filePath the String file path for which to obtain a flag
     * @return the integer flag for {@code filePath}, or -1 if none
     * @see #getColor()
     */
    public abstract int getFlagForPath(String filePath);

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Obtains the absolute file path depicted by this node.
     * This path is set during construction and will <i>not</i> change
     *  even if the node text itself is changed.
     * @return the String absolute file path represented by this node, or {@code null} if none
     * @see #getNodeName()
     * @see #getNodeFile(boolean)
     * @see #getNodePath()
     * @see #getNodePathAsFile(boolean)
     */
    public String getFilePath() {
        return path;
    }

    /**
     * Obtains the display name of this node, excluding any parent names.
     * This <i>may</i> correspond to the folder or file name,
     *  but might be different if the node's display text ever changes.
     * @return the String display name of this node
     * @see #getNodeFile(boolean)
     */
    public String getNodeName() {
        return getUserObject().toString();
    }

    /**
     * Obtains the file represented by the display name of this node, excluding any parent paths.
     * Because this method is based solely on a node's appearance,
     *  the file might be different if the node's display text ever changes.
     * <i>There is no guarantee that the constructed file exists or is a valid pathname.</i>
     * <p>
     * If {@code flat == true}, this method is identical to constructing a file
     *  around {@link #getFilePath()}.
     * @param flat a boolean, {@code true} if the current tree is "flat"
     *  (i.e. not a hierarchical file view)
     * @return the {@link File} represented by {@link #getNodeName()}, or {@code null} if none
     */
    public File getNodeFile(boolean flat) {
        if (flat) {
            return getFilePath() == null ? null : new File(getFilePath());
        }
        return new File(getUserObject().toString());
    }

    /**
     * Obtains the path depicted by this node, as a sum of its parent node paths.
     * This method is equivalent to calling {@link #getNodeName()} on each parent node
     *  (including this node) and concatenating them together, in order.
     * <p>
     * Because this method is based solely on a node's appearance,
     *  the path might be different if the node's display text ever changes.
     * If directory separators (e.g. slashes) must be inserted between each node,
     *  use {@link #getNodePathAsFile(boolean)} instead.
     * @return the String absolute file path represented by this node; will never be {@code null}
     */
    public String getNodePath() {
        StringBuilder builder = new StringBuilder();
        TreeNode[] path = getPath();
        for (int i = 0; i < path.length - 1; i++) {
            builder.append(((DefaultMutableTreeNode) path[i]).getUserObject());
        }

        builder.append(((DefaultMutableTreeNode) path[path.length - 1]).getUserObject());
        return builder.toString();
    }

    /**
     * Obtains the absolute file depicted by this node, as a sum of its parent node file paths.
     * This method can be used when {@link #getFilePath()}
     *  would normally return {@code null} for a node.
     * Because this method is based solely on a node's appearance,
     *  the file might be different if the node's display text ever changes.
     * <i>There is no guarantee that the constructed file exists or is a valid pathname.</i>
     * <p>
     * Intermediate nodes (i.e. directories) will always include a trailing slash.
     * If {@code flat == true}, this method is identical to {@link #getNodeFile(boolean)}.
     * @param flat a boolean, {@code true} if the current tree is "flat"
     *  (i.e. not a hierarchical file view)
     * @return the {@link File} represented by this node; can only be {@code null}
     *  if {@code flat == false}
     * @see #getNodePath()
     */
    public File getNodePathAsFile(boolean flat) {
        if (flat) { return getNodeFile(flat); }

        StringBuilder builder = new StringBuilder();
        TreeNode[] path = getPath();
        for (int i = 0; i < path.length - 1; i++) {
            builder.append(((DefaultMutableTreeNode) path[i]).getUserObject());
            builder.append("/");
        }

        builder.append(((DefaultMutableTreeNode) path[path.length - 1]).getUserObject());
        if (getFilePath() == null) {
            builder.append("/");
        }
        return new File(builder.toString());
    }

    /**
     * Sets the text to be shown when hovering over this node.
     * @param toolTipText the String text to show when hovering over this node;
     *  <i>must already be localized</i>
     */
    protected void setToolTipText(String toolTipText) {
        this.toolTipText = toolTipText;
    }

    /** @return the String text to be shown when hovering over this node */
    protected String getToolTipText() {
        return toolTipText;
    }
}
