package edu.uwyo.crawl.ui.tree;

import java.util.StringTokenizer;

import javax.swing.tree.TreePath;

/**
 * A class used for saving or restoring a tree's expansion state
 *  (i.e. the configuration of all expanded/collapsed nodes).
 * @version 6 March, 2019
 * @author Christopher Stanton
 */
public final class ExpansionState {
    /** Don't allow construction of this class. */
    private ExpansionState() { throw new AssertionError(); }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final String DELIMITER = ",";

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Serializes the current expansion state of {@code tree} into a format
     *  that can be later restored via {@link #restore(FileNodeToolTipTree, String)}.
     * @param tree the {@link FileNodeToolTipTree} whose expansion state should be saved
     * @return the String representation of the tree's expansion state
     */
    public static String save(FileNodeToolTipTree tree) {
        final int rowCount = tree.getRowCount();
        TreePath rowPath = tree.getPathForRow(0);
        StringBuilder buffer = new StringBuilder();

        for (int i = 0; i < rowCount; i++) {
            TreePath path = tree.getPathForRow(i);

            if (i == 0 || isDescendant(path, rowPath)) {
                if (tree.isExpanded(path)) { buffer.append(DELIMITER + i); }
            } else {
                break;
            }
        }
        return buffer.toString();
    }

    /**
     * Sets a tree's expansion state to {@code state},
     *  as was obtained previously by {@link #save(FileNodeToolTipTree)}.
     * <i>Calling this method on a tree containing a different node structure than in {@code state}
     *  may cause unexpected behavior.</i>
     * @param tree the {@link FileNodeToolTipTree} to modify
     * @param state the String serialized tree state that should be restored
     */
    public static void restore(FileNodeToolTipTree tree, String state) {
        StringTokenizer tokenizer = new StringTokenizer(state, DELIMITER);
        while (tokenizer.hasMoreTokens()) {
            int token = Integer.parseInt(tokenizer.nextToken());
            tree.expandRow(token);
        }
    }

    /**
     * Determines whether the tree denoted by {@code path1}
     *  is the descendant of another tree {@code path2}.
     * @param path1 the first {@link TreePath}
     * @param path2 the second {@link TreePath}
     * @return a boolean, {@code true} if {@code path1} is a descendant of {@code path2}
     */
    private static boolean isDescendant(TreePath path1, TreePath path2) {
        int count1 = path1.getPathCount();
        int count2 = path2.getPathCount();
        if(count1 <= count2) {
            return false;
        }

        while (count1 != count2) {
            path1 = path1.getParentPath();
            count1--;
        }
        return path1.equals(path2);
    }
}
