package edu.uwyo.crawl.ui.tree;

import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Set;

import edu.uwyo.crawl.database.DataInput;
import edu.uwyo.crawl.database.DataOutput;
import edu.uwyo.crawl.database.Serializer;

/**
 * A tree view of files.
 * Primarily used to convert flattened file lists (e.g. {@code a/b.txt, a/b/c.txt, a/d/e.txt})
 *  to a hierarchical structure that can be used by a {@link FileNodeToolTipTree}
 *  for graphical browsing.
 * New files are added to the tree via {@link #add(String, String, boolean, long)}.
 * An optional "flat mode" e.g. for viewing collections of duplicate files
 *  can be enabled via {@link #setFlatMode(boolean)}.
 * <p>
 * This class does not discriminate between folders and files within its hierarchy.
 * If a subtree's {@link #getChildren()} returns a zero-size result,
 *  <i>that tree might represent either a file or an empty folder.</i>
 * In other words, <b>there is no way to distinguish between folders and files
 *  given an arbitrary instance of this class</b> unless the constructing class
 *  can guarantee that only files (not folders) will be added.
 * <p>
 * This class is serializable to disk, as used by very large trees
 *  that cannot be otherwise represented in memory.
 * @version 20 February, 2019
 * @author Christopher Stanton
 */
public class FileDisplayTree implements Serializable, Serializer<FileDisplayTree> {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** Used for serialization; <b>do not modify!</b> */
    public static final FileDisplayTree INSTANCE = new FileDisplayTree();

    private static final String DELIMITERS = "[\\\\\\/]+";
    private static final long serialVersionUID = 1994080994513940074L;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private boolean isRoot;
    private boolean flatMode;
    private int size;

    private String filePath;
    private String path;
    private long fileSize;

    private HashMap<String, FileDisplayTree> children;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new root {@link FileDisplayTree}.
     * The {@code root} file can be {@code null} if it is not yet known,
     *  but must be set via {@link #clear(String)} before any files are added to the tree.
     * @param rootPath the String absolute file path containing the root of this tree;
     *  <b>must never contain child paths</b> (e.g. if {@code "my/path/is/provided"},
     *  the root will be {@code "my/path/is/provided"}, not {@code "my"})
     */
    public FileDisplayTree(String rootPath) {
        this(true, rootPath, getRoot(rootPath));
    }

    /**
     * Creates a new {@link FileDisplayTree}.
     * @param isRoot a boolean, {@code true} if this is the root tree
     * @param filePath the String absolute file path referenced by this node
     * @param path the String path name of this node
     */
    private FileDisplayTree(boolean isRoot, String filePath, String path) {
        this();
        this.isRoot = isRoot;
        this.filePath = filePath;
        this.path = path;
    }

    /** Creates a new {@link FileDisplayTree}. */
    private FileDisplayTree() {
        children = new HashMap<String, FileDisplayTree>();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void serialize(DataOutput output, FileDisplayTree instance) throws IOException {
        output.writeBoolean(instance.isRoot);
        output.writeBoolean(instance.flatMode);
        output.writeInt(instance.size);
        output.writeLong(instance.fileSize);

        output.writeBoolean(instance.filePath != null);
        if (instance.filePath != null) {
            output.writeUTF(instance.filePath);
        }

        output.writeBoolean(instance.path != null);
        if (instance.path != null) {
            output.writeUTF(instance.path);
        }

        Set<String> keys = instance.children.keySet();
        output.writeShort((short) keys.size());
        for (String key : keys) {
            output.writeUTF(key);
            serialize(output, instance.children.get(key));
        }
    }

    @Override
    public FileDisplayTree deserialize(DataInput input) throws IOException {
        FileDisplayTree toReturn = new FileDisplayTree();
        toReturn.isRoot = input.readBoolean();
        toReturn.flatMode = input.readBoolean();
        toReturn.size = input.readInt();
        toReturn.fileSize = input.readLong();

        if (input.readBoolean()) {
            toReturn.filePath = input.readUTF();
        }

        if (input.readBoolean()) {
            toReturn.path = input.readUTF();
        }

        int numChildren = input.readShort();
        for (int i = 0; i < numChildren; i++) {
            String path = input.readUTF();
            FileDisplayTree value = deserialize(input);
            toReturn.children.put(path, value);
        }

        return toReturn;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** @return the integer size of this tree, including all child trees */
    public int size() {
        return size;
    }

    /** @return the long total file size of this tree, including all child trees, in bytes */
    public long fileSize() {
        return fileSize;
    }

    /** @return a boolean, {@code true} if this tree contains no children */
    public boolean isEmpty() {
        return children.size() == 0;
    }

    /** @return a boolean, {@code true} if this tree is currently in flat-view mode */
    public boolean isFlat() {
        return flatMode;
    }

    /**
     * Enables flat-view mode.
     * The default representation of this tree is hierarchical;
     *  an added file's parent paths are each separate nodes in the tree,
     *  allowing the user to browse the file tree structure.
     * However, some trees may need to display files in a flat structure
     *  (e.g. when displaying duplicates), which is what "flat mode" is for.
     * <p>
     * Enabling flat mode affects files that are newly added to the tree
     *  via {@link #add(String, String, boolean, long)}.
     * Consequently, <i>the current display mode should only be changed after clearing,
     *  but before adding files to the tree.</i>
     * @param enabled a boolean, {@code true} if flat mode should be enabled
     */
    public void setFlatMode(boolean enabled) {
        flatMode = enabled;
    }

    /**
     * Clears this tree of all children.
     * Children are only deleted if they have not been referenced elsewhere.
     * @param rootPath the new String absolute file path containing the root of this tree;
     *  <b>must never contain child paths</b> (e.g. if {@code "my/path/is/provided"},
     *  the root will be {@code "my/path/is/provided"}, not {@code "my"})
     */
    public void clear(String rootPath) {
        if (isRoot) {
            this.filePath = rootPath;
            this.path = getRoot(rootPath);
        }
        size = 0;
        fileSize = 0L;
        children.clear();
    }

    /**
     * Obtains an array of all file paths represented within this tree, excluding the root.
     * <i>The returned files are guaranteed to be returned in a fixed, depth-first ordering.</i>
     * @return the array of String file paths in this tree and its children
     */
    public String[] getPaths() {
        String[] toReturn = new String[size];
        getPaths(toReturn, 0);
        return toReturn;
    }

    /**
     * Obtains an array of all prefixes represented within this tree.
     * If {@link #isFlat()} returns {@code false}, the resulting array will be full
     *  of {@code null} values.
     * <p>
     * The returned sizes are guaranteed to correspond <u>exactly</u> to the file ordering
     *  returned by {@link #getPaths()}.
     * Consequently, the returned array <i>will</i> contain duplicates if multiple trees exist
     *  under any single prefix.
     * @return the array of String prefixes in this tree and its children
     */
    public String[] getPrefixes() {
        String[] toReturn = new String[size];
        if (!flatMode) {
            return toReturn;
        }
        getPrefixes(toReturn, null, 0);
        return toReturn;
    }

    /**
     * Obtains an array of all file sizes represented within this tree, excluding the root.
     * <p>
     * The returned sizes are guaranteed to correspond <u>exactly</u> to the file ordering
     *  returned by {@link #getPaths()}.
     * Because that method does not return directories, this method will not return
     *  directory sizes either.
     * The combination of {@link #getPaths()} and this method can thus be used to recreate the tree
     *  via successive calls to {@link #add(String, String, boolean, long)}.
     * @return the array of long file sizes in this tree and its children
     */
    public long[] getPathSizes() {
        long[] toReturn = new long[size];
        getPathSizes(toReturn, 0);
        return toReturn;
    }

    /**
     * Obtains an array of all file flags represented within this tree.
     * The returned flags are guaranteed to correspond <u>exactly</u> to the file ordering
     *  returned by {@link #getPaths()}.
     * @param factoryNode the {@link AbstractFileNode} used to determine a file's flags
     * @return the array of integer flags in this tree and its children
     */
    public int[] getFlags(AbstractFileNode factoryNode) {
        int[] toReturn = new int[size];
        getFlags(factoryNode, toReturn, 0);
        return toReturn;
    }

    /**
     * Prints the current tree.
     * @param currIndent the String current indent to use when printing this tree;
     *  <i>this argument is completely ignored if this method is called on the root tree</i>
     */
    public void print(String currIndent) {
        if (isRoot) {
            currIndent = "";
        }

        System.out.println(currIndent + getPath());
        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            children.get(child).print(currIndent + "    ");
        }
    }

    /**
     * Forces the tree's root to be set to {@code newRoot}, also updating all of its children.
     * Used when moving the entire tree to a new directory.
     * @param oldRoot the String old file root; can be {@code null} if called on the root tree
     * @param newRoot the String new file root to use
     */
    public void forceRoot(String oldRoot, String newRoot) {
        if (isRoot) {
            oldRoot = filePath;
            filePath = newRoot;
            path = getRoot(newRoot);
        } else {
            filePath = filePath.replace(oldRoot, newRoot);
        }

        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            children.get(child).forceRoot(oldRoot, newRoot);
        }
    }

    /**
     * Adds a file to this result tree.
     * <b>If this method is called on anything other than the root tree, it does nothing.</b>
     * Parent folders are recursively created such that a full file tree will be displayed:
     * <pre>
     * path/to/root (must be a file/folder)
     * | This
     * | --{@literal >} is
     * | --{@literal >} --{@literal >} a
     * | --{@literal >} --{@literal >} --{@literal >} file
     * | --{@literal >} --{@literal >} --{@literal >} different
     * | --{@literal >} --{@literal >} --{@literal >} --{@literal >} file
     * | Here
     * | --{@literal >} is
     * | --{@literal >} --{@literal >} another
     * | --{@literal >} --{@literal >} --{@literal >} file</pre>
     * <p>
     * If flat mode is enabled, the tree structure will instead look like the following:
     * <pre>
     * rootName (can be any text)
     * | prefix1
     * | --{@literal >} filePath1
     * | --{@literal >} filePath2
     * | prefix2
     * | ...
     * | prefixN</pre>
     * <p>
     * In flat mode, the root name is the text passed into {@link #clear(String)}
     *  and does <i>not</i> need to be a file/folder name.
     * Each {@code prefix} added via this method is a child of the root tree.
     * Finally, {@code filePath} is created as a single node under {@code prefix}.
     * <i>The file path will <u>not</u> be expanded in flat mode;
     *  the tree can only ever be 3 nodes deep.</i>
     * @param prefix the String prefix into which the result should be placed;
     *  only applicable if "flat mode" was enabled via {@link #setFlatMode(boolean)}
     * @param filePath the String absolute file path to add; must begin with the root tree path
     *  or this method does nothing
     * @param isDirectory a boolean, {@code true} if {@code filePath} represents a directory
     * @param size the long size of the file to be added, in bytes
     */
    public void add(String prefix, String filePath, boolean isDirectory, long size) {
        if (!isRoot) { return; }

        if (flatMode) {
            FileDisplayTree childToAdd = children.get(prefix);
            if (childToAdd == null) {
                children.put(prefix, new FileDisplayTree(false, null, prefix));
                childToAdd = children.get(prefix);
            }

            this.size++;
            this.fileSize += size;

            childToAdd.size++;
            childToAdd.fileSize += size;

            FileDisplayTree newChild = new FileDisplayTree(filePath);
            newChild.size = 1;
            newChild.fileSize = size;
            childToAdd.children.put(filePath, newChild);
            return;
        }

        String absolutePath = filePath;
        if (absolutePath.startsWith(path)) {
            absolutePath = absolutePath.substring(path.length());
            while (absolutePath.startsWith("\\") || absolutePath.startsWith("/")) {
                absolutePath = absolutePath.substring(1);
            }
            String[] tokens = absolutePath.split(DELIMITERS);
            add(isDirectory ? null : filePath, this, tokens, 0, size);
        }
    }

    /** @return the String absolute file path of this tree */
    public String getFilePath() {
        return filePath;
    }

    /** @return the String name of this tree's path */
    public String getPath() {
        return path;
    }

    /** @return the long total file size of this tree and its children, in bytes */
    public long getFileSize() {
        return fileSize;
    }

    /**
     * @return this tree's immediate children;
     *  if a file or empty folder, the number of children will be zero
     */
    public HashMap<String, FileDisplayTree> getChildren() {
        return children;
    }

    /**
     * Finds a subtree's children by following the given path descriptor.
     * Primarily used in combination with {@link FileNodeToolTipTree}.
     * <p>
     * <u>Example Usage</u>:
     * <p>
     * To find the children of a folder {@code "my/sub/folder"},
     *  pass an array <code>{"my", "sub", "folder"}</code> into this method.
     * In this example, {@code "my"} is the name of this tree as obtained by {@link #getPath()}.
     * @param paths an array of Object ({@link AbstractFileNode}) paths to follow;
     *  the first path must <u>always</u> be the same as this tree's path {@link #getPath()}
     * @return the {@link HashMap} of child trees that were found, or {@code null} if none
     */
    public HashMap<String, FileDisplayTree> getChildren(Object[] paths) {
        FileDisplayTree tree = getTree(this, paths, 0);
        return tree != null ? tree.children : null;
    }

    /**
     * Finds a subtree by following the given path descriptor.
     * Primarily used in combination with {@link FileNodeToolTipTree}.
     * <p>
     * <u>Example Usage</u>:
     * <p>
     * To find the tree {@code "my/sub/folder"},
     *  pass an array <code>{"my", "sub", "folder"}</code> into this method.
     * In this example, {@code "my"} is the name of this tree as obtained by {@link #getPath()}.
     * @param paths an array of Object ({@link AbstractFileNode}) paths to follow;
     *  the first path must <u>always</u> be the same as this tree's path {@link #getPath()}
     * @return the {@link FileDisplayTree} represented by {@code paths},
     *  or {@code null} if not found
     */
    public FileDisplayTree getTree(Object[] paths) {
        return getTree(this, paths, 0); //Object for compatibility with TreeWillExpandListener
    }

    /**
     * Adds this tree to the list of paths, then recursively adds its children.
     * @param toPut the array of String file paths into which this tree should be added
     * @param index the integer index in {@code toPut} indicating the next unused entry
     * @return the integer next value of {@code index}, after this tree and its children were added
     */
    private int getPaths(String[] toPut, int index) {
        if (filePath != null && !isRoot) {
            toPut[index] = filePath;
            index++;
        }

        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            index = children.get(child).getPaths(toPut, index);
        }
        return index;
    }

    /**
     * Adds this tree to the list of prefixes, then recursively adds it children.
     * @param toPut the array of String prefixes into which this tree should be added
     * @param currPrefix the String prefix that contains this tree
     * @param index the integer index in {@code toPut} indicating the next unused entry
     * @return the integer next value of {@code index}, after this tree and its children were added
     */
    private int getPrefixes(String[] toPut, String currPrefix, int index) {
        if (filePath != null && !isRoot) {
            toPut[index] = currPrefix;
            index++;
        }

        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            if (isRoot) {
                index = children.get(child).getPrefixes(toPut, child, index);
            } else {
                index = children.get(child).getPrefixes(toPut, currPrefix, index);
            }
        }
        return index;
    }

    /**
     * Adds this tree to the list of sizes, then recursively adds its children.
     * Ignores directories.
     * @param toPut the array of long file sizes into which this tree should be added
     * @param index the integer index in {@code toPut} indicating the next unused entry
     * @return the integer next value of {@code index}, after this tree and its children were added
     */
    private int getPathSizes(long[] toPut, int index) {
        if (filePath != null && !isRoot) {
            toPut[index] = fileSize;
            index++;
        }

        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            index = children.get(child).getPathSizes(toPut, index);
        }
        return index;
    }

    /**
     * Adds this tree to the list of flags, then recursively adds its children.
     * @param factoryNode the {@link AbstractFileNode} used to determine a file's flags
     * @param toPut the array of integer flags into which this tree should be added
     * @param index the integer index in {@code toPut} indicating the next unused entry
     * @return the integer next value of {@code index}, after this tree and its children were added
     */
    private int getFlags(AbstractFileNode factoryNode, int[] toPut, int index) {
        if (filePath != null && !isRoot) {
            toPut[index] = factoryNode.getFlagForPath(filePath);
            index++;
        }

        Set<String> childrenSet = children.keySet();
        for (String child : childrenSet) {
            index = children.get(child).getFlags(factoryNode, toPut, index);
        }
        return index;
    }

    /**
     * Converts a root directory into a format suitable for the root of this tree.
     * Used only during object construction.
     * @param rootPath the String unaltered root path
     * @return the String altered {@code rootPath},
     *  as would be suitable for constructing a root tree
     */
    private static String getRoot(String rootPath) {
        if (rootPath == null) { return null; }

        while (rootPath.endsWith("\\") || rootPath.endsWith("/")) {
            rootPath = rootPath.substring(0, rootPath.length() - 1);
        }
        return rootPath;
    }

    /**
     * Recursively adds a file to this result tree.
     * This method creates all the subfolder trees in {@code tokens} if needed.
     * @param filePath the String absolute file path to associate with the added tree
     * @param curr the current {@link FileDisplayTree}
     * @param tokens the array of String tokens representing the file to add
     * @param index the integer index in {@code tokens} of the current path to create
     * @param size the long size of the file to be added, in bytes
     */
    private static void add(String filePath, FileDisplayTree curr, String[] tokens, int index,
            long size) {
        curr.size++;
        curr.fileSize += size;
        if (index == tokens.length - 1) {
            FileDisplayTree leaf = new FileDisplayTree(false, filePath, tokens[index]);
            leaf.fileSize = size;
            curr.children.put(tokens[index], leaf);
            return;
        }

        FileDisplayTree child = curr.children.get(tokens[index]);
        if (child == null) {
            child = new FileDisplayTree(false, null, tokens[index]);
            curr.children.put(tokens[index], child);
        }
        add(filePath, child, tokens, index + 1, size);
    }

    /**
     * Recursively searches this tree for a subtree.
     * @param curr the current {@link FileDisplayTree} to search
     * @param paths the array of {@link AbstractFileNode} tokens
     *  representing the path of the file to search
     * @param index the integer index in {@code paths} of the current path token to find
     * @return the {@link FileDisplayTree} represented by {@code paths},
     *  or {@code null} if not found
     */
    private static FileDisplayTree getTree(FileDisplayTree curr, Object[] paths, int index) {
        String toCompare = (String) (((AbstractFileNode) paths[index]).getUserObject());

        if (index == paths.length - 1) {
            return curr.path.equals(toCompare) ? curr : null;
        }

        if (curr.path.equals(toCompare)) {
            toCompare = (String) (((AbstractFileNode) paths[index + 1]).getUserObject());
            FileDisplayTree child = curr.children.get(toCompare);
            return (child == null) ? null : getTree(child, paths, index + 1);
        }
        return null;
    }
}
