package edu.uwyo.crawl.ui.tree;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.tree.TreePath;

import edu.uwyo.crawl.ui.popup.OpenFileListener;

/**
 * A listener used for showing a context menu when one or more node(s) are clicked.
 * Supports dragging and shift-click to select a range of nodes.
 * @version 21 February, 2019
 * @author Christopher Stanton
 */
public class FileTreeMouseListener extends MouseAdapter {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private static final int SHIFT_CLICK = MouseEvent.SHIFT_DOWN_MASK
            | MouseEvent.BUTTON1_DOWN_MASK;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private int startRow;
    private int endRow;
    private int lastRowCount;
    private FileDisplayTree resultTree;

    private final FileNodeToolTipTree tree;
    private final AbstractFileNodePopup popup;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link FileTreeMouseListener}.
     * @param tree the {@link FileNodeToolTipTree} that will own this listener
     * @param popup the {@link AbstractFileNodePopup} to display when a mouse event occurs
     */
    protected FileTreeMouseListener(FileNodeToolTipTree tree, AbstractFileNodePopup popup) {
        this.tree = tree;
        this.popup = popup;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void mousePressed(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1) {
            if ((e.getModifiersEx() & SHIFT_CLICK) != SHIFT_CLICK) {
                startRow = tree.getClosestRowForLocation(e.getX(), e.getY());
            }
        }
        lastRowCount = tree.getRowCount();
        showPopup(e);
    }

    @Override
    public void mouseDragged(MouseEvent e) {
        endRow = tree.getClosestRowForLocation(e.getX(), e.getY());
        selectRange();
    }

    @Override
    public void mouseReleased(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1) {
            if ((e.getModifiersEx() & MouseEvent.CTRL_DOWN_MASK) == 0) {
                if (lastRowCount != tree.getRowCount()) {
                    lastRowCount = tree.getRowCount();
                    startRow = tree.getClosestRowForLocation(e.getX(), e.getY()); /*
                                                                                   * row may reflect
                                                                                   *  a different node
                                                                                   *  if an expansion
                                                                                   *  or collapse has
                                                                                   *  occurred; must
                                                                                   *  ensure that only
                                                                                   *  the expanded or
                                                                                   *  collapsed node
                                                                                   *  is selected and
                                                                                   *  not an undesired
                                                                                   *  range
                                                                                   */
                    endRow = startRow;
                } else {
                    endRow = tree.getClosestRowForLocation(e.getX(), e.getY());
                    selectRange();
                }
            }
        }
        showPopup(e);
    }

    @Override
    public void mouseClicked(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1 && e.getClickCount() == 2) {
            int[] rows = tree.getSelectionRows();
            AbstractFileNode[] nodes = new AbstractFileNode[rows.length];
            for (int i = 0; i < rows.length; i++) {
                TreePath path = tree.getPathForRow(rows[i]);
                nodes[i] = ((AbstractFileNode) path.getLastPathComponent());
            }
            OpenFileListener.openFiles(nodes);

        }
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the file tree that should be used for obtaining entries in the tree.
     * @param resultTree the {@link FileDisplayTree} to be displayed by this tree
     */
    protected void setTree(FileDisplayTree resultTree) {
        this.resultTree = resultTree;
        popup.setTree(resultTree);
    }

    /**
     * Selects a range of nodes, based on the current values of {@link #startRow}
     *  and {@link #endRow}.
     * Those values will <u>not</u> be modified by this method.
     */
    private void selectRange() {
        if (endRow == startRow) {
            tree.setSelectionRow(startRow);
            return;
        }

        int start = startRow;
        int end = endRow;
        if (start > end) {
            int temp = start;
            start = end;
            end = temp;
        }

        int[] rows = new int[end - start + 1];
        for (int i = 0; i < rows.length; i++) {
            rows[i] = start + i;
        }
        tree.setSelectionRows(rows);
    }

    /**
     * Shows the popup.
     * The popup is only shown if platform-dependent popup triggers (e.g. right-click on Windows)
     *  have been satisfied.
     * @param e the {@link MouseEvent} that triggered this listener
     */
    private void showPopup(MouseEvent e) {
        if (popup != null && e.isPopupTrigger()) {
            final int currRow = tree.getClosestRowForLocation(e.getX(), e.getY());

            boolean currInSelection = false;
            int[] rows = tree.getSelectionRows();
            for (int i = 0; i < rows.length; i++) {
                if (rows[i] == currRow) {
                    currInSelection = true;
                    break;
                }
            }

            if (!currInSelection) {
                tree.setSelectionRow(currRow);
                rows = tree.getSelectionRows();
            }

            AbstractFileNode[] nodes = new AbstractFileNode[rows.length];
            for (int i = 0; i < rows.length; i++) {
                TreePath path = tree.getPathForRow(rows[i]);
                nodes[i] = ((AbstractFileNode) path.getLastPathComponent());
            }
            popup.show(nodes, (resultTree != null && resultTree.isFlat()), e);
        }
    }
}
