package edu.uwyo.crawl.ui.tree;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.NoSuchElementException;
import java.util.Set;

import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;

import net.javamon.dev.Resources;
import net.javamon.dev.StringUtils;
import net.javamon.dev.utils.HTML;

/**
 * A lazy-load listener for {@link FileNodeToolTipTree} designed for use with extremely large
 *  file trees, where the tree structure is defined by a {@link FileDisplayTree}.
 * New nodes are only created when a folder node is about to be expanded,
 *  and are immediately released from memory when a folder node has been closed.
 * <p>
 * To ensure that folders are properly displayed, this class adds a {@link #FOLDER_SLUG}
 *  to all unopened folder nodes.
 * The owner {@link PanelFileTree} must create a root node with this slug to ensure proper behavior
 *  when the tree's contents are reconfigured.
 * The currently displayed tree can be changed via {@link #setTree(FileDisplayTree)}.
 * @version 21 February, 2019
 * @author Christopher Stanton
 */
public class LazyFileLoadListener implements TreeWillExpandListener {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * This slug must be added to all unexpanded folders to ensure that the user
     *  can actually expand them.
     * These slugs will be automatically removed
     *  and replaced with their real contents when the folder is expanded.
     */
    protected static final String FOLDER_SLUG = "defaultFolderSlugForLazyLoadListener";

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private FileDisplayTree resultTree;

    private final DefaultTreeModel model;
    private final AbstractFileNode nodeFactory;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link LazyFileLoadListener}.
     * @param model the {@link DefaultTreeModel} used by the tree
     * @param nodeFactory the {@link AbstractFileNode} used for constructing new nodes in the tree
     */
    protected LazyFileLoadListener(DefaultTreeModel model, AbstractFileNode nodeFactory) {
        this.model = model;
        this.nodeFactory = nodeFactory;
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException {
        HashMap<String, FileDisplayTree> map = resultTree.getChildren(event.getPath().getPath());
        if (map == null || map.size() == 0) {
            throw new ExpandVetoException(event, "folder is empty");
        }

        AbstractFileNode currNode = (AbstractFileNode) event.getPath().getLastPathComponent();
        try {
            AbstractFileNode child = (AbstractFileNode) currNode.getFirstChild();
            if (child.getUserObject().equals(FOLDER_SLUG)) {
                currNode.remove(currNode.getIndex(child));
            }
        } catch (NoSuchElementException e) {}

        Set<String> children = map.keySet();
        add(currNode, children, map, true);
        add(currNode, children, map, false);
        model.reload(currNode);
    }

    @Override
    public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException {
        AbstractFileNode currNode = (AbstractFileNode) event.getPath().getLastPathComponent();
        if (currNode.getChildCount() > 0) {
            currNode.removeAllChildren();
            currNode.add(nodeFactory.create(null, FOLDER_SLUG));
        }
        model.reload(currNode);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Sets the file tree that should be used for obtaining entries in the tree.
     * @param resultTree the {@link FileDisplayTree} to be displayed by this tree
     */
    protected void setTree(FileDisplayTree resultTree) {
        this.resultTree = resultTree;
    }

    /**
     * Loads the current node's children (if any) into this tree.
     * This method is typically called in two passes: first with {@code folders == true}
     *  to load all folders, and second with {@code folders == false} to load all files.
     * This behavior ensures a more natural browsing experience in which all folders appear
     *  at the top and all files beneath.
     * @param currNode the {@link AbstractFileNode} representing the current node to be expanded
     * @param children the {@link Set} of children owned by {@code currNode}
     * @param map the {@link HashMap} of {@link FileDisplayTree} instances that were found
     *  at the current node's path
     * @param folders a boolean, {@code true} if only folders should be added
     *  and {@code false} if only files should be added
     */
    private void add(AbstractFileNode currNode, Set<String> children,
            HashMap<String, FileDisplayTree> map, boolean folders) {
        ArrayList<String> sortedChildren = new ArrayList<String>(children);
        Collections.sort(sortedChildren);

        for (String child : sortedChildren) {
            int childrenSize = map.get(child).getChildren().size();

            if ((folders && childrenSize > 0) || (!folders && childrenSize == 0)) {
                AbstractFileNode childNode = nodeFactory.create(map.get(child).getFilePath(), child);
                if (childNode.excludeFromTree()) {
                    continue;
                }

                if (folders) {
                    childNode.add(nodeFactory.create(null, FOLDER_SLUG));
                }
                currNode.add(childNode);

                if (folders) {
                    childNode.setToolTipText(HTML.wrap(Resources.__("mainNodeFolderTT",
                            map.get(child).size(),
                            StringUtils.getFormattedBytes(map.get(child).fileSize()))));
                } else {
                    childNode.setToolTipText(HTML.wrap(Resources.__("mainNodeFileTT",
                            StringUtils.getFormattedBytes(map.get(child).fileSize()))));
                }
            }
        }
    }
}
