package edu.uwyo.crawl.ui.tree;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;

import javax.swing.JScrollPane;
import javax.swing.ToolTipManager;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import net.javamon.dev.Resources;
import net.javamon.dev.StringUtils;
import net.javamon.dev.utils.HTML;
import net.javamon.dev.utils.JLabelHTML;
import net.javamon.dev.utils.layout.AbstractPanel;

/**
* A panel for displaying a traversable file tree that can be expanded and collapsed
*  similar to file browsers.
* The tree's file listing is updated via {@link #updateView(FileDisplayTree)}.
* In order to use this class:
* <ul>
*  <li>The {@link AbstractFileNode} class (defining how files are represented in the tree)
*  must be extended and a single instance provided to this class's constructor.</li>
*  <li>A {@link FileDisplayTree} must be separately instantiated and maintained
*  with a list of files.</li>
*  <li>(Optional) The {@link AbstractFileEntryRenderer} class can be extended to provide
*  a custom node visualization. Otherwise, the default {@link FileNodeToolTipTree} display
*  will be used.</li>
*  <li>(Optional) The {@link AbstractFileNodePopup} class can be extended to provide a context menu
*  that will be displayed whenever a specific node in the tree is selected.</li>
* </ul>
* <p>
* Files in the tree will be loaded lazily to save memory; a folder's children will not be added
*  to the display until that folder is actually expanded.
* Furthermore, all of a folder's children will be recursively deleted
*  when a folder has been collapsed.
* Consequently, <b>this panel can be used to display extremely large file directories,</b>
*  provided the user doesn't expand too many directories simultaneously.
* <p>
* Scroll bars will automatically appear if the file tree becomes too wide or too tall
*  to display in this panel.
* The panel itself, however, will not automatically resize unless placed in a layout permitting it
*  to do so.
* @version 21 February, 2019
* @author Christopher Stanton
*/
public class PanelFileTree extends AbstractPanel {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private JScrollPane scrollPane;
    private FileNodeToolTipTree tree;
    private AbstractFileNode root;
    private LazyFileLoadListener treeListener;
    private FileTreeMouseListener treeMouseListener;
    private FileDisplayTree resultTree;

    private JLabelHTML noResults;

    private final AbstractFileNode nodeFactory;
    private final AbstractFileEntryRenderer entryRenderer;
    private final boolean equalRowHeights;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link PanelFileTree}.
     * <p>
     * For very large trees, the {@code largeModel} or {@code equalRowHeights} parameters
     *  can be used to improve rendering performance.
     * The {@code equalRowHeights} option nearly always improves performance, but assumes that
     *  all rows in the tree will have identical heights.
     * The {@code largeModel} option works even if rows have different heights, but can sometimes
     *  <i>slow down</i> performance depending on the tree structure.
     * @param nodeFactory the {@link AbstractFileNode} to be used as a factory
     *  for creating other nodes; because this object will only be used to create other nodes,
     *  its internal file, content, etc. can be left as {@code null}
     * @param entryRenderer the {@link AbstractFileEntryRenderer} used to display each node;
     *  if {@code null}, the tree will be rendered identically to a regular {@link FileNodeToolTipTree}
     * @param nodePopup the {@link AbstractFileNodePopup} that should be displayed
     *  when a node is selected via a context event (e.g. right-clicking), or {@code null} if none
     * @param largeModel a boolean; if {@code true}, the tree will be optimized
     *  for very large node counts (e.g. hundreds or thousands of nodes)
     * @param equalRowHeights a boolean; if {@code true}, each row will be assumed
     *  to have equal height; enabling this option can <i>massively</i> speed up rendering
     *  performance for trees containing many children (especially if those children are all
     *  in the same directory), but prevents different children from having different heights
     * @param title the String title to display,
     *  as a key to be obtained from the localized resource bundle;
     *  if not {@code null}, a border will be created around the panel
     * @param preferredWidth the integer width of the panel, in pixels;
     *  if this or {@code preferredHeight} are less than or equal to zero, size is unrestricted
     * @param preferredHeight the integer height of the panel, in pixels;
     *  if this or {@code preferredWidth} are less than or equal to zero, size is unrestricted
     * @param args a variable number of {@link Object} arguments
     *  indicating text substitutions to be made in {@code title}
     */
    public PanelFileTree(AbstractFileNode nodeFactory, AbstractFileEntryRenderer entryRenderer,
            AbstractFileNodePopup nodePopup, boolean largeModel, boolean equalRowHeights,
            String title, int preferredWidth, int preferredHeight, Object... args) {
        super(title, preferredWidth, preferredHeight, args);
        this.nodeFactory = nodeFactory;
        this.entryRenderer = (entryRenderer == null) ? new DefaultFileEntryRenderer()
                : entryRenderer;
        this.equalRowHeights = equalRowHeights;

        setTree(nodePopup, largeModel);
        setNoResults();
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    public void setEnabled(boolean enabled) {
        tree.setEnabled(enabled);
    }

    @Override
    public void validate() {}

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Redraws the entire tree, applying node filtering if applicable.
     * Uses the last tree that was assigned via {@link #updateView(FileDisplayTree)}
     *  and retains the user's current view of expanded/collapsed nodes.
     * <p>
     * This method does nothing if there is no tree or it is not currently visible.
     * @return a boolean, {@code true} if the tree exists and is visible
     */
    public boolean redraw() {
        if (resultTree == null || !tree.isVisible()) { return false; }
        String state = ExpansionState.save(tree);
        updateView(resultTree);
        ExpansionState.restore(tree, state);
        return true;
    }

    /**
     * Updates the view associated with this panel.
     * Call this method whenever the list of files in this tree should change.
     * @param resultTree the {@link FileDisplayTree} whose contents should be displayed
     *  in this panel, or {@code null} if this panel's tree should be completely hidden
     */
    public void updateView(FileDisplayTree resultTree) {
        this.resultTree = resultTree;
        if (resultTree != null && resultTree.isEmpty()) {
            resultTree = null;
        }
        tree.setVisible(resultTree != null);
        scrollPane.setVisible(resultTree != null);
        noResults.setVisible(resultTree == null);
        if (resultTree == null) {
            return;
        }
        treeListener.setTree(resultTree);
        treeMouseListener.setTree(resultTree);

        if (equalRowHeights) {
            Component c = tree.getCellRenderer().getTreeCellRendererComponent(tree, root, false,
                    false, false, 1, false);
            tree.setRowHeight(c.getPreferredSize().height + 2); /*
                                                                 * Calculating the per-cell row
                                                                 *  height is a significant part
                                                                 *  of rendering the tree.
                                                                 * Setting a uniform height
                                                                 *  disables the calculation,
                                                                 *  speeding up performance.
                                                                 */
        }

        root.setUserObject(resultTree.getPath());
        root.removeAllChildren();
        root.add(nodeFactory.create(null, LazyFileLoadListener.FOLDER_SLUG));
        root.setToolTipText(HTML.wrap(Resources.__("mainNodeFolderTT",
                resultTree.size(),
                StringUtils.getFormattedBytes(resultTree.fileSize()))));

        ((DefaultTreeModel) tree.getModel()).reload();
        tree.collapseRow(0);

        scrollPane.setViewportView(tree);
    }



    /**
     * @return the {@link FileDisplayTree} currently being displayed in this panel,
     *  or {@code null} if none
     */
    public FileDisplayTree getTree() {
        return tree.isVisible() ? resultTree : null;
    }

    /** @return the {@link Color} used as the tree's background, if it is opaque */
    public Color getTreeBackground() {
        return tree.getBackground();
    }

    /** Sets the tree and scroll pane. */
    private void setTree(AbstractFileNodePopup nodePopup, boolean largeModel) {
        root = nodeFactory.create(null, "root"); //temporary node to prevent errors

        tree = new FileNodeToolTipTree(root);
        ToolTipManager.sharedInstance().registerComponent(tree);
        treeListener = new LazyFileLoadListener((DefaultTreeModel) tree.getModel(), nodeFactory);
        tree.setCellRenderer(new FileTreeCellRenderer(entryRenderer));
        treeMouseListener = new FileTreeMouseListener(tree, nodePopup);
        tree.addMouseListener(treeMouseListener);
        tree.addMouseMotionListener(treeMouseListener);
        tree.addKeyListener(new FileTreeKeyListener(tree, nodePopup));
        tree.addTreeWillExpandListener(treeListener);
        tree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
        tree.setUI(new WideTreeUI(tree, largeModel));
        tree.setVisible(false);

        scrollPane = new JScrollPane(tree);
        scrollPane.setPreferredSize(new Dimension(
                getPreferredSize().width - left() - right() - borderWidth(),
                getPreferredSize().height - top() - bottom() - borderHeight()));
        layoutFill(scrollPane, 5, panel, panel, panel, panel);
        add(scrollPane);
    }

    /** Creates the no-results label. */
    private void setNoResults() {
        noResults = createLabel("mainNoResults", null);
        noResults.setForeground(Color.RED);
        noResults.setVisible(false);
        layout(WEST, noResults, 5, WEST, panel);
        layout(NORTH, noResults, 5, NORTH, panel);
        add(noResults);
    }
}
