package edu.uwyo.crawl.ui.tree;

import java.awt.Rectangle;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JComponent;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.plaf.basic.BasicTreeUI;
import javax.swing.tree.AbstractLayoutCache.NodeDimensions;

/**
 * An extension of {@link BasicTreeUI} in which all nodes are given a width equal
 *  to the current tree width, excluding per-node indent amounts.
 * This UI allows e.g. {@link javax.swing.JPanel}-based nodes to have
 *  both left- and right-aligned text, which is not possible in the default ragged UI
 *  used by {@link FileNodeToolTipTree}.
 * In conjunction with {@link AbstractFileNode}, nodes can also be "highlighted" in a uniform way
 *  that will extend across the entire width of the tree,
 *  resulting in a more consistent visual experience.
 * <p>
 * This UI assumes that the owner {@link FileNodeToolTipTree} will be placed in a {@link JScrollPane},
 *  although node widths will still be extended normally without one.
 * @version 21 February, 2019
 * @author Christopher Stanton
 */
public class WideTreeUI extends BasicTreeUI {
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                   CONSTANTS                     ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /** A {link Runnable} used for reconfiguring the tree display. */
    private final Runnable configureRunnable = new Runnable() {
        @Override
        public void run() {
            configureLayoutCache();
        }
    };

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    FIELDS                       ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    private JScrollPane scrollPane;
    private boolean resize;

    private final FileNodeToolTipTree tree;
    private final boolean largeModel;

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                  CONSTRUCTORS                   ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Creates a new {@link WideTreeUI}.
     * Will automatically call {@link FileNodeToolTipTree#setUI(javax.swing.plaf.TreeUI)}.
     * @param tree the {@link FileNodeToolTipTree} that will contain this UI
     * @param largeModel a boolean; if {@code true}, tree rendering will be optimized
     *  for very large trees (otherwise, this should be {@code false})
     */
    protected WideTreeUI(FileNodeToolTipTree tree, boolean largeModel) {
        this.tree = tree;
        this.largeModel = largeModel;

        tree.addTreeExpansionListener(new CollapseListener());
        tree.setUI(this);
        setLargeModel(true);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //              OVERRIDDEN METHODS                 ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    @Override
    protected NodeDimensions createNodeDimensions() {
        if (largeModel && !isLargeModel()) {
            setLargeModel(true); /*
                                  * a large-model tree might initially be small with lazy-loading,
                                  *  which causes this flag to get reset internally
                                  */
        }
        return new WideNodeDimensionsHandler();
    }

    @Override
    public void installUI(JComponent c) {
        super.installUI(c);
        tree.addPropertyChangeListener("ancestor", new TreePropertyChangeListener());
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    METHODS                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Indicates that the size of the scroll pane or tree has changed significantly enough
     *  to trigger a recomputation of node dimensions.
     * <p>
     * Because this UI makes all nodes the same width as the tree,
     *  <i>the tree width can <u>never</u> decrease in any way.</i>
     * This behavior is problematic in two cases:
     * <ol>
     *  <li><b>When the window is resized:</b> if the window containing this tree is enlarged,
     *  the tree width will also be enlarged. But when the window is later made smaller,
     *  the tree width will <u>not</u> become smaller.</li>
     *  <li><b>When a node is collapsed:</b> if part of the tree is out of view (requiring a
     *  horizontal scroll bar) and a collapsed node removes the need for that scroll bar,
     *  the tree width will not shrink.</li>
     * </ol>
     * <p>
     * In both cases, the scroll pane will use the incorrect (too wide) width of the tree,
     *  resulting in a horizontal scroll bar even when the entire tree would be visible
     *  without scrolling.
     * More problematically, any right-aligned elements in the nodes will appear
     *  on this far right edge, requiring the user to scroll a potentially large distance
     *  to see them.
     * <p>
     * This method solves the problem by first setting all node widths to that
     *  of the scroll pane viewport, and then reconfiguring the node widths a second time
     *  to ensure that the scroll pane reappears if it is actually needed.
     * <p>
     * <i>To be effective, this fix must be used (1) whenever the window is resized
     *  and (2) whenever a node is collapsed.</i>
     */
    private void reportSizeChange() {
        resize = true;
        configureLayoutCache();
        resize = false;
        SwingUtilities.invokeLater(configureRunnable);
    }

    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    //                    CLASSES                      ////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * A listener that reconfigures the tree size when a node collapse occurs.
     * This listener is necessary to ensure that the horizontal scroll bar (if any)
     *  disappears and the tree width shrinks properly when a node is collapsed.
     */
    private class CollapseListener implements TreeExpansionListener {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public void treeExpanded(TreeExpansionEvent event) {}

        @Override
        public void treeCollapsed(TreeExpansionEvent event) {
            reportSizeChange();
        }
    }

    /**
     * A node dimension handler in which all nodes will have width sufficient to align their
     *  right edge with the tree's current right edge.
     * This effectively makes each node as wide as the tree itself, excluding any indented space.
     * <p>
     * Because node dimensions are cached by the tree, a property-change listener
     *  must be separately created to refresh the tree e.g. when scroll bars appear or disappear.
     */
    private class WideNodeDimensionsHandler extends NodeDimensionsHandler {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public Rectangle getNodeDimensions(Object value, int row, int depth, boolean expanded,
                Rectangle size) {
            Rectangle dimensions = super.getNodeDimensions(value, row, depth, expanded, size);

            int rowX = getRowX(row, depth);
            if (resize && scrollPane != null) {
                dimensions.width = scrollPane.getViewport().getWidth() - rowX;
            } else if (tree.getWidth() > rowX + dimensions.width) {
                dimensions.width = tree.getWidth() - rowX;
            }

            return dimensions;
        }
    }

    /**
     * A listener for updating the tree display when something inside this tree's owner
     *  {@link JScrollPane} changes.
     */
    private class ScrollComponentListener implements ComponentListener {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public void componentHidden(ComponentEvent e) {}

        @Override
        public void componentMoved(ComponentEvent e) {}

        @Override
        public void componentResized(ComponentEvent e) {
            reportSizeChange();
        }

        @Override
        public void componentShown(ComponentEvent e) {}
    }

    /**
     * A listener for updating the tree display when the scroll bars of this tree's
     *  owner {@link JScrollPane} are made visible/invisible or otherwise change configuration.
     */
    private class ScrollAdjustmentListener implements AdjustmentListener {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public void adjustmentValueChanged(AdjustmentEvent e) {
            configureLayoutCache();
        }
    }

    /**
     * A listener for updating the tree's owner {@link JScrollPane} if a property change occurs.
     * Used to ensure that any changes in the layout, such as window resizes, also redraw the tree.
     * <p>
     * This listener is also necessary to ensure that node widths are recomputed e.g. when the tree
     *  expands to a state that requires horizontal or vertical scrolling.
     * If this listener were not applied, any previously-drawn nodes would not be correctly updated
     *  when a scroll bar appears or disappears.
     */
    private class TreePropertyChangeListener implements PropertyChangeListener {
        ///////////////////////////////////////////////////////////////////////////////////////////
        /* OVERRIDDEN METHODS */
        ///////////////////////////////////////////////////////////////////////////////////////////

        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            if (tree.getParent() instanceof JViewport
                    && tree.getParent().getParent() instanceof JScrollPane) {
                scrollPane = (JScrollPane) tree.getParent().getParent();
                scrollPane.addComponentListener(new ScrollComponentListener());

                JScrollBar horizontal = scrollPane.getHorizontalScrollBar();
                horizontal.addAdjustmentListener(new ScrollAdjustmentListener());
            }
        }
    }
}
